package fr.inria.gemo.axml.optimax;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

import org.w3c.dom.Document;


import fr.inria.gemo.axml.optimax.adocument.ADNodeCounter;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.helping.Utilities;
import fr.inria.gemo.axml.optimax.helping.comparators.ComparatorSimpleNetwCost;
import fr.inria.gemo.axml.optimax.info.IDProvider;
import fr.inria.gemo.axml.optimax.info.P2PInformation;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;
import fr.inria.gemo.axml.optimax.info.Statistics;
import fr.inria.gemo.axml.optimax.io.FileManager;
import fr.inria.gemo.axml.optimax.io.output.ADocumentRelationshipGraphModule;
import fr.inria.gemo.axml.optimax.io.output.AxmlFileOutputModule;
import fr.inria.gemo.axml.optimax.io.output.GraphvizModule;
import fr.inria.gemo.axml.optimax.io.output.VerboseADocumentDrawingModule;
import fr.inria.gemo.axml.optimax.strategy.SearchAlgorithmAtom;
import fr.inria.gemo.axml.optimax.strategy.Step;
import fr.inria.gemo.axml.optimax.strategy.Strategy;

public class Optimax {
	private PeerInformationModule peerInformationModuleInstance = null;
//	private P2PInformation p2pInformationInstance = null;
	
	public Optimax(IDProvider givenIDProviderInstance, P2PInformation givenP2PInformationInstance)
	{
	    Parameters.init();
		peerInformationModuleInstance = PeerInformationModule.getInstance();	
		peerInformationModuleInstance.initialize(givenIDProviderInstance, givenP2PInformationInstance);
	}

	/**
	 * One PlanFinder will be created for each step of the strategy.
	 * @param givenDoc The original document.
	 * @param strat The strategy.
	 */
	public void optimizeAndStoreGraphizFiles(Document givenDoc, Strategy strat)
	{		
		PlanFinder current = null;
		PlanFinder previous = null;
		for (int i = 0; i < strat.steps.size(); i ++){
			
			System.out.println(">>> DOING STEP " + i);
			Step step = strat.steps.get(i);		
			if (i == 0){
				current = new PlanFinder(givenDoc);
				current.setStep(step);
			}
			else{
				current = new PlanFinder(previous, step);
			}
			
			// we mark at the statistics class that we
			// have moved to the following step
			Statistics.startOfSteps.add(Statistics.listWithCurrServiceAndWeightedNetworkCostOfAlreadyVisitedPlans.size());
			
			if(current.initialADoc == null)
				break;

			// we start the search
			current.searchForAllPlans(current.initialADoc);
			previous = current;
		}
		
		current.migrateToBeVisitedToAlreadyVisited();
		
		
		FileManager fileManagerInstance = FileManager.getInstance();
		// We check if we are going to print drawings of ADocuments
		if (!Parameters.getProperty("graphType").equals(Parameters.FALSE))
			fileManagerInstance.createOutputImages(current.algo.planRepositoryInstance);
		
		if (Parameters.getProperty("axmlFiles").equals(Parameters.TRUE))
			fileManagerInstance.createOptimizationFiles(current.algo.planRepositoryInstance, givenDoc, new StringBuffer("optim.xml"));
		
		fileManagerInstance.createGnuPlotGraphs(current.algo.planRepositoryInstance);
		
		ADocumentRelationshipGraphModule aDocumentRelationshipGraphModuleInst = new ADocumentRelationshipGraphModule();
		aDocumentRelationshipGraphModuleInst.createRelationshipGraph();
	}
	
	public Document optimize(Document givenDoc)
	{		
		return optimize(givenDoc, 0);
	}
	
	public Document optimize(Document givenDoc, int docNumber)
	{		
		Strategy strat = new Strategy(Parameters.strategyFileName);
		
		PlanFinder current = null;
		PlanFinder previous = null;
		
		for (int i = 0; i < strat.steps.size(); i ++){
			
			System.out.println(">>> DOING STEP " + i);
			Step step = strat.steps.get(i);		
			if (i == 0){
				current = new PlanFinder(givenDoc);
				current.setStep(step);
				
				if (!Parameters.getProperty("graphType").equals(Parameters.FALSE))
					VerboseADocumentDrawingModule.createOutputImage(current.initialADoc, "initialADoc");
				
				if (Parameters.getProperty("axmlFiles").equals(Parameters.TRUE))
					AxmlFileOutputModule.createAxmlFileFromRewrittenADocumentAndOriginalDom(current.initialADoc, givenDoc, new StringBuffer("initialAXML.xml"));
			}
			else{
				current = new PlanFinder(previous, step);
			}
		
			// we mark at the statistcs class that we
			// have moved to the following step
			Statistics.startOfSteps.add(Statistics.listWithCurrServiceAndWeightedNetworkCostOfAlreadyVisitedPlans.size());
			
			if(current.initialADoc == null)
				break;

			// we start the search
			current.searchForAllPlans(current.initialADoc);
			previous = current;
		}
		
		current.migrateToBeVisitedToAlreadyVisited();
		
		LinkedList<ADocument> results = new LinkedList<ADocument>();
		results.addAll(current.algo.planRepositoryInstance.aDocsAlreadyVisited);
		Collections.sort(results, new ComparatorSimpleNetwCost());
		
		if(docNumber < 0 || docNumber >= results.size())
			throw new IllegalArgumentException("docNumber is out of range. The range is between 0 and " + (results.size() - 1));
		
		if (!Parameters.getProperty("graphType").equals(Parameters.FALSE))
			VerboseADocumentDrawingModule.createOutputImage(results.get(docNumber), "selectedADoc");
		
		if (Parameters.getProperty("axmlFiles").equals(Parameters.TRUE))
			AxmlFileOutputModule.createAxmlFileFromRewrittenADocumentAndOriginalDom(results.get(docNumber), givenDoc, new StringBuffer("selectedAXML.xml"));
		
		
		return AxmlFileOutputModule.createAxmlDomFromRewrittenADocumentAndOriginalDom(results.get(docNumber), givenDoc);
	}
	
	public Document optimize(Document givenDoc, String strategyFileName)
	{
		Parameters.strategyFileName = strategyFileName;
		return optimize(givenDoc);
	}
	
	public Document optimize(Document givenDoc, String strategyFileName, int docNumber)
	{
		Parameters.strategyFileName = strategyFileName;
		return optimize(givenDoc, docNumber);
	}
}
