package fr.inria.gemo.axml.optimax;
import java.util.Enumeration;
import java.util.Properties;

import java.io.File;


public class Parameters {

	private static Properties props;
	
	public static String strategyFileName="strategy2.xml";

	public static final String INSTANTIATION="INST";
	public static final String CSE="CSE"; 
	public static final String DELEGATION="DELG";
	public static final String GENERALIZATION="GEN";
	public static final String COMPOSITION="COMP";
	public static final String BREADTH_FIRST="B";
	public static final String DEPTH_FIRST="D";
	public static final String SIMPLE_NETWORK_COST = "SNC";
	
	public static final String RECEIVE="receive";
	public static final String SEND="send";
	public static final String NEWNODE="newNode";
	public static final String GQS="gqs";
	public static final String STS="sts";
	
	public static final String ANY="any";
	
	public static final String TRUE="true";
	public static final String FALSE="false";
	
	public static String axmlURI = "http://futurs.inria.fr/gemo/axml/";
	public static String algebraURI = "http://futurs.inria.fr/gemo/axml/service/Algebra";
	public static String queryURI = "http://futurs.inria.fr/gemo/axml/service/Query";
	public static String streamSrvURI = "n/a";	
	
	public static String materializerForContinuousQueryURI = "fr.inria.gemo.axml.model.sc.materialization.MaterializerForContinuousQUERY";
	public static String materializerForSendURI = "fr.inria.gemo.axml.model.sc.materialization.MaterializerForSEND";
	public static String materializerForReceiveURI = "fr.inria.gemo.axml.model.sc.materialization.MaterializerForRECEIVE";
	
    // this will be modified to add the OS-dependent path separator
    // at Parameters.init()

	public static String OUTPUT_DIR="tmp";

	public static final String UNBOUNDED = "*";
	
	public static final String VERBOSE_GRAPH = "verbose";
	public static final String CONCISE_GRAPH = "concise";

	public static void init() {
	   
		props = new Properties();

		// initializing the default properties
		for (int i = 0; i < defaults.length; i++) {
			props.put(defaults[i][0].toUpperCase(), defaults[i][1]);
		}
		
		System.out.println("Path separator: " + props.get("PATHSEPARATOR"));
		
		File f = new File(OUTPUT_DIR);
		f.mkdirs();
		 
		OUTPUT_DIR = new String(OUTPUT_DIR + props.get("PATHSEPARATOR"));


		// overriding them with system properties
		Properties sysProps = System.getProperties();
		Enumeration propNames = sysProps.propertyNames();
		while (propNames.hasMoreElements()) {
			String pName = (String) propNames.nextElement();
			String pValue = (String) sysProps.get(pName);
			props.put(pName.toUpperCase(), pValue);
		}
	}

	public static String getProperty(String propName){
		if (props == null){
			throw new Error("Properties should have been initialized");
		}
		String res = (String)props.get(propName.toUpperCase());
		if (res == null){
			throw new Error("No property named " + propName);
		}
		return res;
	}

	public static void setProperty(String propName, String propValue) {
		props.put(propName.toUpperCase(), propValue);
	}

	private static String defaults[][] = {
		/**
		 * Directory (under the working dir) where the temporary files should go
		 *  
		 */
		{ "temporaryFileDir", "tmp" },
		
		/**
		 * Detailed graphs, for debugging, or concise graphs for showing
		 * Possible values: Parameters.VERBOSE_GRAPH, Parameters.CONCISE_GRAPH,
		 * Parameters.FALSE
		 */
		{ "graphType", Parameters.VERBOSE_GRAPH},

		/**
		 * Path separator string - for VMs that are totally wrong...
		 */
		{ "pathSeparator", "/"},

		/**
		 * Draw graph cost vs number of plans? The possible values here
		 * can be Parameters.TRUE || Parameters.FALSE 
		 */
		{ "drawCost", Parameters.FALSE},

		/**
		 * Draw aDocument relation graph. The possible values here can be
		 * Parameters.FALSE, Parameters.CONCISE_GRAPH, Parameters.VERBOSE_GRAPH
		 */
		{ "drawRel", Parameters.FALSE},
		
		/**
		 * CSE should be complete or be applied only when it reduces th number
		 * of simple scs. Possible values: Parameters.TRUE || Parameters.FALSE 
		 */
		{ "completeCSE", Parameters.TRUE},
		
		/**
		 * Do we want to print the optimized axml files?
		 * Possible values: Parameters.TRUE || Parameters.FALSE 
		 */
		{ "axmlFiles", Parameters.FALSE}
	};
		
}







