package fr.inria.gemo.axml.optimax;

import java.util.*;

import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Document;

import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;
import fr.inria.gemo.axml.optimax.io.FileManager;
import fr.inria.gemo.axml.optimax.strategy.Step;
import fr.inria.gemo.axml.optimax.strategy.Strategy;

public class PlanFinder extends Object{
	Strategy strat;
	
	// 1 PlanFinder = 1 step = 1 algo
	SearchAlgorithm algo;
	Step step;
	
	Document originalDocument;
	ADocument initialADoc = null;;
	
	/**
	 * Constructor that takes the repository of the previous finder
	 * @param previousFinder
	 */
	PlanFinder(PlanFinder previousFinder, Step step)
	{
		this.originalDocument = previousFinder.originalDocument;
		this.algo = new SearchAlgorithm();
		this.setStep(step);
		// this has created the new Repository of the new Algorithm
		
		// now add all the plans and data from the previous rep, in the new rep. The purpose here
		// is to get the new order in the aDocsToBeVisited list
		this.algo.planRepositoryInstance.copyDataFromGivenPlanFinder(previousFinder.algo.planRepositoryInstance);
		
		if(this.algo.planRepositoryInstance.aDocsToBeVisited.isEmpty())
			this.initialADoc = null;
		else
			this.initialADoc = this.algo.planRepositoryInstance.removeADocToBeVisitedFromFront();
	}
	
	public final void setStep(Step step){
		this.step = step;
		this.algo.setStep(step);
	}
	
	PlanFinder(Document givenDocument)
	{
		originalDocument = givenDocument;
		algo = new SearchAlgorithm();
		
		ADocument skeletonOfDoc = null;
		try {
			skeletonOfDoc = FileManager.constructADocFromAxmlDom(givenDocument);
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		PeerInformationModule PeerInformationModuleInstance = PeerInformationModule.getInstance();
		StringBuffer specificPeer = new StringBuffer(PeerInformationModuleInstance.getCurrPeerId());
		ADocument skeletonOfDocAtSpecificPeer = new ADocument(skeletonOfDoc);
		skeletonOfDocAtSpecificPeer.setPeerThatHostsADoc(specificPeer);
		
		System.out.println(skeletonOfDoc);
		System.out.println(skeletonOfDocAtSpecificPeer);
		
		skeletonOfDocAtSpecificPeer.calculateHashCode();
		skeletonOfDocAtSpecificPeer.calculateNetworkCost();
		initialADoc = skeletonOfDocAtSpecificPeer;
	}
	

	//	 Main program loop
	public void searchForAllPlans(ADocument givenADoc)
	{

		System.out.println("SEARCHING FOR ALL PLANS, algo.limit is " + algo.limit);
		algo.newSearch(givenADoc);
		// variable introduced by Ioana to control the number of plans
		int allPlans = 0;
		
		//		for(int sp=0; (sp<100 && strategyModuleInstance.applyRewrittingStep(true, true)); sp++)
		while(allPlans < algo.limit && algo.applyRewritingStep(true, true))
			{
				System.out.println("ALLPLANS: " + allPlans);
			
				allPlans ++;
				System.out.println("availableProcessors " + Runtime.getRuntime().availableProcessors());
				System.out.println("freeMemory(MB)  " + (float)((float)Runtime.getRuntime().freeMemory()/(float)1000000));
				System.out.println("maxMemory(MB)   " + (float)((float)Runtime.getRuntime().maxMemory()/(float)1000000));
				System.out.println("totalMemory(MB) " + (float)((float)Runtime.getRuntime().totalMemory()/(float)1000000));
			}
//		}
		System.out.println("ALLPLANS: " + allPlans + " limit: " + algo.limit);
	}
	
	boolean isThereAloopWithTheSends(Service givenService, List<String> peersOfPreviousSends)
	{
		if(givenService.getServiceName().toString().equals("send"))
		{
			if(peersOfPreviousSends.contains(givenService.getExecutionPeer().toString()))
				return true;
		}
		
		List<String> copyOfPeersOfPreviousSends;
		
		if(givenService.getServiceName().toString().equals("send"))
		{
			copyOfPeersOfPreviousSends = new ArrayList<String>(peersOfPreviousSends);
			copyOfPeersOfPreviousSends.add(givenService.getExecutionPeer().toString());
		}

		else
			copyOfPeersOfPreviousSends = new ArrayList<String>();
		
		Iterator cldIterator = givenService.getChildren().iterator();
		while(cldIterator.hasNext())
		{
			ADNode tempNodeRef = (ADNode) cldIterator.next();
			if(tempNodeRef instanceof Service)
			{
				Service tempServiceRef = (Service) tempNodeRef;
				if(isThereAloopWithTheSends(tempServiceRef, copyOfPeersOfPreviousSends))
					return true;
			}
		}
		return false;
	}
	
	/**
	 * After the last step of a strategy has run, we would like the 
	 * plans toBeVisited at that moment to be drawn.
	 * Normally, when there is only one step,
	 * the alreadyVisited plans are the only one to be drawn,
	 * and the toBeVisited collection is empty.
	 * However, to accomodate many-steps strategies that still do not
	 * exhaust the possibilities, we may move all plans to the
	 * alreadyVisited collection.
	 */
	public void migrateToBeVisitedToAlreadyVisited(){
		PlanRepository rep = algo.planRepositoryInstance;
		Iterator<ADocument> it = rep.aDocsToBeVisited.iterator();
		while (it.hasNext()){
			rep.aDocsAlreadyVisited.add(it.next());
		}
	}
}
