package fr.inria.gemo.axml.optimax;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;

import fr.inria.gemo.axml.optimax.adocument.ADNodeCounter;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.EqualADocAtom;
import fr.inria.gemo.axml.optimax.adocument.RuleDescriptionContainer;
import fr.inria.gemo.axml.optimax.info.Statistics;

public class PlanRepository {

	Comparator givenCompar;
	
	public LinkedHashSet<ADocument> aDocsAlreadyVisited;
	public LinkedList<ADocument> aDocsToBeVisited;
 	
	private void displayDocsAlreadyVisited(){
		System.out.println("Already visited documents: " + aDocsAlreadyVisited.size());
	}
	
	public PlanRepository(Comparator givenComp)
	{
		this.givenCompar = givenComp;
		aDocsToBeVisited = new LinkedList<ADocument>();
		aDocsAlreadyVisited = new LinkedHashSet<ADocument>();
	}
	
	public PlanRepository(Comparator givenComp, LinkedHashSet<ADocument> givenADocsAlreadyVisited)
	{
		this.givenCompar = givenComp;
		aDocsToBeVisited = new LinkedList<ADocument>();
		aDocsAlreadyVisited = new LinkedHashSet<ADocument>();
	}
	
	public void resetPlanRepository(ADocument givenADoc) 
	{ 
		if(givenADoc == null)
		{
			System.out.println("xoxoxo");
			System.exit(0);
		}
		
		aDocsAlreadyVisited.clear();
		aDocsToBeVisited.clear();
		aDocsToBeVisited.add(givenADoc);
	}
	
	public void preparePlanRepositoryForNextStep(ADocument givenADoc)
	{
		aDocsToBeVisited.add(givenADoc);
	}
	
	public void clearPlanRepository() 
	{ 
		aDocsAlreadyVisited = null;
		aDocsToBeVisited = null;
		givenCompar = null;
	}
	
	public LinkedHashSet<ADocument> getADocsAlreadyVisitedSet() { return aDocsAlreadyVisited; }
	public int getADocsAlreadyVisitedSetSize() { return aDocsAlreadyVisited.size(); }
	public boolean addADocAlreadyVisited(ADocument givenADoc) { return aDocsAlreadyVisited.add(givenADoc); }
	public boolean doesADocExistInAlreadyVisitedSet(ADocument givenADoc) { return aDocsAlreadyVisited.contains(givenADoc); }
	
	public LinkedList<ADocument> getADocsToBeVisited() { return aDocsToBeVisited; }
	public int getADocsToBeVisitedSize() { return aDocsToBeVisited.size(); }

	public ADocument removeADocToBeVisitedFromFront()
	{
		return aDocsToBeVisited.removeFirst();
		//ADocument ad = aDocsToBeVisited.first();
		//aDocsToBeVisited.remove(ad);
		//return ad;
	}
	
	// This method will copy data from givenPlanFinder to this. The data
	// will be the same as in givenPlanFinder. The only thing that is not
	// copied is the comparator. The docs in the aDocsToBeVisited list
	// will have the order of the new comparator
	public void copyDataFromGivenPlanFinder(PlanRepository givenPlanRepository)
	{
		this.aDocsAlreadyVisited.clear();
		this.aDocsAlreadyVisited.addAll(givenPlanRepository.aDocsAlreadyVisited);
		
		// get the docs to be visited
		this.aDocsToBeVisited.clear();
		this.addCollectionToADocToBeVisited(givenPlanRepository.aDocsToBeVisited);
	}
	
	/**
	 *  This method adds givenCollection elements in the ADocToBeVisited
	 */ 
	public void addCollectionToADocToBeVisited(Collection<ADocument> givenCollection)
	{
		System.out.println("Adding to toBeVisited collection "  + givenCollection.size() + 
				" plans");
		System.out.println("Before the addition there were " + aDocsToBeVisited.size() + " plans");
		
		Iterator<ADocument> givenCollectionIter = givenCollection.iterator();
		
		// The complexity of the following loop is: the size of the added collection
		int i = 0;
		while(givenCollectionIter.hasNext())
		{
			ADocument currADoc = givenCollectionIter.next();
			//TODO maybe with more sophisticated rewriting rules !aDocsToBeVisited.contains(currADoc)
			// can be avoided
			
			// We update totalRulesApplied. We are going to insert the rule that was
			// aplied to the previous doc of currADoc in order to get currADoc
			Statistics.totalRulesApplied.add(currADoc.getRulesApplied().get(currADoc.getRulesApplied().size()-1));
			
			if(!aDocsAlreadyVisited.contains(currADoc) && !aDocsToBeVisited.contains(currADoc)){
				aDocsToBeVisited.add(currADoc);
				
				// We update the needed statistic data
				if(Parameters.getProperty("drawCost").equals(Parameters.TRUE))
				{
					// we update the minimumSimpleCostOfAlreadyVisitedPlans if needed 
					if(Statistics.minimumSimpleCostOfAlreadyVisitedPlans == -1 || Statistics.minimumSimpleCostOfAlreadyVisitedPlans > currADoc.getSimpleNetworkCost())
					{
						Statistics.minimumSimpleCostOfAlreadyVisitedPlans = currADoc.getSimpleNetworkCost();
						Statistics.cheapestSimpleCostADoc = currADoc;
					}
					
					// we update the minimumRealCostOfAlreadyVisitedPlans if needed
					if(Statistics.minimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans == -1 || Statistics.minimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans > currADoc.getServiceAndWeightedNetworkCost1())
					{
						Statistics.minimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans = currADoc.getServiceAndWeightedNetworkCost1();
						Statistics.cheapestServiceAndWeightedNetworkCostADoc = currADoc;
					}
				
					// We update the structures that keep the minimum Adocument cost until now 
					Statistics.listWithMinimumSimpleCostOfAlreadyVisitedPlans.add(Statistics.minimumSimpleCostOfAlreadyVisitedPlans);
					Statistics.listWithMinimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans.add(Statistics.minimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans);
					
					// We update the structures that keep the currentADoc cost 
					Statistics.listWithCurrSimpleCostOfAlreadyVisitedPlans.add(currADoc.getSimpleNetworkCost());
					Statistics.listWithCurrServiceAndWeightedNetworkCostOfAlreadyVisitedPlans.add(currADoc.getServiceAndWeightedNetworkCost1());
				}
				
				System.out.println(i + ": added one plan to be visited");
			}
			else{
				EqualADocAtom idPairOfEqualPlans = null;
				
				if(!Parameters.getProperty("drawRel").equals(Parameters.FALSE))
				{
					idPairOfEqualPlans = new EqualADocAtom();
					idPairOfEqualPlans.numberOfADoc1 = currADoc.getADocNumber();
				}
				
				if (aDocsAlreadyVisited.contains(currADoc)){
					System.out.println(i + ": this plan has already been visited");
					if(!Parameters.getProperty("drawRel").equals(Parameters.FALSE))
					{
						Iterator<ADocument> aDocsAlreadyVisitedIter= aDocsAlreadyVisited.iterator();
						while(aDocsAlreadyVisitedIter.hasNext())
						{
							ADocument currADocAlreadyVisited = aDocsAlreadyVisitedIter.next();
							if(currADocAlreadyVisited.equals(currADoc))
							{
								idPairOfEqualPlans.numberOfADoc2 = currADocAlreadyVisited.getADocNumber();
								break;
							}
						}
					}
				}
				if (aDocsToBeVisited.contains(currADoc)){
					System.out.println(i + ": this plan is already on the list to be visited");
					if(!Parameters.getProperty("drawRel").equals(Parameters.FALSE))
					{
						int positionOfEqualADocOfCurrADoc = aDocsToBeVisited.indexOf(currADoc);
						ADocument equalADocOfCurrADoc = aDocsToBeVisited.get(positionOfEqualADocOfCurrADoc);
						idPairOfEqualPlans.numberOfADoc2 = equalADocOfCurrADoc.getADocNumber();
					}
				}
				if(!Parameters.getProperty("drawRel").equals(Parameters.FALSE))
					Statistics.aDocEqualities.add(idPairOfEqualPlans);
			}
			
			i ++;
		}
		System.out.println("Now there are " + aDocsToBeVisited.size() + " docs to be visited");
		System.out.println("Sorting them...");
		Collections.sort(aDocsToBeVisited, this.givenCompar);
	}
	public boolean isADocsToBeVisitedSetEmpty() { return aDocsToBeVisited.isEmpty(); }
	public Comparator getComparator()
	{
		return givenCompar;
	}
	
	public int calculateNodesAlreadyVisited() { return ADNodeCounter.calculateADNodesOfADocumentSet(aDocsAlreadyVisited); }
}
