package fr.inria.gemo.axml.optimax;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.helping.ServiceMapper;
import fr.inria.gemo.axml.optimax.helping.comparators.ComparatorSearchBreadth;
import fr.inria.gemo.axml.optimax.helping.comparators.ComparatorSearchDepth;
import fr.inria.gemo.axml.optimax.helping.comparators.ComparatorSimpleNetwCost;
import fr.inria.gemo.axml.optimax.helping.comparators.ComposedComparator;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;
import fr.inria.gemo.axml.optimax.info.Statistics;
import fr.inria.gemo.axml.optimax.io.FileManager;
import fr.inria.gemo.axml.optimax.rules.CSE;
import fr.inria.gemo.axml.optimax.rules.Composition;
import fr.inria.gemo.axml.optimax.rules.Delegation;
import fr.inria.gemo.axml.optimax.rules.Instantiation;
import fr.inria.gemo.axml.optimax.strategy.AtomStep;
import fr.inria.gemo.axml.optimax.strategy.CardinalityAtom;
import fr.inria.gemo.axml.optimax.strategy.RepetitionStep;
import fr.inria.gemo.axml.optimax.strategy.RuleAtom;
import fr.inria.gemo.axml.optimax.strategy.SearchAlgorithmAtom;
import fr.inria.gemo.axml.optimax.strategy.Step;

/**
 * This class encapsulates the rewriting logic of a strategy step:
 * - the search algorithm (breadth-first, depth-first). This is controlled via the
 * comparator of the planRepositoryInstance.
 * - the rules allowed during this step. This is controlled via a few boolean attributes.
 * - the number of successive rewritings allowed. This is controlled via a long parameter.
 * 
 * This information is taken from a Step object.
 * 
 * @author Spyros
 *
 */
public class SearchAlgorithm {

	PlanRepository planRepositoryInstance;
	Instantiation instatiationInstance;
	Composition compositionInstance;
	Delegation delegationInstance;
	CSE cseInstance;
	
	ServiceMapper serviceMapperInstance;
	List<StringBuffer> listOfPeers;
	public Step step;
	
	/** 
	 * true if this algo has the right to apply instantiation 
	 */
	boolean instOK;
	/**
	 * true if this algo has the right to apply C
	 */
	boolean compOK;
	/**
	 * true if this algo has the right to apply delegation
	 */
	boolean delegOK;
	/**
	 * true if this algo has the right to apply CSE
	 */
	boolean CSEOK;
	
	/**
	 * how many plans this algo is allowed to develop
	 */
	long limit;
	
	/**
	 * how many plans have already been developed
	 */
	long alreadyDeveloped;
	
	SearchAlgorithm() {
		planRepositoryInstance = null;
		instatiationInstance = new Instantiation();
		compositionInstance = new Composition();
		delegationInstance = new Delegation();
		cseInstance = new CSE();
		
		PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
		listOfPeers = peerInformationModuleInstance.getListOfPeers();
	}
	
	/**
	 * Ioana
	 * This method informs the algorithm (out of the Step information)
	 * of what it should do:
	 * - what search algorithm
	 * - what rules
	 * - how many times
	 * @param step
	 */
	public void setStep(Step step){
		this.step = step;
			
		LinkedList<SearchAlgorithmAtom> algoAtoms = step.algos;
		Comparator compOfThisRepository = null;
		for (int i = 0; i < algoAtoms.size(); i ++){
			Comparator ci = this.makeComparator(algoAtoms.get(i));
			if (compOfThisRepository == null){
				compOfThisRepository = ci;
			}
			else{
				compOfThisRepository = new ComposedComparator(compOfThisRepository, ci);
			}
		}
		
		this.planRepositoryInstance = new PlanRepository(compOfThisRepository);
		
		AtomStep as = null;
		RepetitionStep rs = null;
		try{
			as = (AtomStep) step;
			RuleAtom ra = as.a;
			limit = 1;
			String s = ra.name;
			
			if (s.compareTo(Parameters.INSTANTIATION)==0){
				this.instOK = true;
			}
			else{
				if (s.compareTo(Parameters.COMPOSITION)==0){
					this.compOK = true;
				}
				else{
					if (s.compareTo(Parameters.DELEGATION)==0){
						this.delegOK = true;
					}
					else{
						if (s.compareTo(Parameters.CSE)==0){
							this.CSEOK = true;
						}
					}
				}
			}
		}
		catch(ClassCastException e){
			rs = (RepetitionStep)step;
			for (int i = 0; i < rs.atoms.size(); i ++){
				RuleAtom ra = (RuleAtom)rs.atoms.get(i);
				String s = ra.name;
				
				if (s.compareTo(Parameters.INSTANTIATION)==0){
					this.instOK = true;
					System.out.println("INSTANTIATION ALLOWED");
				}
				else{
					if(s.compareTo(Parameters.COMPOSITION)==0){
						this.compOK = true;
						System.out.println("COMPOSITION ALLOWED");
					}
					else {
						if (s.compareTo(Parameters.DELEGATION)==0){
							this.delegOK = true;
							System.out.println("DELEGATION ALLOWED");
						}
						else{
							if (s.compareTo(Parameters.CSE)==0){
								this.CSEOK = true;
								System.out.println("CSE ALLOWED");
							}
						}
					}
				}
			}
			this.limit = rs.card.max;
		}
	}
	/**
	 * Ioana
	 * @param rep, the new repository to use (includes the comparator).
	 */
	public void setRepository(PlanRepository rep){
		this.planRepositoryInstance = rep;
	}
//	/**
//	 * Ioana
//	 * @param newRepository
//	 */
//	public SearchAlgorithm(PlanRepository newRepository){
//		planRepositoryInstance = newRepository;
//		instatiationInstance = new Instantiation();
//		delegationInstance = new Delegation();
//		cseInstance = new CSE();
//		
//		PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
//		listOfPeers = peerInformationModuleInstance.getListOfPeers();
//	}
	
	public void newSearch(ADocument givenADoc)
	{
		if(planRepositoryInstance == null)
		{
			System.out.println("There should be a valid planRepositoryInstance");
			System.exit(0);
		}
		planRepositoryInstance.preparePlanRepositoryForNextStep(givenADoc);
//		givenADoc.calculateNetworkCost();
	}
	
	public boolean applyRewritingStep(boolean consistencyCheck, boolean printPlanRepositoryStatistics)
	{
		if(planRepositoryInstance.isADocsToBeVisitedSetEmpty())
			return false;
		
		ADocument currentADoc = planRepositoryInstance.removeADocToBeVisitedFromFront();
		
		if(consistencyCheck && !currentADoc.isADocumentConsistent())
		{
			System.err.println("The following document is inconsistent!!!!");
			System.err.println(currentADoc);
			FileManager.createDebugOutputImage(currentADoc, "debug");
			System.exit(0);
		}
		
		//System.out.println("Considering a plan of cost: " + currentADoc.getSimpleNetworkCost());
		
		// TODO the check bellow should be deleted. It's here for security reasons
		if(planRepositoryInstance.doesADocExistInAlreadyVisitedSet(currentADoc))
		{
			System.err.println("Something is going wrong with the insertion at the ADocsToBeVisited list");
			System.err.println("currentADoc");
			System.err.println(currentADoc + "," + currentADoc.hashCode());
			Iterator<ADocument> aDocsAlreadyVisitedIter = planRepositoryInstance.aDocsAlreadyVisited.iterator();
//			FileManager.createDebugOutputImage(currentADoc, "debug0");
			int counter = 1;
			while(aDocsAlreadyVisitedIter.hasNext())
			{
				ADocument tempADoc = aDocsAlreadyVisitedIter.next();
				if(currentADoc.equals(tempADoc))
					System.err.println("!!#" + counter);
				else
					System.err.println("#" + counter);
//				FileManager.createDebugOutputImage(tempADoc, new String("debug" + counter));
				System.err.println(tempADoc + "," + tempADoc.hashCode());
				counter++;
			}
			System.exit(0);
		}
		
		if(planRepositoryInstance.addADocAlreadyVisited(currentADoc) != true)
		{
			System.err.println("Something is going wrong with the insertion at the ADocsToBeVisited list");
			System.err.println("currentADoc");
			System.exit(0);
		}
		
		List<ADocument> returnedADocs = applyTransformationStep(currentADoc, consistencyCheck);
		planRepositoryInstance.addCollectionToADocToBeVisited(returnedADocs);
		
		if(printPlanRepositoryStatistics)
			System.out.println("aDocsToBeVisited(" + planRepositoryInstance.getADocsToBeVisitedSize() + "), aDocsAlreadyVisited(" + planRepositoryInstance.getADocsAlreadyVisitedSetSize() + ")");
		
		return true;
	}
	
	private List<ADocument> applyTransformationStep(ADocument currentADoc, boolean consistencyCheck) 
	{
		List<ADocument> newPlans = new ArrayList<ADocument>();	
		
		// Apply instatiation rule
		if (this.instOK){
			System.out.println("INSTATIATING!");
			newPlans.addAll(instatiationInstance.instatiationForADocs(currentADoc));
		}
		
		// Apply the simple delegation rule
		if (this.delegOK){
			System.out.println("DELEGATING!");
			newPlans.addAll(delegationInstance.delegationRuleForADocs(currentADoc, listOfPeers));
		}

		// Apply the query composition rule
		if (this.compOK){
			System.out.println("COMPOSITING");
			newPlans.addAll(compositionInstance.queryComposition(currentADoc));
		}
		
		// Apply the common subexpression elimination rule (factorization)
		if (this.CSEOK){
			System.out.println("CSE!");
			newPlans.addAll(cseInstance.commonSubExpressionElimination(currentADoc));
		}
		return newPlans;
	}

	private Comparator makeComparator(SearchAlgorithmAtom algoAtom){
		if (algoAtom.name.compareTo(Parameters.DEPTH_FIRST)==0){
			// depth first
			return new ComparatorSearchDepth();
		}
		else{
			if (algoAtom.name.compareTo(Parameters.BREADTH_FIRST)==0){
				// breadth first
				return new ComparatorSearchBreadth();
			}
			else{
				if (algoAtom.name.compareTo(Parameters.SIMPLE_NETWORK_COST)==0){
					// simple network cost
					return new ComparatorSimpleNetwCost();
				}
				else{
					throw new Error("Unknown search algorithm: " + algoAtom.name);
				}
			}
		}
	}
	
}
