package fr.inria.gemo.axml.optimax.adocument;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import fr.inria.gemo.axml.optimax.helping.DinstinctNumberGenerator;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;
import fr.inria.gemo.axml.optimax.io.output.GraphvizModule;
import fr.inria.gemo.axml.optimax.io.output.VerboseADocumentDrawingModule;
import fr.inria.gemo.axml.optimax.rules.UsefulMethods;
 
public class ADocument implements Comparable {
	private ADNode rootOfADoc;
	private StringBuffer peerThatHostsADoc;
	private int hashCodeValue;
	private float simpleNetworkCost;
	private float serviceAndWeightedNetworkCost1;
	private int numberOfRewritings;
	private int numberOfGeneralizationInstatiations;
	private int numberOfDelegations;
	private int numberOfCSE;
	private int numberOfNodes;
	private int aDocNumber;
	private List<RuleDescriptionContainer> rulesApplied;
	
	public Map<String,String> oldScCurrNodeIdsToNew = null;
	public Map<String,String> newScCurrNodeIdsToOld = null;
	
	public Map<IdContainer, ADNode> originalIdsOfNodesNotUnderReceivesToADNodeRefs = null;
	
	public ADocument()
	{
		rootOfADoc = null;
		peerThatHostsADoc = new StringBuffer();
		hashCodeValue = 0;
		simpleNetworkCost = 0;
		serviceAndWeightedNetworkCost1 = 0;
		numberOfRewritings = 0;
		numberOfGeneralizationInstatiations = 0;
		numberOfDelegations = 0;
		numberOfCSE = 0;
		numberOfNodes = 0;
		aDocNumber = (new DinstinctNumberGenerator()).getAnADocNumberInt();
		rulesApplied = new ArrayList<RuleDescriptionContainer>();
		oldScCurrNodeIdsToNew = new HashMap<String,String>();
		newScCurrNodeIdsToOld = new HashMap<String,String>();
		originalIdsOfNodesNotUnderReceivesToADNodeRefs = new HashMap<IdContainer, ADNode>();
	}
	
	public ADocument(ADocument givenADoc)
	{
		ADocument newADoc = copyADoc(givenADoc);
		rootOfADoc = newADoc.rootOfADoc;
		peerThatHostsADoc = newADoc.peerThatHostsADoc;
		hashCodeValue = newADoc.hashCodeValue;
		simpleNetworkCost = newADoc.simpleNetworkCost;
		serviceAndWeightedNetworkCost1 = newADoc.serviceAndWeightedNetworkCost1;
		numberOfRewritings = newADoc.numberOfRewritings;
		numberOfGeneralizationInstatiations = newADoc.numberOfGeneralizationInstatiations;
		numberOfDelegations = newADoc.numberOfDelegations;
		numberOfCSE = newADoc.numberOfCSE;
		numberOfNodes = newADoc.numberOfNodes;
		aDocNumber = (new DinstinctNumberGenerator()).getAnADocNumberInt();
		rulesApplied = newADoc.rulesApplied;
		oldScCurrNodeIdsToNew = newADoc.oldScCurrNodeIdsToNew;
		newScCurrNodeIdsToOld = newADoc.newScCurrNodeIdsToOld;
		originalIdsOfNodesNotUnderReceivesToADNodeRefs = newADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs;
		
		// we update the setReferenceToADocument of all the
		// nodes of the document to *this*
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(rootOfADoc);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this);
		}
		
//		System.out.println("ADocument(ADocument givenADoc) oldServiceCurrNodeIdsToNew :" + oldServiceCurrNodeIdsToNew);
//		System.out.println("ADocument(ADocument givenADoc) newServiceCurrNodeIdsToOld :" + newServiceCurrNodeIdsToOld);
	}
	
	public ADNode getRootOfADoc()
	{
		return rootOfADoc;
	}
	
	public float getSimpleNetworkCost()
	{
		return simpleNetworkCost;
	}
	
	public float getServiceAndWeightedNetworkCost1()
	{
		return serviceAndWeightedNetworkCost1;
	}
	
	public int getNumberOfRewritings()
	{
		return numberOfRewritings;
	}
	
	public int getNumberOfGeneralizationInstatiations()
	{
		return numberOfGeneralizationInstatiations;
	}
	
	public int getNumberOfDelegations()
	{
		return numberOfDelegations;
	}
	
	public int getNumberOfCSE()
	{
		return numberOfCSE;
	}
	
	public int getNumberOfNodes()
	{
		return numberOfNodes;
	}
	
	public StringBuffer getPeerThatHostsADoc()
	{
		return peerThatHostsADoc;
	}
	
	public int getADocNumber()
	{
		return aDocNumber;
	}
	
	// TODO for now this is used only in equivalence
	// TODO check. If we want to use it somewhere else
	// TODO we should create a valid instance at least to
	// TODO originalIdsOfNodesNotUnderReceivesToADNodeRefs
	public void setRootOfADoc(ADNode givenRoot)
	{
		rootOfADoc = copyADNodes(givenRoot);
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(rootOfADoc);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this);
		}
	}
	
	public void setPeerThatHostsADoc(StringBuffer givenPeer)
	{
		peerThatHostsADoc = new StringBuffer(givenPeer);
	}

	public static ADocument copyADoc(ADocument originalADoc)
	{
		ADocument newADoc = new ADocument();
		newADoc.oldScCurrNodeIdsToNew = new HashMap<String,String>(originalADoc.oldScCurrNodeIdsToNew);
		newADoc.newScCurrNodeIdsToOld = new HashMap<String,String>(originalADoc.newScCurrNodeIdsToOld);
		newADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs = new HashMap<IdContainer, ADNode>();
		newADoc.rootOfADoc = copyADNodes(originalADoc.rootOfADoc);
		newADoc.peerThatHostsADoc = new StringBuffer(originalADoc.peerThatHostsADoc);
		newADoc.hashCodeValue = originalADoc.hashCodeValue;
		newADoc.simpleNetworkCost = originalADoc.simpleNetworkCost;
		newADoc.serviceAndWeightedNetworkCost1 = originalADoc.serviceAndWeightedNetworkCost1;
		newADoc.numberOfRewritings = originalADoc.numberOfRewritings;
		newADoc.numberOfGeneralizationInstatiations = originalADoc.numberOfGeneralizationInstatiations;
		newADoc.numberOfDelegations = originalADoc.numberOfDelegations;
		newADoc.numberOfCSE = originalADoc.numberOfCSE;
		newADoc.numberOfNodes = originalADoc.numberOfNodes;
		newADoc.aDocNumber = originalADoc.aDocNumber;
		
		// TODO
		// What is writen bellow has to be done in a more wise way
		
		// We insert the new records at originalIdsOfNodesNotUnderReceivesToADNodeRefs
		// of newDoc
		LinkedList<ADNode> oldNodesToBeVisited = new LinkedList<ADNode>();
		LinkedList<ADNode> newNodesToBeVisited = new LinkedList<ADNode>();
		oldNodesToBeVisited.add(originalADoc.rootOfADoc);
		newNodesToBeVisited.add(newADoc.rootOfADoc);
		while(!newNodesToBeVisited.isEmpty())
		{
			ADNode currOldNode = oldNodesToBeVisited.removeFirst();
			ADNode currNewNode = newNodesToBeVisited.removeFirst();
			
			oldNodesToBeVisited.addAll(currOldNode.getChildren());
			newNodesToBeVisited.addAll(currNewNode.getChildren());
			
			if(originalADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.containsValue(currOldNode))
			{
				IdContainer id = new IdContainer(currOldNode.getOrigNodeID().toString(), currOldNode.getOrigDocID().toString(), currOldNode.getOrigPeerID().toString());
				if(originalADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(id) == currOldNode)
					newADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(id, currNewNode);
			}
		}
		
		// We create the originalIdWithHostAddressesThatPointToThisNode of the nodes
		// of the newDoc
		// we also update the pointers of the service nodes - simple nodes of newDoc to
		// point to the new Doc 
		oldNodesToBeVisited = new LinkedList<ADNode>();
		newNodesToBeVisited = new LinkedList<ADNode>();
		oldNodesToBeVisited.add(originalADoc.rootOfADoc);
		newNodesToBeVisited.add(newADoc.rootOfADoc);
		while(!newNodesToBeVisited.isEmpty())
		{
			ADNode currOldNode = oldNodesToBeVisited.removeFirst();
			ADNode currNewNode = newNodesToBeVisited.removeFirst();
			
			oldNodesToBeVisited.addAll(currOldNode.getChildren());
			newNodesToBeVisited.addAll(currNewNode.getChildren());

			currNewNode.setReferenceToADocument(newADoc);
			
			if(!currOldNode.getOriginalIdWithHostAddressContainersThatPointToThisNode().isEmpty())
			{
				Iterator<SimpleNode> addressesThatPointToThisNodeIter = currOldNode.getOriginalIdWithHostAddressContainersThatPointToThisNode().iterator();
				while(addressesThatPointToThisNodeIter.hasNext())
				{
					SimpleNode addressThatPointsToThisNode = addressesThatPointToThisNodeIter.next();
					currNewNode.getOriginalIdWithHostAddressContainersThatPointToThisNode().add((SimpleNode)ADocument.returnRespectiveReferenceToNewADoc(originalADoc, newADoc, addressThatPointsToThisNode));
				}
			}
		}
		
		newADoc.rulesApplied.addAll(originalADoc.rulesApplied);
		
		return newADoc;
	}
	
	private static ADNode copyADNodes(ADNode givenADNode)
	{
		ADNode aDNodeToBeReturned = null;
		
		if(givenADNode instanceof Service)
			aDNodeToBeReturned = new Service((Service)givenADNode);
		else if(givenADNode instanceof SimpleNode)
			aDNodeToBeReturned = new SimpleNode((SimpleNode)givenADNode);
		
		return aDNodeToBeReturned;
	}
	
	public static ADNode returnRespectiveReferenceToNewADoc(ADocument oldADoc, ADocument newADoc, ADNode innerRefToOldADoc)
	{
		LinkedList<ADNode> oldNodesToBeVisited = new LinkedList<ADNode>();
		oldNodesToBeVisited.add(oldADoc.rootOfADoc);
		
		LinkedList<ADNode> newNodesToBeVisited = new LinkedList<ADNode>();
		newNodesToBeVisited.add(newADoc.rootOfADoc);
		
		while(!oldNodesToBeVisited.isEmpty())
		{
			ADNode currOldNode = oldNodesToBeVisited.removeFirst();
			oldNodesToBeVisited.addAll(currOldNode.getChildren());
			
			ADNode currNewNode = newNodesToBeVisited.removeFirst();
			newNodesToBeVisited.addAll(currNewNode.getChildren());
			
			while(oldNodesToBeVisited.size() > newNodesToBeVisited.size())
				oldNodesToBeVisited.removeLast();
			
			while(oldNodesToBeVisited.size() < newNodesToBeVisited.size())
				newNodesToBeVisited.removeLast();
			
			if(oldNodesToBeVisited.size() != newNodesToBeVisited.size())
				System.err.println("Serious problem at returnRespectiveReferenceToNewADoc!!!");
			
			if(innerRefToOldADoc == currOldNode)
			{
			
//				if(!currOldNode.equals(currNewNode))
//				{
//					System.out.println("returnRespectiveReferenceToNewADoc");
//					System.out.println("currOldNode :" + currOldNode);
//					System.out.println("currNewNode :" + currNewNode);
//					
//					System.out.println("oldADoc");
//					System.out.println(oldADoc);
//					
//					System.out.println("newADoc");
//					System.out.println(newADoc);
//				}
//				else
					return currNewNode;
			}
		}
		
		return null;
	}
	
	// TODO this has to be done more properly
	public static StringBuffer returnCallingPeerOfSpecificWebService(Service givenServiceCall, ADocument givenADoc)
	{
		if(givenServiceCall.getServiceName().toString().equals("send") || givenServiceCall.getServiceName().toString().equals("receive"))
			return new StringBuffer(givenServiceCall.getExecutionPeer().toString());
		
		LinkedList<ADNode> path = new LinkedList<ADNode>();
		ADNode currentNode = givenServiceCall;
		while(currentNode != null)
		{
			path.addFirst(currentNode);
			currentNode = currentNode.getFather();
		}
		
//		System.out.println("<-----------------------");
//		System.out.println("Checking for service:" + givenServiceCall);
		StringBuffer callingPeer = new StringBuffer(givenADoc.getPeerThatHostsADoc());
		while(!path.isEmpty())
		{
			currentNode = path.removeFirst();
//			System.out.println(currentNode);
			if(currentNode instanceof Service)
			{
				Service serviceCurrentNode = (Service) currentNode;
				String serviceNameOfCurrentNode = serviceCurrentNode.getServiceName().toString();
				
				if(serviceNameOfCurrentNode.equals("newNode") && serviceCurrentNode != givenServiceCall)
					callingPeer = new StringBuffer(serviceCurrentNode.getExecutionPeer().toString());

				else if(serviceNameOfCurrentNode.equals("receive"))
					return null;
				
				// if serviceCurrentNode is any other service
				else
				{
					int afterActivatedServicePosition = -1;
					if(serviceCurrentNode.getAfterActivated().size() == 0)
						afterActivatedServicePosition = checkIfThereIsADNodeWithSpecificIdAtTheADNodeList(serviceCurrentNode.getAfterActivated(), path);

					int afterTerminatedServicePosition = -1;
					if(serviceCurrentNode.getAfterTerminated().size() == 0)
						afterTerminatedServicePosition = checkIfThereIsADNodeWithSpecificIdAtTheADNodeList(serviceCurrentNode.getAfterTerminated(), path);
					
					if(afterActivatedServicePosition != -1 || afterTerminatedServicePosition != -1)
					{
						int elementsToRemove = 0;
						if(afterActivatedServicePosition > afterTerminatedServicePosition)
							elementsToRemove = afterActivatedServicePosition;
						else if(afterActivatedServicePosition < afterTerminatedServicePosition)
							elementsToRemove = afterTerminatedServicePosition;
						// if they are equal
						else
							if(afterActivatedServicePosition != -1)
								elementsToRemove = afterActivatedServicePosition;
						
						for(int i=0; i<elementsToRemove; i++)
							path.remove();
					}
				}
			}
		}
//		System.out.println("----------------------->");
		return callingPeer;
	}
	
	// The following method checks if there is a node in givenADNodeList with ID in the givenIdSet. If such node
	// exists then its position in the list is being returned (0 means that is the first one). -1 is
	// returned if no such node is has been found. If more than one ids appear in the givenIdSet then
	// the last node of givenADNodeList that has id in the givenIdSet will be returned
	private static int checkIfThereIsADNodeWithSpecificIdAtTheADNodeList(Set<String> givenIdSet, List<ADNode> givenADNodeList)
	{
		Set<String> copyOfGivenIdSet = new HashSet<String>(givenIdSet);
		LinkedList<ADNode> copyOfGivenADNodeList = new LinkedList<ADNode>(givenADNodeList);
		int position = 0;
		int result = -1;
		while(!copyOfGivenADNodeList.isEmpty() && copyOfGivenIdSet.size() != 0)
		{
			ADNode currADNode = copyOfGivenADNodeList.removeFirst();
			if(givenIdSet.contains(currADNode.getCurrNodeID().toString()))
			{
				copyOfGivenIdSet.remove(currADNode.getCurrNodeID().toString());
				result = position;
			}
			else
				position++;
		}
		
		return result;
	}
	
	// This function finds all services of a specific ADocument and
	// returns a list of references to them
	public static List<Service> findServicesOfADoc(ADocument givenADoc)
	{
		List<Service> returningList = new ArrayList<Service>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenADoc.rootOfADoc);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			
			if(currNode instanceof Service)
				returningList.add((Service) currNode);
		}
		return returningList;
	}
	
	public String toString()
	{
		StringBuffer returnStr = new StringBuffer();
		returnStr.append("Doc@");
		returnStr.append(peerThatHostsADoc);
		returnStr.append("/");
		returnStr.append(rootOfADoc.toString());
		return returnStr.toString();
	}
	
	public boolean equals(Object givenObject)
	{
		if(givenObject instanceof ADocument)
		{
			ADocument givenADocument = (ADocument)givenObject;
			
//			if(givenADocument.getADocNumber() == 22 && this.getADocNumber() == 24)
//			{
//				VerboseADocumentDrawingModule.createOutputImage(givenADocument, Integer.toString(givenADocument.getADocNumber()));
//				VerboseADocumentDrawingModule.createOutputImage(this, Integer.toString(this.getADocNumber()));
//				System.out.println("xaxa");
//			}
			
			if(givenADocument.peerThatHostsADoc.toString().equals(peerThatHostsADoc.toString()))
			{
				if(givenADocument.rootOfADoc.equals(rootOfADoc))
					return true;
			}
		}
		return false;
	}
	
	
	public int hashCode()
	{
		//TODO this has to be changed
//		return 1;
		calculateHashCode();
		return hashCodeValue;
	}
	
	public void calculateHashCode()
	{
		long newHashCode = Math.abs(peerThatHostsADoc.toString().hashCode());
		newHashCode += Math.abs(rootOfADoc.hashCode());
		newHashCode = newHashCode % Integer.MAX_VALUE;
		
		hashCodeValue = (int)newHashCode;
	}
	
	// In order for a document to be consistent, it must have all 
	// his references correct. More specifically all the children 
	// references of a node must be pointing to nodes that would 
	// have the specific node as a father 
	public boolean isADocumentConsistent()
	{
		if(this.rootOfADoc.getFather() != null)
		{
			System.err.println("Inconsistency at the root of the ADoc");
			return false;
		}
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(this.rootOfADoc);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			for(int i=0; i<currNode.getChildren().size(); i++)
			{
				if(currNode.getChildren().get(i).getFather() != currNode)
				{
					System.err.println("Inconsistency at '" + currNode + "' , child '" + currNode.getChildren().get(i) + "'");
					return false;
				}
			}
		}
		
		// we check also if all the nodes are pointing to
		// *this*
		return checkIfAllNodesOfGivenDocPointToThisDoc();
	}
	
	public static boolean isSubTreeConsistent(ADNode givenADNode)
	{
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenADNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			for(int i=0; i<currNode.getChildren().size(); i++)
			{
				if(currNode != givenADNode && currNode.getChildren().get(i).getFather() != currNode)
				{
					System.err.println("Inconsistency at '" + currNode + "' , child '" + currNode.getChildren().get(i) + "'");
					return false;
				}
			}
		}
		return true;
	}
	
	private int calculateTheNumberOfNodesOfADoc(ADocument givenADoc)
	{
		int returningNumber = 0;
		LinkedList<ADNode> aDNodesToBeVisited = new LinkedList<ADNode>();
		aDNodesToBeVisited.add(givenADoc.getRootOfADoc());
		while(!aDNodesToBeVisited.isEmpty())
		{
			returningNumber++;
			aDNodesToBeVisited.addAll(aDNodesToBeVisited.removeFirst().getChildren());
		}
		return returningNumber;
	}
	
	private int calculateTheNumberOfNewNodeScOfADoc(ADocument givenADoc)
	{
		int returningNumber = 0;
		LinkedList<ADNode> aDNodesToBeVisited = new LinkedList<ADNode>();
		aDNodesToBeVisited.add(givenADoc.getRootOfADoc());
		while(!aDNodesToBeVisited.isEmpty())
		{
			ADNode currNode = aDNodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service servCurrNode = (Service) currNode;
				if(servCurrNode.getServiceName().toString().equals("newNode"))
					returningNumber++;
			}
			aDNodesToBeVisited.addAll(currNode.getChildren());
		}
		return returningNumber;
	}
	
	public int compareTo(Object otherADoc)
	{
		if( ! ( otherADoc instanceof ADocument ) )
			throw new ClassCastException("Invalid object");
		
		int nodesOfLocalDoc = calculateTheNumberOfNodesOfADoc(this);
		int nodesOfOtherDoc = calculateTheNumberOfNodesOfADoc((ADocument) otherADoc);
		
		if(nodesOfLocalDoc > nodesOfOtherDoc)
			return 1;
		else if(nodesOfLocalDoc < nodesOfOtherDoc)
			return -1;
		else
			return 0;
	}
	
	public void calculateNetworkCost()
	{
		simpleNetworkCost = 0;
		serviceAndWeightedNetworkCost1 = 0;
		
		PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
		
		// find all the service calls that are not under receive scs
		List<Service> scToBeExamined = UsefulMethods.returnServicesUnderneathASpecificNodeStopOnReceive(this.rootOfADoc);
		
		Iterator<Service> scToBeExaminedIter = scToBeExamined.iterator();
		while(scToBeExaminedIter.hasNext())
		{
			Service currSc = scToBeExaminedIter.next();
			if(currSc.getServiceName().toString().equals("send"))
			{
				// for every send we will locate the respective receive.
				// If these two scs have different calling peers then we
				// increase the counters (note that the calling peer and
				// the execution peer for send - receive is the same peer)
				SimpleNode secondChildOfSend = (SimpleNode)currSc.getChildren().get(1);
				IdContainer origIdOfRespectiveReceive = secondChildOfSend.getRemoteOriginalIdAndHostingPeer().id;
				Service receiveRef = (Service) this.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(origIdOfRespectiveReceive);
				
				// we compare the execution peers of the send and the receive
				if(!currSc.getExecutionPeer().toString().equals(receiveRef.getExecutionPeer().toString()))
				{
					simpleNetworkCost++;
					Float realCost = peerInformationModuleInstance.getComCost(currSc.getExecutionPeer().toString(), receiveRef.getExecutionPeer().toString());
					if(realCost != null)
						serviceAndWeightedNetworkCost1 += realCost;	
				}
			}
			// we ignore the receives
			else if(currSc.getServiceName().toString().equals("receive"));
			// if it is a newNode we add something small
			else if(currSc.getServiceName().toString().equals("newNode"))
			{
				
				StringBuffer callingPeerOfCurrSc = returnCallingPeerOfSpecificWebService(currSc, this);
				if(!callingPeerOfCurrSc.toString().equals(currSc.getExecutionPeer().toString()))
				{
					simpleNetworkCost += 0.2;
					
					Float tempCost = peerInformationModuleInstance.getComCost(callingPeerOfCurrSc.toString(), currSc.getExecutionPeer().toString());
					if(tempCost != null)
						serviceAndWeightedNetworkCost1 += 0.2 * tempCost;
				}
			}
			// and for all the other simple scs and newNode scs we check if
			// the calling peer is different from the executing peer. If yes
			// the we increase the counters
			// For the serviceAndWeightedNetworkCost1 we consider also some
			// computation cost for the simple scs
			else
			{
				StringBuffer callingPeerOfCurrSc = returnCallingPeerOfSpecificWebService(currSc, this);
				if(!callingPeerOfCurrSc.toString().equals(currSc.getExecutionPeer().toString()))
				{
					simpleNetworkCost++;
					
					Float tempCost = peerInformationModuleInstance.getComCost(callingPeerOfCurrSc.toString(), currSc.getExecutionPeer().toString());
					if(tempCost != null)
						serviceAndWeightedNetworkCost1 += tempCost;
					
					// if the sc has other scs underneath then we must also consider
					// the transfering of the variables to the service of currSc
					if(currSc.checkInDescendantsForReceiveOrSimpleSc())
					{
						simpleNetworkCost++;
						if(tempCost != null)
							serviceAndWeightedNetworkCost1 += tempCost;
					}
				}
				// some computation cost for the simple scs
				serviceAndWeightedNetworkCost1 += 0.5;
			}
		}
	}
	
//	// The method bellow should run when the document is given to us clean from any
//	// newNode, send and receive services
//	// NOTE The method bellow has to be rewritten if send and receives exist
//	// in the initial document
//	public void calculateNetworkCostInitializationOLD(ADocument givenADoc)
//	{
//		simpleNetworkCost = 0;
//		realNetworkCost = 0;
//		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
//		nodesToBeVisited.add(givenADoc.getRootOfADoc());
//		while(!nodesToBeVisited.isEmpty())
//		{
//			ADNode currNode = nodesToBeVisited.removeFirst();
//			nodesToBeVisited.addAll(currNode.getChildren());
//			
//			if(currNode instanceof Service)
//			{
//				Service serviceCurrNode = (Service) currNode;
//				String callingPeerOfServiceCurrNode = ADocument.returnCallingPeerOfSpecificWebService(serviceCurrNode, givenADoc).toString();
//				if(!callingPeerOfServiceCurrNode.equals(serviceCurrNode.getExecutionPeer().toString()))
//				{
//					simpleNetworkCost++;
//					PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
//					Integer realCost = peerInformationModuleInstance.getComCost(callingPeerOfServiceCurrNode, serviceCurrNode.getExecutionPeer().toString());
//					if(realCost != null)
//						realNetworkCost += realCost;
//					else
//					{
//						System.out.println("Not all network costs between peers were defined");
//						System.exit(0);
//					}
//				}
//			}
//		}
//	}
//	
//	// The following method should be called every time we apply a GeneralizationInstatiation
//	// step at an ADocument
//	// The arguments callingPeer, oldExecutionPeer, newExecutionPeer are related to the service
//	// call to which we applied the Generalization and Instatiation rule
//	public void updateNetworkCostGeneralizationInstatiationOLD(String callingPeer, String oldExecutionPeer, String newExecutionPeer)
//	{
//		PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
//		if(!callingPeer.equals(oldExecutionPeer))
//		{
//			simpleNetworkCost--;
//			Integer realCost = peerInformationModuleInstance.getComCost(callingPeer, oldExecutionPeer);
//			if(realCost != null)
//				realNetworkCost -= realCost;
//			else
//			{
//				System.out.println("Not all network costs between peers were defined");
//				System.exit(0);
//			}
//		}
//		if(!callingPeer.equals(newExecutionPeer))
//		{
//			simpleNetworkCost++;
//			Integer realCost = peerInformationModuleInstance.getComCost(callingPeer, newExecutionPeer);
//			if(realCost != null)
//				realNetworkCost += realCost;
//			else
//			{
//				System.out.println("Not all network costs between peers were defined");
//				System.exit(0);
//			}
//		}
//	}
//	
//	// The following method should be called every time we apply a Delegation
//	// step at an ADocument
//	// The arguments oldCallingPeer, newCallingPeer, executionPeer are related to the service
//	// call to which we applied the Delegation rule
//	public void updateNetworkCostDelegationOLD(String oldCallingPeer, String newCallingPeer, Service delegatedService)
//	{
//		// delegation works in a way that newNode and send service calls have
//		// different calling peers that the delegated service's calling peer
//		simpleNetworkCost = simpleNetworkCost + 2;
//		PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
//		Integer realCost = peerInformationModuleInstance.getComCost(oldCallingPeer, newCallingPeer);
//		realCost = realCost * 2;
//		if(realCost != null)
//			realNetworkCost += realCost;
//		else
//		{
//			System.out.println("Not all network costs between peers were defined");
//			System.exit(0);
//		}
//		
//		// The delegated service and all its descendants will be visited in order to see
//		// if the delegation affected the already calculated costs.
//		LinkedList<ADNode> aDNodesToBeVisited = new LinkedList<ADNode>();
//		aDNodesToBeVisited.add(delegatedService);
//		while(!aDNodesToBeVisited.isEmpty())
//		{
//			ADNode currADNode = aDNodesToBeVisited.removeFirst();
//			aDNodesToBeVisited.addAll(currADNode.getChildren());
//			
//			if(!(currADNode instanceof Service))
//				continue;
//			
//			Service currServiceNode = (Service) currADNode;
//			// TODO we have to write it in a better way
//			if(currServiceNode.getServiceName().toString().equals("receive"))
//			{
//				System.out.println("A receive was found among the descendants");
//				//////////////////////////////////////
////				System.exit(0);
//			}
//			
//			if(!oldCallingPeer.equals(currServiceNode.getExecutionPeer().toString()))
//			{
//				simpleNetworkCost--;
//				Integer realCost2 = peerInformationModuleInstance.getComCost(oldCallingPeer, currServiceNode.getExecutionPeer().toString());
//				if(realCost2 != null)
//					realNetworkCost -= realCost2;
//				else
//				{
//					System.out.println("Not all network costs between peers were defined");
//					System.exit(0);
//				}
//			}
//			if(!newCallingPeer.equals(currServiceNode.getExecutionPeer().toString()))
//			{
//				simpleNetworkCost++;
//				Integer realCost2 = peerInformationModuleInstance.getComCost(newCallingPeer, currServiceNode.getExecutionPeer().toString());
//				if(realCost2 != null)
//					realNetworkCost += realCost2;
//				else
//				{
//					System.out.println("Not all network costs between peers were defined");
//					System.exit(0);
//				}
//			}
//		}
//
//	}
//	
//	// The following method should be called every time we apply a Common Subexpression
//	// Elimination step at an ADocument
//	// The list eliminatedService contains the eliminated services with their calling peers.
//	// The masterServiceCallingPeer contains the calling peer of the master service. This is
//	// the service that it going to provide its data to the others
//	public void updateNetworkCostCSEOLD(List<ServiceAndCallingPeer> eliminatedServices, String masterServiceCallingPeer)
//	{
//		Iterator<ServiceAndCallingPeer> eliminatedServicesIter = eliminatedServices.iterator();
//		while(eliminatedServicesIter.hasNext())
//		{
//			ServiceAndCallingPeer currServiceAndCallingPeer = eliminatedServicesIter.next();
//			
//			PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
//			
//			// Here we calculate the savings that occur from the ellimination of not needed services
//			LinkedList<ADNode> aDNodesToBeVisited = new LinkedList<ADNode>();
//			aDNodesToBeVisited.add(currServiceAndCallingPeer.service);
//			while(!aDNodesToBeVisited.isEmpty())
//			{
//				ADNode currADNode = aDNodesToBeVisited.removeFirst();
//				aDNodesToBeVisited.addAll(currADNode.getChildren());
//				
//				if(!(currADNode instanceof Service))
//					continue;
//				
//				Service currServiceNode = (Service) currADNode;
//				//TODO we have to write this in a better way				
//				if(currServiceNode.getServiceName().toString().equals("receive"))
//				{
//					System.out.println("A receive was found among the descendants");
//////////////////////////////////////////////////////////////////////	
//					//					System.exit(0);
//				}
//				
//				if(!currServiceAndCallingPeer.callingPeer.equals(currServiceNode.getExecutionPeer().toString()))
//				{
//					simpleNetworkCost--;
//					Integer realCost = peerInformationModuleInstance.getComCost(currServiceAndCallingPeer.callingPeer, currServiceNode.getExecutionPeer().toString());
//					if(realCost != null)
//						realNetworkCost -= realCost;
//					else
//					{
//						System.out.println("Not all network costs between peers were defined");
//						System.exit(0);
//					}
//				}
//			}
//			
//			// For the send & receive communication cost
//			if(!currServiceAndCallingPeer.callingPeer.equals(masterServiceCallingPeer))
//			{
//				simpleNetworkCost++;
//				Integer realCost = peerInformationModuleInstance.getComCost(currServiceAndCallingPeer.callingPeer, masterServiceCallingPeer);
//				if(realCost != null)
//					realNetworkCost += realCost;
//				else
//				{
//					System.out.println("Not all network costs between peers were defined");
//					System.exit(0);
//				}
//
//			}
//		}
//	}
	
	// Let's say that givenNode is a root at a tree A. The following method
	// searches for original ids that are in the receiving destination of the
	// receive service calls nodes of A.
	// These are nodes (send sc and the newNode sc that are usually above sends)
	// that shouldn't be considered in equivalence check
	public static Set<IdContainer> findOriginalIdsThatShouldBeOmmited(ADNode givenNode)
	{
		Set<IdContainer> setToBeReturned = new HashSet<IdContainer>();
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currADNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currADNode.getChildren());
			
			if(currADNode instanceof Service)
			{
				Service serviceCurrADNode = (Service) currADNode;
				if(serviceCurrADNode.getServiceName().toString().equals("receive"))
				{
					// I select the last child which should contain the original id 
					// of the related send
					ADNode addressContainer = serviceCurrADNode.getChildren().get(serviceCurrADNode.getChildren().size() - 1);
					if(!(addressContainer instanceof SimpleNode))
					{
						System.err.println("A SimpleNode should be at the last child of a receive sc");
						System.exit(0);
					}
					SimpleNode addressContainerSimpleNode = (SimpleNode) addressContainer;
					setToBeReturned.add(addressContainerSimpleNode.getRemoteOriginalIdAndHostingPeer().id);
				}
			}
		}
		return setToBeReturned;
	}
	
	public static List<ADNode> findChildrenOfADNodeForEquivalenceCheck(ADNode givenADNode, Set<IdContainer> originalIDsOfNodesThatShouldNotExistInResult)
	{
		// Here we are going to find all the children of givenADNode.
		// Because we are searching for equivalence and not
		// equality the description of the receiving data of the 
		// receive service calls will be taken in to account. The 
		// receive sc as a sc will not be taken in to account in the
		// comparison
		LinkedList<ADNode> givenADNodeChildren = new LinkedList<ADNode>();
		Iterator<ADNode> givenADNodeGetChildrenIter = givenADNode.getChildren().iterator();
		while(givenADNodeGetChildrenIter.hasNext())
		{
			ADNode currChild = givenADNodeGetChildrenIter.next();
			if(currChild instanceof Service)
			{
				Service serviceCurrChild = (Service) currChild;
				if(serviceCurrChild.getServiceName().toString().equals("receive"))
				{
					givenADNodeChildren.addAll(serviceCurrChild.getChildren());
					// We remove the last child because it is the address of the send
					// that is transmitting data
					givenADNodeChildren.removeLast(); 
				}
				else
					givenADNodeChildren.add(currChild);
			}
			else
				givenADNodeChildren.add(currChild);
		}
		Iterator<ADNode> givenADNodeChildrenIter = givenADNodeChildren.iterator();
		while(givenADNodeChildrenIter.hasNext())
		{
			ADNode currChild = givenADNodeChildrenIter.next();
			if(currChild instanceof Service)
			{
				Service serviceCurrChild = (Service)currChild;
				IdContainer currChildIdContainer = null;
				// If it is a newNode sc then we check if the first child is in the set
				// We do it because newNode usually is above send (delegation)
				if(serviceCurrChild.getServiceName().toString().equals("newNode"))
				{
					currChild = currChild.getChildren().get(0);
					currChildIdContainer = new IdContainer(currChild.getOrigNodeID().toString(), currChild.getOrigDocID().toString(), currChild.getOrigPeerID().toString());
				}
				else
					currChildIdContainer = new IdContainer(currChild.getOrigNodeID().toString(), currChild.getOrigDocID().toString(), currChild.getOrigPeerID().toString());
				
				if(originalIDsOfNodesThatShouldNotExistInResult.contains(currChildIdContainer))
					givenADNodeChildrenIter.remove();
			}
		}
		
		return givenADNodeChildren;
	}
	
	public void increaseNumberOfRewritings()
	{
		numberOfRewritings++;
	}

	public void increaseNumberOfGeneralizationInstatiations()
	{
		numberOfGeneralizationInstatiations++;
	}

	public void increaseNumberOfDelegations()
	{
		numberOfDelegations++;
	}

	public void increaseNumberOfCSE()
	{
		numberOfCSE++;
	}
	
	public void insertAppliedRuleToADoc(String nameOfRule, int numberOfOldADoc, String arg1, String arg2, int numberOfNewADoc, boolean incrCost, boolean decrCost, float oldCost, float newCost)
	{
		RuleDescriptionContainer newEntry = new RuleDescriptionContainer();
		newEntry.nameOfRule = nameOfRule;
		newEntry.numberOfOldADoc = numberOfOldADoc;
		newEntry.arg1 = arg1;
		newEntry.arg2 = arg2;
		newEntry.numberOfNewADoc = numberOfNewADoc;
		newEntry.increasesCost = incrCost;
		newEntry.decreasesCost = decrCost;
		newEntry.oldCost = oldCost;
		newEntry.oldCost = newCost;
		rulesApplied.add(newEntry);
	}
	
	public String getRulesAppliedToString()
	{
		return rulesApplied.toString();
	}
	
	public List<RuleDescriptionContainer> getRulesApplied()
	{
		return rulesApplied;
	}
	
	// This method gets a subtree rooted by givenNode and it is going to
	// eliminate all transfers (send & reseives). It will also erase 
	// the execution peer of all service calls because it is not needed in
	// equivalence check. As a result it will return a document with root
	// the transfers free subtree
	static public ADNode constructEquivalenceDataTree(ADNode givenNode)
	{
		ADNode returningNode = null;
		if(givenNode instanceof Service)
		{
			Service serviceGivenNode = (Service) givenNode;
			if(serviceGivenNode.getServiceName().toString().equals("receive"))
			{
				// we will find the respective send of serviceGivenNode
				
				// we find send's address (original Id with hosting peer)
				SimpleNode sendsAddressContainer = (SimpleNode) serviceGivenNode.getChildren().get(serviceGivenNode.getChildren().size() - 1);
				IdWithHostContainer sendsAddress = sendsAddressContainer.getRemoteOriginalIdAndHostingPeer();
				
				// we locate the send
				Service referenceToSend = (Service) serviceGivenNode.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(sendsAddress.id);
				
				// we check if the what part of the send is a pointer
				ADNode firstChildOfSend = referenceToSend.getChildren().get(0);
				ADNode newFirstChildOfSend = firstChildOfSend;
				if(firstChildOfSend instanceof SimpleNode)
				{
					SimpleNode simpleNodeFirstChildOfSend = (SimpleNode) firstChildOfSend;
					// the send's what part is pointing to another node
					if(simpleNodeFirstChildOfSend.getNodeData().length() == 0 && !simpleNodeFirstChildOfSend.getRemoteOriginalIdAndHostingPeer().isEmpty())
					{
						// we locate the other service and we store its reference at newFirstChildOfSend
						IdWithHostContainer pointedServiceAddress = simpleNodeFirstChildOfSend.getRemoteOriginalIdAndHostingPeer();
						newFirstChildOfSend = simpleNodeFirstChildOfSend.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(pointedServiceAddress.id);
					}
				}
				
				returningNode = constructEquivalenceDataTree(newFirstChildOfSend);
			}
			else if(serviceGivenNode.getServiceName().toString().equals("send"));
			else if(serviceGivenNode.getServiceName().toString().equals("newNode"));

			// if it is a simple service
			else
			{
				returningNode = givenNode.returnNewNodeWithNoChildrenAndFatherButWithOldCurrIds();
				// we eliminate the execution peer of the returned simple service calls
				// we do this because in equivalence check we do not care where the
				// sc will be executed. We just care about the result
				((Service) returningNode).setExecutionPeer(new StringBuffer());

				for(int i=0; i<givenNode.getChildren().size(); i++)
				{
					ADNode newChildOfReturningNode = constructEquivalenceDataTree(givenNode.getChildren().get(i));
					returningNode.insertAsChild(newChildOfReturningNode);
				}
			}
		}
		// if it is a simpleNode
		else
		{
			returningNode = givenNode.returnNewNodeWithNoChildrenAndFatherButWithOldCurrIds();
			for(int i=0; i<givenNode.getChildren().size(); i++)
			{
				if(givenNode.getCurrNodeID().toString().equals("40") && i == 1 && givenNode.getReferenceToADocument().getADocNumber() == 10)
					System.out.println("xaxa");
				ADNode newChildOfReturningNode = constructEquivalenceDataTree(givenNode.getChildren().get(i));
				if(newChildOfReturningNode != null)
					returningNode.insertAsChild(newChildOfReturningNode);
			}
		}
		return returningNode;
	}
	
	public boolean checkIfAllNodesOfGivenDocPointToThisDoc()
	{
		boolean noProblemFound = true;
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(this.getRootOfADoc());
		
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			
			if(currNode.getReferenceToADocument() != this)
			{
				System.err.println("Node is not pointing to the correct document");
				System.err.println("node currId:" + currNode.getCurrNodeID());
				noProblemFound = false;
			}
		}
		
		return noProblemFound;
	}
	/*
	public int estimateMinCostOfADocNoSendsAndReceives(ADocument givenADoc)
	{
		int returningCost = 0;

		// First we are going to find the leaves of the document
		LinkedList<ADNode> leavesOfTheADoc = new LinkedList<ADNode>();
		
		// Path already visited
		
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenADoc.rootOfADoc);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currADNode = nodesToBeVisited.removeFirst();
			if(currADNode instanceof Service)
			{
				Service serviceCurrADNode = (Service) currADNode;
				if(serviceCurrADNode.getIsSpecialService())
					return -1;
			}
			
			if(currADNode.getChildren().isEmpty())
				leavesOfTheADoc.add(currADNode);
		}
		
		// Now we are going to ascend the tree from the leaves
		// and we are counting how many times we change execution
		// peer for services
		while(!leavesOfTheADoc.isEmpty())
		{
			ADNode currNode = leavesOfTheADoc.remove();
			int counter = 0;
			String lastExecutionPeer = null;
			while(currNode != null)
			{
				if(currNode instanceof Service)
				{
					Service serviceCurrNode = (Service) currNode;
					if(lastExecutionPeer == null)
					{
						lastExecutionPeer = serviceCurrNode.getExecutionPeer().toString();
					}
					else if(!serviceCurrNode.getExecutionPeer().toString().equals(lastExecutionPeer))
					{
						counter++;
						lastExecutionPeer = serviceCurrNode.getExecutionPeer().toString();
					}
				}
			}
			
			if(lastExecutionPeer != null)
			{
				if(!lastExecutionPeer.equals(givenADoc.getPeerThatHostsADoc().toString()))
					counter++;
			}
		}
		
	}
	*/
}
