package fr.inria.gemo.axml.optimax.adocument;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import fr.inria.gemo.axml.optimax.helping.DinstinctNumberGenerator;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;
import fr.inria.gemo.axml.optimax.io.output.GraphvizModule;

public class Service implements ADNode 
{
	private ADocument hostingDocumentReference;
	
	private StringBuffer currNodeID;
	private StringBuffer currDocID;
	private StringBuffer currPeerID;
	
	private StringBuffer origNodeID;
	private StringBuffer origDocID;
	private StringBuffer origPeerID;
	
	private ADNode father;
	private ExtendedLinkedList<ADNode> children;
	
	private boolean isSpecialService;
	public StringBuffer serviceName;
	public StringBuffer executionPeer;
	private Set<String> afterActivated;
	private Set<String> afterTerminated;
	private StringBuffer returning;
	private StringBuffer xpath;
	private int hashCodeOfTreeRootedByCurrService;
	
	private ExtendedLinkedList<SimpleNode> originalIdWithHostAddressContainersThatPointToThisNode; 
	
	private PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
	
	public Service()
	{
		ADNodeCounter.increaseNumberOfNodes();
		DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
		currNodeID = dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer();
		currDocID = new StringBuffer(peerInformationModuleInstance.getCurrDocId());
		currPeerID = new StringBuffer(peerInformationModuleInstance.getCurrPeerId());
		
		origNodeID = new StringBuffer(currNodeID);
		origDocID = new StringBuffer(currDocID);
		origPeerID = new StringBuffer(currPeerID);
		
		father = null;
		children = new ExtendedLinkedList<ADNode>();
		
		isSpecialService = false;
		serviceName = new StringBuffer();
		executionPeer = new StringBuffer();
		afterActivated = new HashSet<String>();
		afterTerminated = new HashSet<String>();
		returning = new StringBuffer();
		xpath = new StringBuffer();
		hashCodeOfTreeRootedByCurrService = 0;
		
		originalIdWithHostAddressContainersThatPointToThisNode = new ExtendedLinkedList<SimpleNode>();
	}

	public Service(StringBuffer givenOrigNodeID, StringBuffer givenOrigDocID, StringBuffer givenOrigPeerID, boolean givenIsSpecialService, Set<String> givenAfterActivated, Set<String> givenAfterTerminated, StringBuffer givenServiceName, StringBuffer givenExecutionPeer, StringBuffer givenReturning, StringBuffer givenXPath, int givenHashCodeOfTreeRootedByCurrService, ADNode givenFather, ExtendedLinkedList<ADNode> givenChildren)
	{
		ADNodeCounter.increaseNumberOfNodes();
		DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
		currNodeID = dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer();
		currDocID = new StringBuffer(peerInformationModuleInstance.getCurrDocId());
		currPeerID = new StringBuffer(peerInformationModuleInstance.getCurrPeerId());
		
		if(givenOrigNodeID == null || givenOrigNodeID.length() == 0)
			origNodeID = new StringBuffer(currNodeID);
		else
			origNodeID = new StringBuffer(givenOrigNodeID);
		
		if(givenOrigDocID == null || givenOrigDocID.length() == 0)
			origDocID = new StringBuffer(currDocID);
		else
			origDocID = new StringBuffer(givenOrigDocID);
		
		if(givenOrigPeerID == null || givenOrigPeerID.length() == 0)
			origPeerID = new StringBuffer(currPeerID);
		else
			origPeerID = new StringBuffer(givenOrigPeerID);
		
		originalIdWithHostAddressContainersThatPointToThisNode = new ExtendedLinkedList<SimpleNode>();
		
		this.isSpecialService = givenIsSpecialService;
		this.afterActivated = new HashSet<String>(givenAfterActivated);
		this.afterTerminated = new HashSet<String>(givenAfterTerminated);
		this.serviceName = new StringBuffer(givenServiceName);
		this.executionPeer = new StringBuffer(givenExecutionPeer);
		this.returning = new StringBuffer(givenReturning);
		this.xpath = new StringBuffer(givenXPath);
		this.hashCodeOfTreeRootedByCurrService = givenHashCodeOfTreeRootedByCurrService;
		this.father = givenFather;
		this.children = new ExtendedLinkedList<ADNode>();
		Iterator myIterator = givenChildren.iterator();
		while(myIterator.hasNext())
		{
			ADNode currentGivenNode = (ADNode) myIterator.next();
			ADNode myNewNode;
			if(currentGivenNode instanceof SimpleNode)
				myNewNode = new SimpleNode((SimpleNode)currentGivenNode);
			
			else
				myNewNode = new Service((Service)currentGivenNode);
			
			myNewNode.setFather(this);
			this.children.add(myNewNode);
		}
	}
	
	public Service(StringBuffer givenCurrNodeId, StringBuffer givenCurrDocId, StringBuffer givenCurrPeerId, StringBuffer givenOrigNodeID, StringBuffer givenOrigDocID, StringBuffer givenOrigPeerID, boolean givenIsSpecialService, Set<String> givenAfterActivated, Set<String> givenAfterTerminated, StringBuffer givenServiceName, StringBuffer givenExecutionPeer, StringBuffer givenReturning, StringBuffer givenXPath, int givenHashCodeOfTreeRootedByCurrService, ADNode givenFather, ExtendedLinkedList<ADNode> givenChildren)
	{
		this(givenOrigNodeID, givenOrigDocID, givenOrigPeerID, givenIsSpecialService, givenAfterActivated, givenAfterTerminated, givenServiceName, givenExecutionPeer, givenReturning, givenXPath, givenHashCodeOfTreeRootedByCurrService, givenFather, givenChildren);
		this.currNodeID = new StringBuffer(givenCurrNodeId);
		this.currDocID = new StringBuffer(givenCurrDocId);
		this.currPeerID = new StringBuffer(givenCurrPeerId);
	}
	
	public Service(Service givenService)
	{
		this(givenService.currNodeID, givenService.currDocID, givenService.currPeerID, givenService.origNodeID, givenService.origDocID, givenService.origPeerID, givenService.isSpecialService, givenService.afterActivated, givenService.afterTerminated,  givenService.serviceName, givenService.executionPeer, givenService.returning, givenService.xpath, givenService.hashCodeOfTreeRootedByCurrService, givenService.father, givenService.children);
	}
	
	public Service(Service givenService, boolean withNewIds, boolean markAllAsSpecialServices)
	{
		this(givenService);
		
		if(withNewIds || markAllAsSpecialServices)
		{
			DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
			
			LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
			nodesToBeVisited.add(this);
			while(!nodesToBeVisited.isEmpty())
			{
				ADNode currNode = nodesToBeVisited.removeFirst();
				if(withNewIds)
					currNode.setCurrNodeID(dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer());
				if(markAllAsSpecialServices && currNode instanceof Service)
				{
					Service currServiceNode = (Service) currNode;
					currServiceNode.setIsSpecialService(true);
				}
				nodesToBeVisited.addAll(currNode.getChildren());
			}
		}
	}
	
	public Service(Service givenService, boolean withNewIds)
	{
		this(givenService, withNewIds, false);
	}
	
	public void finalize()
	{
			ADNodeCounter.decreaseNumberOfNodes();
	}
	
	public String toString()
	{
		StringBuffer returnStr;

		returnStr = new StringBuffer(serviceName);
		returnStr.append("@");
		returnStr.append(executionPeer);
		returnStr.append("(");
		
		for(int i=0; i<children.size(); i++)
		{
			returnStr.append(children.get(i).toString());
			if(i != (children.size() -1))
				returnStr.append(", ");
		}
		
		returnStr.append(")");
		
		return returnStr.toString();
	}

	public String toStringWithId()
	{
		StringBuffer returnStr;
		
		returnStr = new StringBuffer("[" + this.hashCode() + "]{" + this.currNodeID + "." + this.currDocID + "." + this.currPeerID + "}" + serviceName);
		returnStr.append("@");
		returnStr.append(executionPeer);
		returnStr.append("(");
		
		for(int i=0; i<children.size(); i++)
		{
			returnStr.append(children.get(i).toStringWithId());
			if(i != (children.size() -1))
				returnStr.append(", ");
		}
		
		returnStr.append(")");
		
		return returnStr.toString();
	}
	
	public boolean equals(Object o)
	{
		if (!(o instanceof Service)) 
			return false;
		
		Service givenService = (Service)o;
		
//		if(this.getReferenceToADocument() != null && this.getReferenceToADocument().getADocNumber() == 30 && givenService.getReferenceToADocument().getADocNumber() == 44)
//		{
//			if(this.getCurrNodeID().toString().equals("1003") && givenService.getCurrNodeID().toString().equals("1516"))
//			{
//				GraphvizModule.createOutputImage(this.getReferenceToADocument(), "this1");
//				GraphvizModule.createOutputImage(givenService.getReferenceToADocument(), "givenService1");
//			}
//		}
		
		if(this.serviceName.toString().equals(givenService.serviceName.toString()))
		{
			if(this.executionPeer.toString().equals(givenService.executionPeer.toString()))
			{	
				if(this.hashCodeOfTreeRootedByCurrService == givenService.hashCodeOfTreeRootedByCurrService)
				{
					if(this.children.size() == givenService.children.size())
					{
						// here we are going to check if this or givenService are send or 
						// receives that have in their what part an original id pointer
						// to another service call (or node). If this is the case we follow the
						// pointer and we continue the equality check from that point.
						if(this.serviceName.toString().equals("send") || this.serviceName.toString().equals("receive"))
						{
							boolean thisFirstChildIsPointer = false;
							ADNode thisFirstChild = this.getChildren().get(0);
							if(thisFirstChild instanceof SimpleNode)
							{
								SimpleNode simpleThisFirstChild = (SimpleNode) thisFirstChild;
								if(simpleThisFirstChild.getNodeData().length() == 0 && !simpleThisFirstChild.getRemoteOriginalIdAndHostingPeer().isEmpty())
									thisFirstChildIsPointer = true;
							}
							
							boolean givenServiceFirstChildIsPointer = false;
							ADNode givenServiceFirstChild = givenService.getChildren().get(0);
							if(givenServiceFirstChild instanceof SimpleNode)
							{
								SimpleNode simpleGivenServiceFirstChild = (SimpleNode) givenServiceFirstChild;
								if(simpleGivenServiceFirstChild.getNodeData().length() == 0 && !simpleGivenServiceFirstChild.getRemoteOriginalIdAndHostingPeer().isEmpty())
									givenServiceFirstChildIsPointer = true;
							}
							
							// we are going to check for equality only the first child. The
							// second one is a pointer to a respective send - receive and we
							// do not check it.
							ADNode newThisFirstChild = null;
							ADNode newGivenServiceFirstChild = null;
							
							if(thisFirstChildIsPointer)
							{
								IdWithHostContainer actualPointer = ((SimpleNode) thisFirstChild).getRemoteOriginalIdAndHostingPeer();
								newThisFirstChild = this.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(actualPointer.id);
							}
							else
								newThisFirstChild = thisFirstChild;
							
							if(givenServiceFirstChildIsPointer)
							{
								IdWithHostContainer actualPointer = ((SimpleNode) givenServiceFirstChild).getRemoteOriginalIdAndHostingPeer();
								newGivenServiceFirstChild = givenService.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(actualPointer.id);
							}
							else
								newGivenServiceFirstChild = givenServiceFirstChild;
							
							// and now we check for equality the first children (at this point we do
							// not really care if these children are pointers or actual data
							return newThisFirstChild.equals(newGivenServiceFirstChild);
						}
						
						// if this and givenService are not send or receive
						// then we continue normally our computation (we do
						// not expect pointers at another place of the document
						Iterator<ADNode> thisIt = this.children.iterator();
						List<ADNode> givenServiceChildren = new ArrayList<ADNode>();
						givenServiceChildren.addAll(givenService.children);
						while(thisIt.hasNext())
						{
							ADNode thisCurrentNode = thisIt.next();
							
							Iterator<ADNode> givenServiceChildrenIter = givenServiceChildren.iterator();
							
							boolean found = false;
							while(givenServiceChildrenIter.hasNext())
							{
								ADNode givenServiceCurrentNode = (ADNode) givenServiceChildrenIter.next();
								
								if(thisCurrentNode.equals(givenServiceCurrentNode) == true)
								{
									found = true;
									givenServiceChildren.remove(givenServiceCurrentNode);
									break;
								}
							}
							if(found == false)
								return false;
						}
						return true;
					}
				}
			}
		}
		return false;
	}
	
	public boolean equivalent(Object o)
	{
		if (!(o instanceof ADNode)) 
			return false;
		
		// We create the equivalence data trees of both nodes. Note that these trees do
		// not have send & receives and the execution peer of the simple services is
		// deleted. These trees can be compared for equality and if they are equal, we
		// can safelly assume that this & o are equivalent
		ADNode thisEquivalenceDataTree = ADocument.constructEquivalenceDataTree(this);
		ADNode oEquivalenceDataTree = ADocument.constructEquivalenceDataTree((ADNode) o);
		
		if(!ADocument.isSubTreeConsistent(thisEquivalenceDataTree))
		{
			System.err.println("service equivalent: Inconsistency found");
			System.exit(0);
		}
		
		if(!ADocument.isSubTreeConsistent(oEquivalenceDataTree))
		{
			System.err.println("service equivalent: Inconsistency found");
			System.exit(0);
		}
		
		return oEquivalenceDataTree.equals(thisEquivalenceDataTree);
	}
	
	public int hashCode()
	{
		long returningHashCode = Math.abs(this.serviceName.toString().hashCode());
		returningHashCode += Math.abs(this.executionPeer.toString().hashCode());
		returningHashCode = returningHashCode % Integer.MAX_VALUE;
		returningHashCode += Math.abs(this.hashCodeOfTreeRootedByCurrService);
		returningHashCode = returningHashCode % Integer.MAX_VALUE;
		// For the receive service call we do not take into account
		// its children in the computation of the hash code (for more
		// info look at the explanation given in the equals method)
		if(this.getServiceName().toString().equals("receive"))
			return (int)returningHashCode;
		
		Iterator myIterator = this.children.iterator();
		while(myIterator.hasNext())
		{
			returningHashCode += Math.abs(myIterator.next().hashCode());
			returningHashCode = returningHashCode % Integer.MAX_VALUE;
		}
		
		return (int)returningHashCode;
//		return 0;
	}
	
	public int originalHashCode()
	{
		return super.hashCode();
	}
	
	public StringBuffer getCurrNodeID()
	{
		return currNodeID;
	}
	
	public StringBuffer getCurrDocID()
	{
		return currDocID;
	}

	public StringBuffer getCurrPeerID()
	{
		return currPeerID;
	}
	
	public StringBuffer getOrigNodeID()
	{
		return origNodeID;
	}
	
	public StringBuffer getOrigDocID()
	{
		return origDocID;
	}

	public StringBuffer getOrigPeerID()
	{
		return origPeerID;
	}
	
	public boolean getIsSpecialService()
	{
		return isSpecialService;
	}
	
	public Set<String> getAfterActivated()
	{
		return afterActivated;
	}
	
	public Set<String> getAfterTerminated()
	{
		return afterTerminated;
	}
	
	public StringBuffer getServiceName()
	{
		return serviceName;
	}
	
	public StringBuffer getExecutionPeer()
	{
		return executionPeer;
	}
	
	public StringBuffer getReturning()
	{
		return returning;
	}
	
	public StringBuffer getXPath()
	{
		return xpath;
	}
	
	public int getHashCodeOfTreeRootedByCurrService()
	{
		return hashCodeOfTreeRootedByCurrService;
	}

	public ADNode getFather()
	{
		return father;
	}
	
	public ExtendedLinkedList<ADNode> getChildren()
	{
		return children;
	}
	
	public StringBuffer getCopyOfServiceName()
	{
		return new StringBuffer(serviceName);
	}
	
	public StringBuffer getCopyOfExecutionPeer()
	{
		return new StringBuffer(executionPeer);
	}

	public List<ADNode> getCopyOfChildren()
	{
		return new ArrayList<ADNode>(children);
	}
	
	public void setIsSpecialService(boolean givenIsSpecialService)
	{
		this.isSpecialService = givenIsSpecialService;
	}

	public void setCurrNodeID(StringBuffer newCurrNodeID)
	{
		this.currNodeID = new StringBuffer(newCurrNodeID);
	}
	
	public void setCurrDocID(StringBuffer newCurrDocID)
	{
		this.currDocID = new StringBuffer(newCurrDocID);
	}
	
	public void setCurrPeerID(StringBuffer newCurrPeerID)
	{
		this.currPeerID = new StringBuffer(newCurrPeerID);
	}
	
	public void setAfterActivated(Set<String> givenAfterActivated)
	{
		this.afterActivated = new HashSet<String>(givenAfterActivated);
	}
	
	public void setAfterTerminated(Set<String> givenAfterTerminated)
	{
		this.afterTerminated = new HashSet<String>(givenAfterTerminated);
	}
	
	public void setServiceName(StringBuffer givenServiceName)
	{
		this.serviceName = new StringBuffer(givenServiceName);
	}
	
	public void setExecutionPeer(StringBuffer givenExecutionPeer)
	{
		this.executionPeer = new StringBuffer(givenExecutionPeer);
	}
	
	public void setXPath(StringBuffer givenXPath)
	{
		this.xpath = new StringBuffer(givenXPath);
	}
	
	public void setHashCodeOfTreeRootedByCurrService(int givenValue)
	{
		this.hashCodeOfTreeRootedByCurrService = givenValue;
	}

	public void setArguments(List<ADNode> givenArguments)
	{
		this.children.clear();
		
		Iterator<ADNode> givenArgumentsIterator = givenArguments.iterator();
		while(givenArgumentsIterator.hasNext())
		{
			ADNode tempArgumentObjectRef = givenArgumentsIterator.next();
			if(tempArgumentObjectRef instanceof SimpleNode)
				this.children.add(new SimpleNode((SimpleNode)tempArgumentObjectRef));
			else
				this.children.add(new Service((Service)tempArgumentObjectRef));
		}
	}
	
	public void setFather(ADNode givenFather)
	{
		this.father = givenFather;
	}
	
	public void setSpecificChildren(List<ADNode> givenArguments)
	{
		this.children.clear();
		this.children.addAll(givenArguments);
		for(int i=0;i<this.children.size(); i++)
		{
			this.children.get(i).setFather(this);
		}
	}
	
	public void setService(Service givenService)
	{
//		delegate = givenService;
		
		this.currNodeID = givenService.currNodeID;
		this.currDocID = givenService.currDocID;
		this.currPeerID = givenService.currPeerID;
		this.origNodeID = givenService.origNodeID;
		this.origDocID = givenService.origDocID;
		this.origPeerID = givenService.origPeerID;
		this.isSpecialService = givenService.getIsSpecialService();
		this.serviceName = givenService.getServiceName();
		this.executionPeer = givenService.getExecutionPeer();
		this.returning = givenService.getReturning();
		this.xpath = givenService.getXPath();
		this.hashCodeOfTreeRootedByCurrService = givenService.hashCodeOfTreeRootedByCurrService;
		this.father = givenService.getFather();
		this.children = givenService.getChildren();
		
		if(this.children != null)
		{
			for(int i=0; i<this.children.size(); i++)
				this.children.get(i).setFather(this);
		}
//		setServiceName(givenService.getServiceName());
//		setExecutionPeer(givenService.getExecutionPeer());
//		setArguments(givenService.getArguments());
	}
	
	public ExtendedLinkedList<SimpleNode> getOriginalIdWithHostAddressContainersThatPointToThisNode()
	{
		return originalIdWithHostAddressContainersThatPointToThisNode;
	}
	
	public ADocument getReferenceToADocument()
	{
		return hostingDocumentReference;
	}
	
	public void setReferenceToADocument(ADocument givenRefToADoc)
	{
		hostingDocumentReference = givenRefToADoc;
	}
	
	public ADNode returnNewNodeWithNoChildrenAndFatherButWithOldCurrIds()
	{
		Service newService = new Service(this.origNodeID, this.origDocID, this.origPeerID, this.currNodeID, this.currDocID, this.currPeerID, this.isSpecialService, this.afterActivated, this.afterTerminated, this.serviceName, this.executionPeer, this.returning, this.xpath, this.hashCodeOfTreeRootedByCurrService, null, new ExtendedLinkedList<ADNode>());
		return newService;
	}
	
	public ADNode returnNewNodeWithNoChildrenAndFatherButWithNewCurrIds()
	{
		Service newService = new Service(this.origNodeID, this.origDocID, this.origPeerID, this.isSpecialService, this.afterActivated, this.afterTerminated, this.serviceName, this.executionPeer, this.returning, this.xpath, this.hashCodeOfTreeRootedByCurrService, null, new ExtendedLinkedList<ADNode>());
		return newService;
	}
	
	public void removeNode()
	{
		if(this.getFather() != null)
			this.getFather().getChildren().refRemove(this);

		this.hostingDocumentReference = null;
		
		this.setFather(null);
	}
	
	public void insertAsChildInBeginning(ADNode givenSubTree)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		this.children.add(0, givenSubTree);
		givenSubTree.setFather(this);
	}
	
	public void insertAsChild(ADNode givenSubTree)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		this.children.add(givenSubTree);
		givenSubTree.setFather(this);
	}
	
	public void insertAsChildNextToGivenChild(ADNode givenSubTree, ADNode givenChild)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		for(int i=0; i<this.children.size(); i++)
		{
			if(this.children.get(i) == givenChild)
			{
				this.children.add(i+1, givenSubTree);
				givenSubTree.setFather(this);
			}
				
		}
	}
	
	public boolean checkInDescendantsForReceiveOrSimpleSc()
	{
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.addAll(this.children);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				if(!serviceCurrNode.getIsSpecialService())
					return true;
				else
				{
					if(serviceCurrNode.getServiceName().equals("receive"))
						return true;
				}
			}
		}
		return false;
	}
}