package fr.inria.gemo.axml.optimax.adocument;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import fr.inria.gemo.axml.optimax.helping.DinstinctNumberGenerator;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.helping.Utilities;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;

public class SimpleNode implements ADNode
{
	private ADocument hostingDocumentReference;
	
	private StringBuffer currNodeID;
	private StringBuffer currDocID;
	private StringBuffer currPeerID;
	
	private StringBuffer origNodeID;
	private StringBuffer origDocID;
	private StringBuffer origPeerID;
	
	private ADNode father;
	private ExtendedLinkedList<ADNode> children;
	
	private StringBuffer nodeData;
	
	public Document declaration;	// if this SimpleNode represents a declaration Node
									// then attach here a copy of this declaration
	
	public String paramName;		// if this SimpleNode represents a param node then
									// we add here the name of this param Node

	private IdWithHostContainer remoteOriginalIdAndHostingPeer;
									// here we are going to store the id of another
									// node. We do not store it at nodeData because
									// we want two simplenodes that contain different 
									// addresses to be equal
	private List<StringBuffer> xPathList;
	
	private ExtendedLinkedList<SimpleNode> originalIdWithHostAddressContainersThatPointToThisNode;
	
	private PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
	
	SimpleNode()
	{
		ADNodeCounter.increaseNumberOfNodes();
		DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
		currNodeID = dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer();
		currDocID = new StringBuffer(peerInformationModuleInstance.getCurrDocId());
		currPeerID = new StringBuffer(peerInformationModuleInstance.getCurrPeerId());
		
		origNodeID = new StringBuffer();
		origDocID = new StringBuffer();
		origPeerID = new StringBuffer();
		
		father = null;
		children = new ExtendedLinkedList<ADNode>();
		nodeData = new StringBuffer();
		remoteOriginalIdAndHostingPeer = new IdWithHostContainer();
		xPathList = new ArrayList<StringBuffer>();
		
		originalIdWithHostAddressContainersThatPointToThisNode = new ExtendedLinkedList<SimpleNode>();
	}
	
	public SimpleNode(StringBuffer givenOrigNodeID, StringBuffer givenOrigDocID, StringBuffer givenOrigPeerID, StringBuffer givenNodeData, IdWithHostContainer givenRemoteOriginalIdAndHostingPeer, List<StringBuffer> givenXPathList, ADNode givenFather, ExtendedLinkedList<ADNode> givenChildren, String givenParamName, Document givenDeclaration)
	{
		this();
		origNodeID = new StringBuffer(givenOrigNodeID);
		origDocID = new StringBuffer(givenOrigDocID);
		origPeerID = new StringBuffer(givenOrigPeerID);
		this.remoteOriginalIdAndHostingPeer = new IdWithHostContainer(givenRemoteOriginalIdAndHostingPeer);
		this.xPathList = new ArrayList<StringBuffer>(givenXPathList);
		this.nodeData.append(givenNodeData);
		if(givenParamName != null)
			this.paramName = new String(givenParamName);
		else 
			this.paramName = null;
		
		if(givenDeclaration != null)
		{
			DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
			DocumentBuilder loader;
			try {
				loader = dfactory.newDocumentBuilder();
				this.declaration = loader.newDocument();
				Node rootNodeForDeclarationDom = this.declaration.importNode(givenDeclaration.getDocumentElement(), true);
				this.declaration.appendChild(rootNodeForDeclarationDom);
			} catch (ParserConfigurationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		else
			this.declaration = null;
		
		this.father = givenFather;
		this.children = new ExtendedLinkedList<ADNode>();
		Iterator myIterator = givenChildren.iterator();
		while(myIterator.hasNext())
		{
			ADNode currentGivenNode = (ADNode) myIterator.next();
			ADNode myNewNode;
			if(currentGivenNode instanceof SimpleNode)
				myNewNode = new SimpleNode((SimpleNode)currentGivenNode);
			
			else
				myNewNode = new Service((Service)currentGivenNode);
			
			myNewNode.setFather(this);
			this.children.add(myNewNode);
		}
	}
	
	SimpleNode(StringBuffer givenOrigNodeID, StringBuffer givenOrigDocID, StringBuffer givenOrigPeerID, StringBuffer givencurrNodeID, StringBuffer givenCurrDocID, StringBuffer givenCurrPeerID, StringBuffer givenNodeData, IdWithHostContainer givenRemoteOriginalIdAndHostingPeer, List<StringBuffer> givenXPathList, ADNode givenFather, ExtendedLinkedList<ADNode> givenChildren, String givenParamName, Document givenDeclaration)
	{
		this(givenOrigNodeID, givenOrigDocID, givenOrigPeerID, givenNodeData, givenRemoteOriginalIdAndHostingPeer, givenXPathList, givenFather, givenChildren, givenParamName, givenDeclaration);
		this.currNodeID = new StringBuffer(givencurrNodeID);
		this.currDocID = new StringBuffer(givenCurrDocID);
		this.currPeerID = new StringBuffer(givenCurrPeerID);
	}
	
	SimpleNode(SimpleNode givenSimpleNode)
	{
		this(givenSimpleNode.origNodeID, givenSimpleNode.origDocID, givenSimpleNode.origPeerID, givenSimpleNode.currNodeID, givenSimpleNode.currDocID, givenSimpleNode.currPeerID, givenSimpleNode.nodeData, givenSimpleNode.remoteOriginalIdAndHostingPeer, givenSimpleNode.xPathList, givenSimpleNode.father, givenSimpleNode.children, givenSimpleNode.paramName, givenSimpleNode.declaration);
	}
	
	public SimpleNode(SimpleNode givenSimpleNode, boolean withNewIds)
	{
		this(givenSimpleNode);
		if(withNewIds)
		{
			DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
			this.currNodeID = dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer();
			
			LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
			nodesToBeVisited.addAll(this.children);
			while(!nodesToBeVisited.isEmpty())
			{
				ADNode currNode = nodesToBeVisited.removeFirst();
				currNode.setCurrNodeID(dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer());
				nodesToBeVisited.addAll(currNode.getChildren());
			}
		}
	}
	
	public SimpleNode(SimpleNode givenSimpleNode, boolean withNewIds, boolean markAllAsSpecialServices)
	{
		this(givenSimpleNode);
		
		if(withNewIds || markAllAsSpecialServices)
		{
			DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
			
			LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
			nodesToBeVisited.add(this);
			while(!nodesToBeVisited.isEmpty())
			{
				ADNode currNode = nodesToBeVisited.removeFirst();
				if(withNewIds)
					currNode.setCurrNodeID(dinstinctNumberGeneratorInstance.getANodeNumberStringBuffer());
				if(markAllAsSpecialServices && currNode instanceof Service)
				{
					Service currServiceNode = (Service) currNode;
					currServiceNode.setIsSpecialService(true);
				}
				nodesToBeVisited.addAll(currNode.getChildren());
			}
		}
	}
	
	SimpleNode(StringBuffer givenNodeData)
	{
		this();
		this.nodeData.append(givenNodeData);
	}
	
	public SimpleNode(String givenNodeData)
	{
		this(new StringBuffer(givenNodeData));
	}
	
	public SimpleNode(StringBuffer givenNodeData, IdWithHostContainer givenRemoteOriginalIdAndHostingPeer, List<StringBuffer> givenXPathList)
	{
		this();
		this.nodeData.append(givenNodeData);
		this.remoteOriginalIdAndHostingPeer = new IdWithHostContainer(givenRemoteOriginalIdAndHostingPeer);
		this.xPathList = new ArrayList<StringBuffer>(givenXPathList);
	}
	
//	SimpleNode(SimpleNode givenSimpleNode, ADNode stopReference)
//	{
//		this();
//		this.currNodeID = givenSimpleNode.currNodeID;
//		this.currDocID = givenSimpleNode.currDocID;
//		this.currPeerID = givenSimpleNode.currPeerID;
//		this.origNodeID = givenSimpleNode.origNodeID;
//		this.origDocID = givenSimpleNode.origDocID;
//		this.origPeerID = givenSimpleNode.origPeerID;
//		this.remoteOriginalIdAndHostingPeer = new IdWithHostContainer(givenSimpleNode.remoteOriginalIdAndHostingPeer);
//		this.xPathList = new ArrayList<StringBuffer>(givenSimpleNode.xPathList);
//		this.nodeData.append(givenSimpleNode.nodeData);
//		this.father = givenSimpleNode.father;
//		this.children = new ExtendedLinkedList<ADNode>();
//		if(givenSimpleNode != stopReference)
//		{
//			Iterator myIterator = givenSimpleNode.children.iterator();
//			while(myIterator.hasNext())
//			{
//				ADNode currentGivenNode = (ADNode) myIterator.next();
//				ADNode myNewNode;
//				if(currentGivenNode instanceof SimpleNode)
//					myNewNode = new SimpleNode((SimpleNode)currentGivenNode, stopReference);
//				
//				else
//					myNewNode = new Service((Service)currentGivenNode, stopReference);
//				
//				myNewNode.setFather(this);
//				this.children.add(myNewNode);
//			}
//		}
////		else
////			System.out.println("Egine Stop!!!!!!!!!!!!!!!!!!!!! (SimpleNode)");
//	}
	
	public void finalize()
	{
			ADNodeCounter.decreaseNumberOfNodes();
	}
	
	public String toString()
	{
		StringBuffer returnStr = new StringBuffer();
		if(remoteOriginalIdAndHostingPeer.isEmpty())
			returnStr.append(nodeData);
		else
			returnStr.append("*" + remoteOriginalIdAndHostingPeer.id.nodeID + "." + remoteOriginalIdAndHostingPeer.id.docID + "." + remoteOriginalIdAndHostingPeer.id.peerID + "@" + remoteOriginalIdAndHostingPeer.host); // we write it like that in order to distinguish that this is nodeDataNotUsedInEqualFunction
		
		if(declaration != null)
		{
			returnStr.append("[");
			returnStr.append(Utilities.DomToStringSerializerWithoutIdentationAndDeclar(declaration));
			returnStr.append("]");
		}
		
		returnStr.append("(");
		
		for(int i=0; i<children.size(); i++)
		{
			returnStr.append(children.get(i).toString());
			if(i != (children.size() -1))
				returnStr.append(", ");
		}
		
		returnStr.append(")");
		
		return returnStr.toString();
	}
	
	public String toStringWithId()
	{
		StringBuffer returnStr;
		
		returnStr = new StringBuffer("[" + this.hashCode() + "]{" + this.currNodeID + "." + this.currDocID + "." + this.currPeerID + "}" + nodeData);
		returnStr.append("(");
		
		for(int i=0; i<children.size(); i++)
		{
			returnStr.append(children.get(i).toStringWithId());
			if(i != (children.size() -1))
				returnStr.append(", ");
		}
		
		returnStr.append(")");
		
		return returnStr.toString();
	}
	
	public boolean equals(Object givenObject)
	{
		if(givenObject instanceof SimpleNode)
		{
			SimpleNode givenSimpleNode = (SimpleNode) givenObject;
			if (this.nodeData.toString().equals(givenSimpleNode.nodeData.toString()) == true)
			{
				if(this.children.size() == givenSimpleNode.children.size())
				{
					Iterator<ADNode> thisChildrenIt = this.children.iterator();
					List<ADNode> givenSimpleNodeChildren = new ArrayList<ADNode>();
					givenSimpleNodeChildren.addAll(givenSimpleNode.children);
					while(thisChildrenIt.hasNext())
					{
						ADNode thisNodeSpecificChild = thisChildrenIt.next();
						
						Iterator givenSimpleNodeChildrenIt = givenSimpleNodeChildren.iterator();

						boolean found = false;
						while(givenSimpleNodeChildrenIt.hasNext())
						{
							ADNode givenSimpleNodeSpecificChild = (ADNode) givenSimpleNodeChildrenIt.next();
							if(thisNodeSpecificChild.equals(givenSimpleNodeSpecificChild) == true)
							{
								found = true;
								givenSimpleNodeChildren.remove(givenSimpleNodeSpecificChild);
								break;
							}
						}
						if(found == false)
							return false;
					}
					return true;
				}				
			}
		}
		return false;
	}
	
	public int hashCode()
	{
//		return 0;
		long returningHashCode = Math.abs(this.nodeData.toString().hashCode()) % Integer.MAX_VALUE;
		Iterator myIterator = this.children.iterator();
		while(myIterator.hasNext())
		{
			returningHashCode += Math.abs(myIterator.next().hashCode()) % Integer.MAX_VALUE;
			returningHashCode = returningHashCode % Integer.MAX_VALUE;
		}
		return (int)returningHashCode;
	}
	
	public boolean equivalent(Object o)
	{
		if (!(o instanceof ADNode)) 
			return false;
		
		// We create the equivalence data trees of both nodes. Note that these trees do
		// not have send & receives and the execution peer of the simple services is
		// deleted. These trees can be compared for equality and if they are equal, we
		// can safelly assume that this & o are equivalent
		ADNode thisEquivalenceDataTree = ADocument.constructEquivalenceDataTree(this);
		ADNode oEquivalenceDataTree = ADocument.constructEquivalenceDataTree((ADNode) o);
		
		if(!ADocument.isSubTreeConsistent(thisEquivalenceDataTree))
		{
			System.err.println("service equivalent: Inconsistency found");
			System.exit(0);
		}
		
		if(!ADocument.isSubTreeConsistent(oEquivalenceDataTree))
		{
			System.err.println("service equivalent: Inconsistency found");
			System.exit(0);
		}

		return oEquivalenceDataTree.equals(thisEquivalenceDataTree);
	}
	
	public int originalHashCode()
	{
		return super.hashCode();
	}
	
	public StringBuffer getAcopyOfTheNodeData()
	{
		return new StringBuffer(nodeData);
	}
	
	public StringBuffer getNodeData()
	{
		return nodeData;
	}
	
	public IdWithHostContainer getRemoteOriginalIdAndHostingPeer()
	{
		return remoteOriginalIdAndHostingPeer;
	}
	
	public void setRemoteOriginalIdAndHostingPeer(IdWithHostContainer givenIdWithHostContainer)
	{
		remoteOriginalIdAndHostingPeer = new IdWithHostContainer(givenIdWithHostContainer);
	}
	
	public List<StringBuffer> getXPathList()
	{
		return xPathList;
	}
	
	public StringBuffer getCurrNodeID()
	{
		return currNodeID;
	}
	
	public StringBuffer getCurrDocID()
	{
		return currDocID;
	}

	public StringBuffer getCurrPeerID()
	{
		return currPeerID;
	}
	
	public StringBuffer getOrigNodeID()
	{
		return origNodeID;
	}
	
	public StringBuffer getOrigDocID()
	{
		return origDocID;
	}

	public StringBuffer getOrigPeerID()
	{
		return origPeerID;
	}
	
	public ADNode getFather()
	{
		return father;
	}
	
	public ExtendedLinkedList<ADNode> getChildren()
	{
		return children;
	}
	
//	public nodeDataObject getnodeData()
//	{
//		return this;
//	}
	
	public void setCurrNodeID(StringBuffer newCurrNodeID)
	{
		this.currNodeID = new StringBuffer(newCurrNodeID);
	}
	
	public void setCurrDocID(StringBuffer newCurrDocID)
	{
		this.currDocID = new StringBuffer(newCurrDocID);
	}
	
	public void setCurrPeerID(StringBuffer newCurrPeerID)
	{
		this.currPeerID = new StringBuffer(newCurrPeerID);
	}
	
	public void setFather(ADNode givenFather)
	{
		this.father = givenFather;
	}
	
	public void setSpecificChildren(List<ADNode> givenArguments)
	{
		this.children.clear();
		this.children.addAll(givenArguments);
		for(int i=0;i<this.children.size(); i++)
		{
			this.children.get(i).setFather(this);
		}
	}
	
	public ExtendedLinkedList<SimpleNode> getOriginalIdWithHostAddressContainersThatPointToThisNode()
	{
		return originalIdWithHostAddressContainersThatPointToThisNode;
	}
	
	public ADocument getReferenceToADocument()
	{
		return hostingDocumentReference;
	}
	
	public void setReferenceToADocument(ADocument givenRefToADoc)
	{
		hostingDocumentReference = givenRefToADoc;
	}
	
	public ADNode returnNewNodeWithNoChildrenAndFatherButWithOldCurrIds()
	{
		SimpleNode newSimpleNode = new SimpleNode(this.origNodeID, this.origDocID, this.origPeerID, this.currNodeID, this.currDocID, this.currPeerID, this.nodeData, this.remoteOriginalIdAndHostingPeer, this.xPathList, null, new ExtendedLinkedList<ADNode>(), this.paramName, this.declaration);
		return newSimpleNode;
	}
	
	public ADNode returnNewNodeWithNoChildrenAndFatherButWithNewCurrIds()
	{
		SimpleNode newSimpleNode = new SimpleNode(this.origNodeID, this.origDocID, this.origPeerID, this.nodeData, this.remoteOriginalIdAndHostingPeer, this.xPathList, null, new ExtendedLinkedList<ADNode>(), this.paramName, this.declaration);
		return newSimpleNode;
	}
	
	public void removeNode()
	{
		if(this.getFather() != null)
			this.getFather().getChildren().refRemove(this);

		this.hostingDocumentReference = null;
		
		this.setFather(null);
	}
	
	public void insertAsChildInBeginning(ADNode givenSubTree)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		this.children.add(0, givenSubTree);
		givenSubTree.setFather(this);
	}
	
	public void insertAsChild(ADNode givenSubTree)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		this.children.add(givenSubTree);
		givenSubTree.setFather(this);
	}
	
	public void insertAsChildNextToGivenChild(ADNode givenSubTree, ADNode givenChild)
	{
		// we update the ReferenceToADocument of the nodes
		// of the givenSubtree
		LinkedList<ADNode> nodesToBeVisited =  new LinkedList<ADNode>();
		nodesToBeVisited.add(givenSubTree);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			nodesToBeVisited.addAll(currNode.getChildren());
			currNode.setReferenceToADocument(this.getReferenceToADocument());
		}
		
		for(int i=0; i<this.children.size(); i++)
		{
			if(this.children.get(i) == givenChild)
			{
				this.children.add(i+1, givenSubTree);
				givenSubTree.setFather(this);
			}
				
		}
	}
}

