package fr.inria.gemo.axml.optimax.info;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import fr.inria.gemo.axml.optimax.Parameters;


public class PeerInformationModule {

	class TwoPeerNamesClass
	{
		String nameOfPeer1;
		String nameOfPeer2;
		
		public boolean equals(Object givenObject)
		{
			if(!(givenObject instanceof TwoPeerNamesClass))
				return false;
			
			TwoPeerNamesClass twoPeerNamesClassGivenObject  = (TwoPeerNamesClass)givenObject;
			if(twoPeerNamesClassGivenObject.nameOfPeer1.equals(this.nameOfPeer1) && twoPeerNamesClassGivenObject.nameOfPeer2.equals(this.nameOfPeer2))
				return true;
			if(twoPeerNamesClassGivenObject.nameOfPeer1.equals(this.nameOfPeer2) && twoPeerNamesClassGivenObject.nameOfPeer2.equals(this.nameOfPeer1))
				return true;
			
			return false;
		}
		
		public int hashCode()
		{
			Integer returningCode = 0;
			returningCode = Math.abs(this.nameOfPeer1.hashCode()) % (Integer.MAX_VALUE/2);
			returningCode += Math.abs(this.nameOfPeer2.hashCode()) % (Integer.MAX_VALUE/2);
			returningCode = returningCode % Integer.MAX_VALUE;
			return returningCode;
		}
	}
	
	Map<String,Map<String,String>> peerInfoContainerPeerNameToServiceNameToEndPoint = null;
	Map<String,List<String>> peerInfoContainerEndPointToPeerNameAndServiceName = null;
	Map<TwoPeerNamesClass,Float> communicationCostsBetweenPeers = null;
	String currDocId = null;
	String currPeerId = null;
	
	private static final PeerInformationModule Instance = new PeerInformationModule();
	
	private PeerInformationModule() 
	{
		peerInfoContainerPeerNameToServiceNameToEndPoint = new HashMap<String,Map<String,String>>();
		peerInfoContainerEndPointToPeerNameAndServiceName = new HashMap<String,List<String>>();
		communicationCostsBetweenPeers = new HashMap<TwoPeerNamesClass,Float>();
	}
	
	public static PeerInformationModule getInstance() { return Instance; }
	
	public void initialize(IDProvider givenIDProviderInstance, P2PInformation givenP2PInformationInstance)
	{
		currDocId = givenIDProviderInstance.getCurrentDocID();
		currPeerId = givenIDProviderInstance.getCurrentPeerID();
		
		List<List<String>> listOfPeerIDServiceNameEndPoint  = givenP2PInformationInstance.getPeerIDServiceNameEndPoint();
		Iterator<List<String>> listOfPeerIDServiceNameEndPointIter = listOfPeerIDServiceNameEndPoint.iterator();
		while(listOfPeerIDServiceNameEndPointIter.hasNext())
		{
			List<String> peerIDServiceNameEndPoint = listOfPeerIDServiceNameEndPointIter.next();
			if(peerIDServiceNameEndPoint.size() != 3)
				System.exit(0);
			else
				insertPeerNameServiceNameEndpoint(peerIDServiceNameEndPoint.get(0), peerIDServiceNameEndPoint.get(1), peerIDServiceNameEndPoint.get(2));
		}
		
		List<List<String>> listOfPairsOfPeersWithComCost  = givenP2PInformationInstance.getComCostForEachPairOfPeers();
		Iterator<List<String>> listOfPairsOfPeersWithComCostIter = listOfPairsOfPeersWithComCost.iterator();
		while(listOfPairsOfPeersWithComCostIter.hasNext())
		{
			List<String> pairOfPeersWithComCost = listOfPairsOfPeersWithComCostIter.next();
			if(pairOfPeersWithComCost.size() != 3)
				System.exit(0);
			else
				insertPairOfPeersWithComCost(pairOfPeersWithComCost.get(0), pairOfPeersWithComCost.get(1), Integer.valueOf(pairOfPeersWithComCost.get(2)));
		}
	}
	
	public List<StringBuffer> getListOfPeers()
	{
		List<StringBuffer> returningList = new LinkedList<StringBuffer>();
		Iterator<Entry<String,Map<String,String>>> myIter = peerInfoContainerPeerNameToServiceNameToEndPoint.entrySet().iterator();
		
		while(myIter.hasNext())
		{
			Entry<String,Map<String,String>> myEntry = myIter.next();
			if(!myEntry.getKey().equals(Parameters.ANY))
				returningList.add(new StringBuffer(myEntry.getKey()));
		}

		return returningList;
	}
	
	public void setAvailaiblePeers(List<StringBuffer> givenPeers)
	{
		Iterator<StringBuffer> givenPeersIter = givenPeers.iterator();
		while(givenPeersIter.hasNext())
		{
			StringBuffer tempPeerName = givenPeersIter.next();
			if(!peerInfoContainerPeerNameToServiceNameToEndPoint.containsKey(tempPeerName.toString()))
				peerInfoContainerPeerNameToServiceNameToEndPoint.put(tempPeerName.toString(), new HashMap<String,String>());
		}
	}
	
	boolean insertPeerNameServiceNameEndpoint(String peerName, String serviceName, String endPoint)
	{
		if(peerInfoContainerPeerNameToServiceNameToEndPoint.containsKey(peerName))
			if(peerInfoContainerPeerNameToServiceNameToEndPoint.get(peerName).containsKey(serviceName))
				return false;
			else
			{
				peerInfoContainerPeerNameToServiceNameToEndPoint.get(peerName).put(serviceName, endPoint);
				List<String> peerNameAndServiceNameHolder = new LinkedList<String>();
				peerNameAndServiceNameHolder.add(peerName);
				peerNameAndServiceNameHolder.add(serviceName);
				peerInfoContainerEndPointToPeerNameAndServiceName.put(endPoint, peerNameAndServiceNameHolder);
				return true;
			}
		
		else
		{
			Map<String,String> tempServiceEndPointMap = new HashMap<String,String>();
			tempServiceEndPointMap.put(serviceName, endPoint);
			peerInfoContainerPeerNameToServiceNameToEndPoint.put(peerName, tempServiceEndPointMap);
			List<String> peerNameAndServiceNameHolder = new LinkedList<String>();
			peerNameAndServiceNameHolder.add(peerName);
			peerNameAndServiceNameHolder.add(serviceName);
			peerInfoContainerEndPointToPeerNameAndServiceName.put(endPoint, peerNameAndServiceNameHolder);
			return true;
		}
	}
	
	boolean insertPairOfPeersWithComCost(String peerName1, String peerName2, float comCost)
	{
		TwoPeerNamesClass tempTwoPeerNamesClass1 = new TwoPeerNamesClass();
		tempTwoPeerNamesClass1.nameOfPeer1 = peerName1;
		tempTwoPeerNamesClass1.nameOfPeer2 = peerName1;
		communicationCostsBetweenPeers.put(tempTwoPeerNamesClass1, (float)0);
			
		TwoPeerNamesClass tempTwoPeerNamesClass2 = new TwoPeerNamesClass();
		tempTwoPeerNamesClass2.nameOfPeer1 = peerName2;
		tempTwoPeerNamesClass2.nameOfPeer2 = peerName2;
		communicationCostsBetweenPeers.put(tempTwoPeerNamesClass2, (float)0);
		
		TwoPeerNamesClass tempTwoPeerNamesClass3 = new TwoPeerNamesClass();
		tempTwoPeerNamesClass3.nameOfPeer1 = peerName1;
		tempTwoPeerNamesClass3.nameOfPeer2 = peerName2;
		if(communicationCostsBetweenPeers.containsKey(tempTwoPeerNamesClass3))
			return false;
		communicationCostsBetweenPeers.put(tempTwoPeerNamesClass3, comCost);
		return true;
	}
	
	public String getEndPoint(String peerName, String serviceName)
	{
		Map<String,String> serviceEndPointMap = peerInfoContainerPeerNameToServiceNameToEndPoint.get(peerName);
		if(serviceEndPointMap != null)
		{
			String endPoint = serviceEndPointMap.get(serviceName);
			if(endPoint != null)
				return endPoint;
			else
				return new String();
		}
		else
			return new String();
	}
	
	public List<String> getPeerNameServiceName(String endPoint)
	{
		return peerInfoContainerEndPointToPeerNameAndServiceName.get(endPoint);
	}
	
	public String getCurrDocId()
	{
		return currDocId;
	}
	
	public String getCurrPeerId()
	{
		return currPeerId;
	}
	
	public float getComCost(String peerName1, String peerName2)
	{
		TwoPeerNamesClass tempTwoPeerNamesClass = new TwoPeerNamesClass();
		tempTwoPeerNamesClass.nameOfPeer1 = peerName1;
		tempTwoPeerNamesClass.nameOfPeer2 = peerName2;
		return this.communicationCostsBetweenPeers.get(tempTwoPeerNamesClass);
	}
}
