package fr.inria.gemo.axml.optimax.io;

import java.io.StringWriter;
import java.util.List;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPathExpressionException;

import org.w3c.dom.Document;

import fr.inria.gemo.axml.optimax.PlanRepository;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.io.input.FileInputModule;
import fr.inria.gemo.axml.optimax.io.output.FileOutputModule;

public class FileManager {
	private FileInputModule fileInputModuleInstance;
	private FileOutputModule fileOutputModuleInstance;
	
	private static final FileManager Instance = new FileManager();
	private FileManager() 
	{
		fileInputModuleInstance = new FileInputModule();
		fileOutputModuleInstance = new FileOutputModule();
	}
	
	public static FileManager getInstance() { return Instance; }
	
	public void createOptimizationFiles(PlanRepository givenPlanRepository, Document originalDom, StringBuffer originalFileName)
	{
		fileOutputModuleInstance.createOptimizationFiles(givenPlanRepository, originalDom, originalFileName);
	}
	
//	public ADocument getFullDocFromSavedDoc(String docName) 
//	{ 
//		return fileInputModuleInstance.getFullDocFromSavedDoc(docName); 
//	}

	public void createOutputImages(PlanRepository givenPlanRepository)
	{
		fileOutputModuleInstance.createOutputImages(givenPlanRepository);
	}
	
	public void createGnuPlotGraphs(PlanRepository givenPlanRepository)
	{
		fileOutputModuleInstance.createGnuPlotGraphs(givenPlanRepository);
	}
	
	public List<StringBuffer> getListOfPeers(String listOfPeersFileName)
	{
		return fileInputModuleInstance.getListOfPeers(listOfPeersFileName);
	}
	
//	public ADocument getSkeletonDocFromSavedDoc(String docName)
//	{
//		return fileInputModuleInstance.getSkeletonDocFromSavedDoc(docName);
//	}
	
//	public Map<Service, ArrayList<Service>> returnMappings(String nameOfMappingFile)
//	{
//		return fileInputModuleInstance.returnMappings(nameOfMappingFile);
//	}
	
	public Document getDomFromXmlFile(String fileName)
	{
		return fileInputModuleInstance.getDomFromXmlFile(fileName);
	}
	
	public static void createDebugOutputImage(ADocument givenADocument, String fileName)
	{
		FileOutputModule.createDebugOutputImage(givenADocument, fileName);
	}
	
	public String getDomAsString(Document givenDom)
	{
		String returningString = new String();
		
	    try
	    {
	        TransformerFactory tranFactory = TransformerFactory.newInstance(); 
	        Transformer aTransformer = tranFactory.newTransformer();
	        
	        
	        StringWriter out = new StringWriter();
	        
	        Source src = new DOMSource(givenDom); 
	        Result dest = new StreamResult(out); 
	        aTransformer.transform(src, dest);
	        
	        returningString = out.toString();
	    }catch(Exception e)
	    {
	          System.out.println(e.getMessage());
	    }
	    
	    return returningString;
	}
	
	public static ADocument constructADocFromAxmlDom(Document givenDom) throws XPathExpressionException
	{
		return FileInputModule.constructADocFromAxmlDom(givenDom);
	}
}
