package fr.inria.gemo.axml.optimax.io.input;

import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import com.sun.org.apache.xerces.internal.util.NamespaceSupport;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.IdContainer;
import fr.inria.gemo.axml.optimax.adocument.IdWithHostContainer;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.helping.Utilities;
import fr.inria.gemo.axml.optimax.helping.NamespaceContextWrapper;
import fr.inria.gemo.axml.optimax.info.DefaultNamespaceDeclarator;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;

public class FileInputModule {

	public static byte[] readDataFromBinaryFile(String fileName) throws IOException
	{
	    String completeFileName = fileName;
	    if (!fileName.startsWith(Parameters.OUTPUT_DIR)){
		completeFileName = new String(Parameters.OUTPUT_DIR + fileName);
	    }
		FileInputStream fileIn = new FileInputStream(completeFileName);
		byte[] biteBuffer = new byte[fileIn.available()];
		fileIn.read(biteBuffer);
		return biteBuffer;
	}
	
	public Document getDomFromXmlFile(String fileName)
	{
		Document doc = null;
		
		try{
			//Create instance of DocumentBuilderFactory
	        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	        factory.setIgnoringElementContentWhitespace(true);
	        factory.setValidating(false);
	        factory.setNamespaceAware(true);
	        
	        //Get the DocumentBuilder
	        DocumentBuilder docBuilder = factory.newDocumentBuilder();
	        Validator handler = new Validator();
	        docBuilder.setErrorHandler(handler);
	        
	        doc = docBuilder.parse(fileName);
		}
		catch(Exception e)
		{
			System.out.println(e.getMessage());
		}
		return doc;
	}
	
	private class Validator extends DefaultHandler {
		public boolean validationError = false;  
		public SAXParseException saxParseException = null; 
		public void error(SAXParseException exception) throws SAXException 
		{
			validationError = true;
			saxParseException = exception;
			System.err.println("Parse error at line " + exception.getLineNumber() + ": " + exception.getMessage());
		}     
		public void fatalError(SAXParseException exception) throws SAXException 
		{
			validationError = true;	    
			saxParseException = exception;
			System.err.println("Parse error at line " + exception.getLineNumber() + ": " + exception.getMessage());
		}		    
		public void warning(SAXParseException exception) throws SAXException 
		{ 
			System.err.println("Parse error at line " + exception.getLineNumber() + ": " + exception.getMessage());
		}	
	}
	
	public List<StringBuffer> getListOfPeers(String listOfPeersFileName)
	{
		Document dom = getDomFromXmlFile(listOfPeersFileName);
		return getListOfPeersFromDom(dom);
	}
	
	private List<StringBuffer> getListOfPeersFromDom(Document givenDocument)
	{
		List<StringBuffer> listToBeReturned = new ArrayList<StringBuffer>();
		
		Node peersNode = givenDocument.getDocumentElement();
		if(!peersNode.getNodeName().toString().equals("peers"))
		{
			System.err.println("This is not a 'peers' document");
			return null;
		}

		if(!peersNode.hasChildNodes())
		{
			System.err.println("This an empty 'peers' document");
			return null;
		}
		
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		
		try 
		{
			NodeList peerNodeList = (NodeList) xPath.evaluate("//peer" ,givenDocument, XPathConstants.NODESET);
			for(int i=0; i<peerNodeList.getLength(); i++)
				listToBeReturned.add(new StringBuffer((String) xPath.evaluate("@name" ,peerNodeList.item(i), XPathConstants.STRING)));
		} 
		catch (XPathExpressionException e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return listToBeReturned;
	}

	public static ADocument constructADocFromAxmlDom(Document givenDom) throws XPathExpressionException
	{
		ADocument returningADoc = new ADocument();
		
		removingNotNeededSpacesAndCommentsFromDom(givenDom);
		
		class NodesWithPath 
		{ 
			List<Integer> path = null; 
			Node node = null; 
			NodesWithPath(Node givenNode, List<Integer> givenPath) { node = givenNode; path = givenPath;} 
		}
		Map<Node,ADNode> axmlDomNodesToADocNodesMap = new HashMap<Node,ADNode>();

		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
		Node origRoot = givenDom.getDocumentElement();
		LinkedList<NodesWithPath> pNodesToBeVisited = new LinkedList<NodesWithPath>();
		
		List<Integer> origRootPath = new ArrayList<Integer>();
		origRootPath.add(1);
		pNodesToBeVisited.add(new NodesWithPath(origRoot, origRootPath));
		// In this loop (first) we construct the ADNodes that correspond to the dom nodes.
		// We also fill the axmlDomNodesToADocNodesMap
		while(!pNodesToBeVisited.isEmpty())
		{
			NodesWithPath currPNode = pNodesToBeVisited.removeFirst();

			if(currPNode.node.getNodeType() == Node.ELEMENT_NODE)
			{
//				System.err.println("!!!!!!!!!!NodeName:" + currPNode.node.getNodeName());
//				System.err.println("!!!!!!!!!!Prefix:" + currPNode.node.getPrefix());
//				System.err.println("!!!!!!!!!!NamespaceURI:" + currPNode.node.getNamespaceURI());

				String axmlPrefix = myNamespaceSupport.getPrefix(Parameters.axmlURI);
				
//				System.out.println("axmlPrefix " + axmlPrefix);
//				System.out.println(new String(axmlPrefix + ":sc"));
//				System.exit(0);
				
				if(currPNode.node.getNodeName().equals(new String(axmlPrefix + ":sc")))
				{
					String endPoint = new String();
					String currNodeId = new String();
					String origNodeId = new String();
					String origDocId = new String();
					String origPeerId = new String();
					Node wsSoapRestNode = null;
					
					Node wssoapNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-soap") , currPNode.node, XPathConstants.NODE);
					if(wssoapNode != null)
						wsSoapRestNode = wssoapNode;
				
					Node wsRestNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-rest") , currPNode.node, XPathConstants.NODE);
					if(wsRestNode != null)
						wsSoapRestNode = wsRestNode;

					endPoint = (String) xPath.evaluate("@endpoint" , wsSoapRestNode, XPathConstants.STRING);
					currNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":id") , currPNode.node, XPathConstants.STRING);
					origNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":nodeID") , currPNode.node, XPathConstants.STRING);
					origDocId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":docID") , currPNode.node, XPathConstants.STRING);
					origPeerId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":peerID") , currPNode.node, XPathConstants.STRING);
					
					// We calculate the hashcode of xml subtree that we are going to use at the specific service
					int hashCodeForService = calculateTheHashCodeForService(currPNode.node, axmlPrefix);
					
					PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
					// We get a peer and service codename that are related with that endpoint 
					List<String> peerNameAndServiceNameContainer = peerInformationModuleInstance.getPeerNameServiceName(endPoint);
					// We create a new service with the information that we extracted from the axml file
					System.out.println("peerNameAndServiceNameContainer:" + peerNameAndServiceNameContainer);
					System.out.println("endPoint:" + endPoint);
					System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!");
					Service newServiceNode = new Service(new StringBuffer(origNodeId), new StringBuffer(origDocId) , new StringBuffer(origPeerId), false, new HashSet<String>(), new HashSet<String>(), new StringBuffer(peerNameAndServiceNameContainer.get(1)), new StringBuffer(peerNameAndServiceNameContainer.get(0)), new StringBuffer(), constructXPathFromPathList(currPNode.path), hashCodeForService, null, new ExtendedLinkedList<ADNode>());
					newServiceNode.setReferenceToADocument(returningADoc);
					returningADoc.newScCurrNodeIdsToOld.put(newServiceNode.getCurrNodeID().toString(), currNodeId);
					returningADoc.oldScCurrNodeIdsToNew.put(currNodeId, newServiceNode.getCurrNodeID().toString());
					
					axmlDomNodesToADocNodesMap.put(currPNode.node, newServiceNode);
					
//					System.out.println(constructXPathFromPathList(currPNode.path));
//					System.err.println("---->endPoint " + endPoint);
//					System.err.println("---->newServiceNode " + newServiceNode);
				}
				else if(!currPNode.node.getNodeName().equals(new String(axmlPrefix + ":activation")) && !currPNode.node.getNodeName().equals(new String(axmlPrefix + ":return")) && !currPNode.node.getNodeName().equals(new String(axmlPrefix + "meta")))
				{
//					NodeList scNodeList1 = null;
//					NodeList scNodeList2 = null;
					String origNodeId = null;
					String origNodeId2 = null;
					String origDocId = null;
					String origDocId2 = null;
					String origPeerId = null;
					String origPeerId2 = null;
					String currNodeId = null;
					String currNodeId2 = null;
					
					currNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":id") , currPNode.node, XPathConstants.STRING);
					origNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":nodeID") , currPNode.node, XPathConstants.STRING);
					origDocId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":docID") , currPNode.node, XPathConstants.STRING);
					origPeerId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":peerID") , currPNode.node, XPathConstants.STRING);
					currNodeId2 = (String) xPath.evaluate(new String("*//@" + axmlPrefix + ":id") , currPNode.node, XPathConstants.STRING);
					origNodeId2 = (String) xPath.evaluate(new String("*//@" + axmlPrefix + ":nodeID") , currPNode.node, XPathConstants.STRING);
					origDocId2 = (String) xPath.evaluate(new String("*//@" + axmlPrefix + ":docID") , currPNode.node, XPathConstants.STRING);
					origPeerId2 = (String) xPath.evaluate(new String("*//@" + axmlPrefix + ":peerID") , currPNode.node, XPathConstants.STRING);

					// If there is no sc element under currPNode I will not visit its children
					// (that's why I have added 'continue' at the end of the following if.
					// I also do not want the node to be a query declaration node neither a
					// query param node
					String queryPrefix = myNamespaceSupport.getPrefix(Parameters.queryURI);
					if(!(currPNode.node.getNodeName().equals(new String(queryPrefix + ":declaration"))) && !(currPNode.node.getNodeName().equals(new String(queryPrefix + ":param"))) && (currNodeId.length() == 0 && currNodeId2.length() == 0 && origNodeId.length() == 0 && origNodeId2.length() == 0 && origDocId.length() == 0 && origDocId2.length() == 0 && origPeerId.length() == 0 && origPeerId2.length() == 0))
					{
							boolean addedAtASibling = false;
							// I search if there is another simplenode without children where I can append the
							// xpath to the specific dom node
							Node currDomNode = currPNode.node;
							for(Node prevSibling = currDomNode.getPreviousSibling(); prevSibling != null && !prevSibling.equals(currPNode.node); prevSibling = currDomNode.getPreviousSibling())
							{
								currDomNode = prevSibling;
								if(prevSibling.getNodeType() == Node.ELEMENT_NODE)
								{
									ADNode prevSiblingADNode = axmlDomNodesToADocNodesMap.get(prevSibling);
									if( prevSiblingADNode != null && prevSiblingADNode instanceof SimpleNode)
									{
										SimpleNode simpleNodePrevSiblingADNode = (SimpleNode) prevSiblingADNode;
										
										// If the size of the xpath list is more than 1 then this node definatelly
										// does not have children
										if(simpleNodePrevSiblingADNode.getXPathList().size() > 1)
										{
											SimpleNode prevSimpleNode = (SimpleNode)prevSiblingADNode;
											prevSimpleNode.getXPathList().add(constructXPathFromPathList(currPNode.path));
											prevSimpleNode.getNodeData().append(", " + currPNode.node.getNodeName());
											addedAtASibling = true;
											break;
										}
										// If the size of the xpath list is 1 then this node may have some services
										// as descendants that are not yet attached
										else if(simpleNodePrevSiblingADNode.getXPathList().size() == 1)
										{
											Node tempDomNode = (Node) xPath.evaluate(simpleNodePrevSiblingADNode.getXPathList().get(0).toString() , givenDom.getDocumentElement(), XPathConstants.NODE);
											String tempNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":id") , tempDomNode, XPathConstants.STRING);
											String tempNodeId2 = (String) xPath.evaluate(new String("*//@" + axmlPrefix + ":id") , tempDomNode, XPathConstants.STRING);
											if(tempNodeId.length() == 0 && tempNodeId2.length() == 0)
											{
												SimpleNode prevSimpleNode = (SimpleNode)prevSiblingADNode;
												prevSimpleNode.getXPathList().add(constructXPathFromPathList(currPNode.path));
												prevSimpleNode.getNodeData().append(", " + currPNode.node.getNodeName());
												addedAtASibling = true;
												break;
											}
										}
									}
								}
							}
							// If I do not find such a simplenode, I create a new simplenode that will host the
							// specific xpath
							if(!addedAtASibling)
							{
								List<StringBuffer> currXPathList = new ArrayList<StringBuffer>();
								currXPathList.add(constructXPathFromPathList(currPNode.path));
								SimpleNode newSimpleNode = new SimpleNode(new StringBuffer(), new StringBuffer(), new StringBuffer(), new StringBuffer(currPNode.node.getNodeName()), new IdWithHostContainer(), currXPathList, null, new ExtendedLinkedList<ADNode>(), new String(), null);
								newSimpleNode.setReferenceToADocument(returningADoc);
								axmlDomNodesToADocNodesMap.put(currPNode.node, newSimpleNode);
							}
							
							continue;
					}
					// if there is-are service calls under currPNode
					// or it is a query declaration either a query param node
					else
					{
						Document declarationDom = null;
						String nameOfParam = null;
						
						if(currPNode.node.getNodeName().equals(new String(queryPrefix + ":declaration")))
						{
							DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
							DocumentBuilder loader;
							
							try {
								loader = dfactory.newDocumentBuilder();
								declarationDom = loader.newDocument();
								Node rootNodeForDeclarationDom = declarationDom.importNode(currPNode.node, true);
								declarationDom.appendChild(rootNodeForDeclarationDom);
								
//								System.out.println("xoxo");
//								System.out.println(Utilities.DomToStringSerializerWithIdentation(declarationDom));
//								System.out.println("xoxo");
//								System.exit(0);
								
							} catch (ParserConfigurationException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
						
						if(currPNode.node.getNodeName().equals(new String(queryPrefix + ":param")))
						{
							nameOfParam = (String) xPath.evaluate(new String("@" + "name") , currPNode.node, XPathConstants.STRING);
//							System.out.println("lala");
//							System.out.println(nameOfParam);
//							System.out.println("lala");
//							System.exit(1);
						}
						
						List<StringBuffer> currXPathList = new ArrayList<StringBuffer>();
						currXPathList.add(constructXPathFromPathList(currPNode.path));
						SimpleNode newSimpleNode = new SimpleNode(new StringBuffer(origNodeId), new StringBuffer(origDocId), new StringBuffer(origPeerId), new StringBuffer(currPNode.node.getNodeName()), new IdWithHostContainer(), currXPathList, null, new ExtendedLinkedList<ADNode>(), new String(), null);
						newSimpleNode.declaration = declarationDom;
						newSimpleNode.paramName = nameOfParam;
						newSimpleNode.setReferenceToADocument(returningADoc);
						//TODO maybe I have to look again why I have commented the bellow
						//TODO (I have done it because I do not care if I have correspondance
						//TODO between new and old currNodeIds
//						if(currNodeId.length() != 0)
//						{
//							returningADoc.newServiceCurrNodeIdsToOld.put(newSimpleNode.getCurrNodeID().toString(), currNodeId);
//							returningADoc.oldServiceCurrNodeIdsToNew.put(currNodeId, newSimpleNode.getCurrNodeID().toString());
//						}
						axmlDomNodesToADocNodesMap.put(currPNode.node, newSimpleNode);
					}
				}
			}
			
			// if currNode has children then it has to visit them
			if(currPNode.node.hasChildNodes())
			{
				NodeList tempNodeList = currPNode.node.getChildNodes();
				int childNumber = 0;
				for(int i=0;i<tempNodeList.getLength();i++)
				{
					Node tempChildNode = tempNodeList.item(i);
					if(tempChildNode.getNodeType() == Node.ELEMENT_NODE)
						childNumber++;
					List<Integer> tempChildPath = new ArrayList<Integer>(currPNode.path);
					tempChildPath.add(childNumber);
					pNodesToBeVisited.addLast(new NodesWithPath(tempChildNode, tempChildPath));
				}
			}
		}
		
		System.out.println(returningADoc.newScCurrNodeIdsToOld);
		System.out.println(returningADoc.oldScCurrNodeIdsToNew);
		
		// In this (second) loop we construct the ADNode tree having as reference the dom tree
		// that was given to us.
		LinkedList<Node> nodesToBeVisited = new LinkedList<Node>();
		nodesToBeVisited.add(origRoot);
		while(!nodesToBeVisited.isEmpty())
		{
			Node currNode = nodesToBeVisited.removeFirst();
			
			if(currNode.hasChildNodes())
			{
				ADNode currADNode = axmlDomNodesToADocNodesMap.get(currNode);
				if(currADNode == null)
					continue;
				else
				{
					System.out.println("=======================");
					
					if(currADNode instanceof Service)
						System.out.println(((Service)currADNode).getServiceName());
					else
						System.out.println(((SimpleNode)currADNode).getNodeData());						
					
					System.out.println(currADNode.getCurrNodeID());
					System.out.println(currADNode.getCurrDocID());
					System.out.println(currADNode.getCurrPeerID());
					System.out.println(currADNode.getOrigNodeID());
					System.out.println(currADNode.getOrigDocID());
					System.out.println(currADNode.getOrigPeerID());
					System.out.println("=======================");
				}
				for(Node tempChildNode = currNode.getFirstChild(); tempChildNode != currNode.getLastChild(); tempChildNode = tempChildNode.getNextSibling())
				{
					ADNode tempChildADNode = axmlDomNodesToADocNodesMap.get(tempChildNode);
					if(tempChildADNode == null)
						continue;
					currADNode.insertAsChild(tempChildADNode);
				}
				ADNode tempLastChildADNode = axmlDomNodesToADocNodesMap.get(currNode.getLastChild());
				if(tempLastChildADNode != null)
				{
					currADNode.insertAsChild(tempLastChildADNode);
				}
			}
			
			// if currNode has children then it has to visit them
			if(currNode.hasChildNodes())
			{
				NodeList tempNodeList = currNode.getChildNodes();
				for(int i=0;i<tempNodeList.getLength();i++)
				{
					Node tempChildNode = tempNodeList.item(i);
					nodesToBeVisited.addLast(tempChildNode);
				}
			}
		}
		
		if(axmlDomNodesToADocNodesMap.get(origRoot) != null)
			returningADoc.setRootOfADoc(axmlDomNodesToADocNodesMap.get(origRoot));
		else
		{
			System.err.println("<---------------------->");
		}
		
		// Here we are going to calculate the hashCodeOfTreeRootedByCurrService of 
		// all the services in the tree. For this reason we search the tree in breadth
		// first manner. For each service we do the following computation:
		// hashCodeOfTreeRootedByCurrService = (hashCodeOfTreeRootedByCurrService - forEachDirectPredecessor.hashCodeOfTreeRootedByCurrService)modInteger.MAX_VALUE
		// Note that before this computation every node has in hashCodeOfTreeRootedByCurrService
		// the modHashCode value of the xml tree to which is a root (this tree includes
		// and the descendant service calls which we do not want to be included in the
		// computation of the hashCodeOfTreeRootedByCurrService for that service call)
		LinkedList<ADNode> aDNodesToBeVisited = new LinkedList<ADNode>();
		aDNodesToBeVisited.add(returningADoc.getRootOfADoc());
		while(!aDNodesToBeVisited.isEmpty())
		{
			ADNode currADNode = aDNodesToBeVisited.removeFirst();
			aDNodesToBeVisited.addAll(currADNode.getChildren());
			if(currADNode instanceof Service)
			{
				// if it is a service call then we are going to search for
				// all the immediate descendant service calls in order to
				// calculate the correct value for hashCodeOfTreeRootedByCurrService
				// of that sc
				Service serviceCurrADNode = (Service)currADNode;
				LinkedList<ADNode> innerADNodesToBeVisited = new LinkedList<ADNode>();
				innerADNodesToBeVisited.addAll(currADNode.getChildren());
				while(!innerADNodesToBeVisited.isEmpty())
				{
					ADNode innerCurrADNode = innerADNodesToBeVisited.removeFirst();
					if(innerCurrADNode instanceof Service)
					{
						// Here for each immediate descendant service of currADNode
						// we update its hashCodeOfTreeRootedByCurrService
						Service serviceInnerCurrADNode = (Service) innerCurrADNode;
						serviceCurrADNode.setHashCodeOfTreeRootedByCurrService(Utilities.amodN(serviceCurrADNode.getHashCodeOfTreeRootedByCurrService() - serviceInnerCurrADNode.getHashCodeOfTreeRootedByCurrService(), Integer.MAX_VALUE).intValue());
					}
					else
						innerADNodesToBeVisited.addAll(innerCurrADNode.getChildren());
				}
			}
		}
		
		returningADoc.calculateHashCode();
		
//		GraphvizModule.createOutputImage(returningADoc, "constructADocFromAxmlDom");
		
//		getServerAddressFromUri("http://localhost:6969/axml/services/myPriceFilterService1");
//		System.out.println(HelpingMethods.DomToStringSerializer(AxmlFileOutputModule.createAxmlFileFromRewrittenADocumentAndOriginalDom(returningADoc, domOfAxml)));
		
		return returningADoc;
	}

	private static void removingNotNeededSpacesAndCommentsFromDom(Document givenDom)
	{
		Node origRoot = givenDom.getDocumentElement();
		LinkedList<Node> nodesToBeVisited = new LinkedList<Node>();
		
		nodesToBeVisited.add(origRoot);
		while(!nodesToBeVisited.isEmpty())
		{
			Node currNode = nodesToBeVisited.removeFirst();
			
			switch(currNode.getNodeType()) {
			case Node.TEXT_NODE : 
				if(currNode.getNodeValue().trim().length() == 0)
					currNode.getParentNode().removeChild(currNode);
				
				break;
			case Node.COMMENT_NODE : 
				currNode.getParentNode().removeChild(currNode);
				break;
			default: 
			}
			
			// if currNode has children then it has to visit them
			if(currNode.hasChildNodes())
			{
				NodeList tempNodeList = currNode.getChildNodes();
				for(int i=0;i<tempNodeList.getLength();i++)
					nodesToBeVisited.addLast(tempNodeList.item(i));
			}
		}
	}
	// This method takes as input a Node and calculates the hashcode of the xml tree rooted by
	// that node. In order to do this it removes from the xml tree the attributes: id, activated, 
	// nodeID, docID, peerID and the element activation. This method is really important in order
	// to determine equivalence between two different services
	private static int calculateTheHashCodeForService(Node givenNode, String axmlPrefix) throws XPathExpressionException
	{
		int hashCodeForService = 0;
		
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
		Document tempDom = null;
		DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder loader;
		try {
			loader = dfactory.newDocumentBuilder();
			tempDom = loader.newDocument();
			Node rootOfTempDom = tempDom.importNode(givenNode, true);
			tempDom.appendChild(rootOfTempDom);
			
			// Here I will delete the endpoints from all the services that are from rootOfTempDom and bellow 
			List<Node> listWithNodesToBeRemoved = new ArrayList<Node>();
			Node wsSoapRestNode = null;
			Node wssoapNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-soap") , rootOfTempDom, XPathConstants.NODE);
			if(wssoapNode != null)
				wsSoapRestNode = wssoapNode;
			Node wsRestNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-rest") , rootOfTempDom, XPathConstants.NODE);
			if(wsRestNode != null)
				wsSoapRestNode = wsRestNode;
			NodeList nodesToBeRemoved = (NodeList) xPath.evaluate("@endpoint" , wsSoapRestNode, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate("*//@endpoint" , wsSoapRestNode, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			// Here I will delete all the id attributes that are in from rootOfTempDom and underneath
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("@" + axmlPrefix + ":id") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//@" + axmlPrefix + ":id") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			// the same for activated attributes
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("@" + axmlPrefix + ":activated") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//@" + axmlPrefix + ":activated") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));

			// the same for nodeID attributes
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("@" + axmlPrefix + ":nodeID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//@" + axmlPrefix + ":nodeID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			// the same for docID attributes
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("@" + axmlPrefix + ":docID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//@" + axmlPrefix + ":docID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			// the same for peerID attributes
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("@" + axmlPrefix + ":peerID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//@" + axmlPrefix + ":peerID") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			Iterator<Node> listWithNodesToBeRemovedIter = listWithNodesToBeRemoved.iterator();
			while(listWithNodesToBeRemovedIter.hasNext())
			{
				Node nodeForDeletion = listWithNodesToBeRemovedIter.next();
				System.out.println(nodeForDeletion);
				System.out.println(nodeForDeletion.getParentNode());
				System.out.println(((Attr)nodeForDeletion).getOwnerElement());
				((Element)((Attr)nodeForDeletion).getOwnerElement()).removeAttributeNode((Attr)nodeForDeletion);
			}
			
			listWithNodesToBeRemoved.clear();
			
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String(axmlPrefix + ":activation") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			nodesToBeRemoved = (NodeList) xPath.evaluate(new String("*//" + axmlPrefix + ":activation") , rootOfTempDom, XPathConstants.NODESET);
			for(int i=0; i<nodesToBeRemoved.getLength(); i++)
				listWithNodesToBeRemoved.add(nodesToBeRemoved.item(i));
			
			listWithNodesToBeRemovedIter = listWithNodesToBeRemoved.iterator();
			while(listWithNodesToBeRemovedIter.hasNext())
			{
				Node nodeForDeletion = listWithNodesToBeRemovedIter.next();
				System.out.println(nodeForDeletion);
				nodeForDeletion.getParentNode().removeChild(nodeForDeletion);
			}
			
			hashCodeForService = Utilities.modHashCode(Utilities.DomToStringSerializerWithoutIdentationAndDeclar(tempDom).toString());
		} catch (ParserConfigurationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return hashCodeForService;
	}
	
	private static StringBuffer constructXPathFromPathList(List<Integer> givenPath)
	{
		StringBuffer path = new StringBuffer();
		Iterator myPathIterator = givenPath.iterator();
		while(myPathIterator.hasNext())
		{
			path.append("/*[");
			path.append(myPathIterator.next());
			path.append("]");
		}
		return path;
	}

	public static ADocument constructADocForPicsFromAxmlDom(Document givenDom) throws XPathExpressionException
	{
		ADocument returningADoc = new ADocument();
		removingNotNeededSpacesAndCommentsFromDom(givenDom);
		
		class NodesWithPath 
		{ 
			List<Integer> path = null; 
			Node node = null; 
			NodesWithPath(Node givenNode, List<Integer> givenPath) { node = givenNode; path = givenPath;} 
		}
		Map<Node,ADNode> axmlDomNodesToADocNodesMap = new HashMap<Node,ADNode>();

		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
		// Here we store the prefixes that we will use
		String axmlPrefix = myNamespaceSupport.getPrefix(Parameters.axmlURI);
		String algebraPrefix = myNamespaceSupport.getPrefix(Parameters.algebraURI);
		String queryPrefix = myNamespaceSupport.getPrefix(Parameters.queryURI);
		String streamSrvPrefix = myNamespaceSupport.getPrefix(Parameters.streamSrvURI);
		
		// Here we are going to store the tags of the nodes which we do not want
		// to represent in the adoc
		HashSet<String> nodesThatShouldNotBeRepresented = new HashSet<String>();
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":receive"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":send"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":newNode"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":receive"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":address"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":originalID"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":peerID"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":docID"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":nodeID"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":ws-soap"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":activation"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":return"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":meta"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":append"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":data"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":reference"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":no-merge"));
		nodesThatShouldNotBeRepresented.add(new String(algebraPrefix + ":endpoint"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":void"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":replace"));
		nodesThatShouldNotBeRepresented.add(new String(axmlPrefix + ":afterActivated"));
		nodesThatShouldNotBeRepresented.add(new String(queryPrefix + ":continuousQuery"));
		nodesThatShouldNotBeRepresented.add(new String(queryPrefix + ":param"));
		nodesThatShouldNotBeRepresented.add(new String(queryPrefix + ":declaration"));
		nodesThatShouldNotBeRepresented.add(new String(streamSrvPrefix + ":max-timeout"));
		nodesThatShouldNotBeRepresented.add(new String(streamSrvPrefix + ":streamToMe"));
		
		Node origRoot = givenDom.getDocumentElement();
		LinkedList<NodesWithPath> pNodesToBeVisited = new LinkedList<NodesWithPath>();
		
		List<Integer> origRootPath = new ArrayList<Integer>();
		origRootPath.add(1);
		pNodesToBeVisited.add(new NodesWithPath(origRoot, origRootPath));
		// In this loop (first) we construct the ADNodes that correspond to the dom nodes.
		// We also fill the axmlDomNodesToADocNodesMap
		while(!pNodesToBeVisited.isEmpty())
		{
			NodesWithPath currPNode = pNodesToBeVisited.removeFirst();

			if(currPNode.node.getNodeType() == Node.ELEMENT_NODE)
			{
				if(currPNode.node.getNodeName().equals(new String(axmlPrefix + ":sc")))
				{
					String endPoint = new String();
					String currNodeId = new String();
					String origNodeId = new String();
					String origDocId = new String();
					String origPeerId = new String();
					Node wsSoapRestNode = null;
					
					Node wssoapNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-soap") , currPNode.node, XPathConstants.NODE);
					if(wssoapNode != null)
						wsSoapRestNode = wssoapNode;
				
					Node wsRestNode = (Node) xPath.evaluate(new String(axmlPrefix + ":ws-rest") , currPNode.node, XPathConstants.NODE);
					if(wsRestNode != null)
						wsSoapRestNode = wsRestNode;

					endPoint = (String) xPath.evaluate("@endpoint" , wsSoapRestNode, XPathConstants.STRING);
					currNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":id") , currPNode.node, XPathConstants.STRING);
					origNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":nodeID") , currPNode.node, XPathConstants.STRING);
					origDocId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":docID") , currPNode.node, XPathConstants.STRING);
					origPeerId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":peerID") , currPNode.node, XPathConstants.STRING);
					
					// We calculate the hashcode of xml subtree that we are going to use at the specific service
					int hashCodeForService = calculateTheHashCodeForService(currPNode.node, axmlPrefix);
					
					PeerInformationModule peerInformationModuleInstance = PeerInformationModule.getInstance();
					// We get a peer and service codename that are related with that endpoint 
					List<String> peerNameAndServiceNameContainer = peerInformationModuleInstance.getPeerNameServiceName(endPoint);
					
					// We create a new service with the information that we extracted from the axml file
					System.out.println("peerNameAndServiceNameContainer:" + peerNameAndServiceNameContainer);
					System.out.println("endPoint:" + endPoint);
					System.out.println("!!!!!!!!!!!!!!!!!!!!!!!!!!");
					Service newServiceNode = new Service(new StringBuffer(origNodeId), new StringBuffer(origDocId) , new StringBuffer(origPeerId), false, new HashSet<String>(), new HashSet<String>(), new StringBuffer(peerNameAndServiceNameContainer.get(1)), new StringBuffer(peerNameAndServiceNameContainer.get(0)), new StringBuffer(), constructXPathFromPathList(currPNode.path), hashCodeForService, null, new ExtendedLinkedList<ADNode>());
					newServiceNode.setReferenceToADocument(returningADoc);
					newServiceNode.setCurrNodeID(new StringBuffer(currNodeId));

					String nameOfService = peerNameAndServiceNameContainer.get(1);
					if(nameOfService.equals("send") || nameOfService.equals("receive"))
					{
						Element secOrigNodeIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":address/" + algebraPrefix + ":originalID/" + algebraPrefix + ":nodeID" , currPNode.node, XPathConstants.NODE);
						StringBuffer secOrigNodeId = new StringBuffer(secOrigNodeIdCont.getTextContent());
						while(secOrigNodeId.charAt(0) == ' ' || secOrigNodeId.charAt(0) == '\n' || secOrigNodeId.charAt(0) == '\t')
							secOrigNodeId.deleteCharAt(0);
						while(secOrigNodeId.charAt(secOrigNodeId.length()-1) == ' ' || secOrigNodeId.charAt(secOrigNodeId.length()-1) == '\n' || secOrigNodeId.charAt(secOrigNodeId.length()-1) == '\t')
							secOrigNodeId.deleteCharAt(secOrigNodeId.length()-1);
						
						Element secOrigDocIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":address/" + algebraPrefix + ":originalID/" + algebraPrefix + ":docID" , currPNode.node, XPathConstants.NODE);
						StringBuffer secOrigDocId = new StringBuffer(secOrigDocIdCont.getTextContent());
						while(secOrigDocId.charAt(0) == ' ' || secOrigDocId.charAt(0) == '\n' || secOrigDocId.charAt(0) == '\t')
							secOrigDocId.deleteCharAt(0);
						while(secOrigDocId.charAt(secOrigDocId.length()-1) == ' ' || secOrigDocId.charAt(secOrigDocId.length()-1) == '\n' || secOrigDocId.charAt(secOrigDocId.length()-1) == '\t')
							secOrigDocId.deleteCharAt(secOrigDocId.length()-1);
						
						Element secOrigPeerIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":address/" + algebraPrefix + ":originalID/" + algebraPrefix + ":peerID" , currPNode.node, XPathConstants.NODE);
						StringBuffer secOrigPeerId = new StringBuffer(secOrigPeerIdCont.getTextContent());
						while(secOrigPeerId.charAt(0) == ' ' || secOrigPeerId.charAt(0) == '\n' || secOrigPeerId.charAt(0) == '\t')
							secOrigPeerId.deleteCharAt(0);
						while(secOrigPeerId.charAt(secOrigPeerId.length()-1) == ' ' || secOrigPeerId.charAt(secOrigPeerId.length()-1) == '\n' || secOrigPeerId.charAt(secOrigPeerId.length()-1) == '\t')
							secOrigPeerId.deleteCharAt(secOrigPeerId.length()-1);

						Element secEndPointCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":address/" + algebraPrefix + ":endpoint", currPNode.node, XPathConstants.NODE);
						StringBuffer secEndPoint = new StringBuffer(secEndPointCont.getTextContent());
						while(secEndPoint.charAt(0) == ' ' || secEndPoint.charAt(0) == '\n' || secEndPoint.charAt(0) == '\t')
							secEndPoint.deleteCharAt(0);
						while(secEndPoint.charAt(secEndPoint.length()-1) == ' ' || secEndPoint.charAt(secEndPoint.length()-1) == '\n' || secEndPoint.charAt(secEndPoint.length()-1) == '\t')
							secEndPoint.deleteCharAt(secEndPoint.length()-1);
						
						List<String> secPeerNameAndServiceNameContainer = peerInformationModuleInstance.getPeerNameServiceName(secEndPoint.toString());
						
						IdContainer sendDestinationAddressIdContainer = new IdContainer(secOrigNodeId.toString(), secOrigDocId.toString(), secOrigPeerId.toString());
						IdWithHostContainer sendDestinationAddressIdWithHostContainer = new IdWithHostContainer(sendDestinationAddressIdContainer, secPeerNameAndServiceNameContainer.get(0));
						SimpleNode sendDestinationAddress = new SimpleNode(new StringBuffer(), sendDestinationAddressIdWithHostContainer, new LinkedList<StringBuffer>());				
						newServiceNode.insertAsChildInBeginning(sendDestinationAddress);
						
						// Here we are going to search if there is a reference tag
						Element referenceNodeIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":reference", currPNode.node, XPathConstants.NODE);
						if(referenceNodeIdCont != null)
						{
							Element refOrigNodeIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":reference/" + algebraPrefix + ":originalID/" + algebraPrefix + ":nodeID" , currPNode.node, XPathConstants.NODE);
							StringBuffer refOrigNodeId = new StringBuffer(refOrigNodeIdCont.getTextContent());
							while(refOrigNodeId.charAt(0) == ' ' || refOrigNodeId.charAt(0) == '\n' || refOrigNodeId.charAt(0) == '\t')
								refOrigNodeId.deleteCharAt(0);
							while(refOrigNodeId.charAt(refOrigNodeId.length()-1) == ' ' || refOrigNodeId.charAt(refOrigNodeId.length()-1) == '\n' || refOrigNodeId.charAt(refOrigNodeId.length()-1) == '\t')
								refOrigNodeId.deleteCharAt(refOrigNodeId.length()-1);
							
							Element refOrigDocIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":reference/" + algebraPrefix + ":originalID/" + algebraPrefix + ":docID" , currPNode.node, XPathConstants.NODE);
							StringBuffer refOrigDocId = new StringBuffer(refOrigDocIdCont.getTextContent());
							while(refOrigDocId.charAt(0) == ' ' || refOrigDocId.charAt(0) == '\n' || refOrigDocId.charAt(0) == '\t')
								refOrigDocId.deleteCharAt(0);
							while(refOrigDocId.charAt(refOrigDocId.length()-1) == ' ' || refOrigDocId.charAt(refOrigDocId.length()-1) == '\n' || refOrigDocId.charAt(refOrigDocId.length()-1) == '\t')
								refOrigDocId.deleteCharAt(refOrigDocId.length()-1);
							
							Element refOrigPeerIdCont = (Element) xPath.evaluate(axmlPrefix + ":ws-soap/" + algebraPrefix + ":" + nameOfService + "/" + algebraPrefix + ":reference/" + algebraPrefix + ":originalID/" + algebraPrefix + ":peerID" , currPNode.node, XPathConstants.NODE);
							StringBuffer refOrigPeerId = new StringBuffer(refOrigPeerIdCont.getTextContent());
							while(refOrigPeerId.charAt(0) == ' ' || refOrigPeerId.charAt(0) == '\n' || refOrigPeerId.charAt(0) == '\t')
								refOrigPeerId.deleteCharAt(0);
							while(refOrigPeerId.charAt(refOrigPeerId.length()-1) == ' ' || refOrigPeerId.charAt(refOrigPeerId.length()-1) == '\n' || refOrigPeerId.charAt(refOrigPeerId.length()-1) == '\t')
								refOrigPeerId.deleteCharAt(refOrigPeerId.length()-1);

							IdContainer refAddressIdContainer = new IdContainer(refOrigNodeId.toString(), refOrigDocId.toString(), refOrigPeerId.toString());
							IdWithHostContainer refAddressIdWithHostContainer = new IdWithHostContainer(refAddressIdContainer, peerNameAndServiceNameContainer.get(0));
							SimpleNode refAddress = new SimpleNode(new StringBuffer(), refAddressIdWithHostContainer, new LinkedList<StringBuffer>());				
							newServiceNode.insertAsChildInBeginning(refAddress);
						}
					}
					
					axmlDomNodesToADocNodesMap.put(currPNode.node, newServiceNode);
					
				}
				else if(!nodesThatShouldNotBeRepresented.contains(currPNode.node.getNodeName()))
				{
					String origNodeId = null;
					String origDocId = null;
					String origPeerId = null;
					
					origNodeId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":nodeID") , currPNode.node, XPathConstants.STRING);
					origDocId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":docID") , currPNode.node, XPathConstants.STRING);
					origPeerId = (String) xPath.evaluate(new String("@" + axmlPrefix + ":peerID") , currPNode.node, XPathConstants.STRING);

					Document declarationDom = null;
					String nameOfParam = null;

					if(currPNode.node.getNodeName().equals(new String(streamSrvPrefix + ":query")))
					{
						DocumentBuilderFactory dfactory = DocumentBuilderFactory.newInstance();
						DocumentBuilder loader;
						
						try {
							loader = dfactory.newDocumentBuilder();
							declarationDom = loader.newDocument();
							Node rootNodeForDeclarationDom = declarationDom.importNode(currPNode.node, true);
							declarationDom.appendChild(rootNodeForDeclarationDom);
							
						} catch (ParserConfigurationException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
					}
					
					List<StringBuffer> currXPathList = new ArrayList<StringBuffer>();
					currXPathList.add(constructXPathFromPathList(currPNode.path));
					SimpleNode newSimpleNode = new SimpleNode(new StringBuffer(origNodeId), new StringBuffer(origDocId), new StringBuffer(origPeerId), new StringBuffer(currPNode.node.getNodeName()), new IdWithHostContainer(), currXPathList, null, new ExtendedLinkedList<ADNode>(), new String(), null);
					newSimpleNode.declaration = declarationDom;
					newSimpleNode.paramName = nameOfParam;
					newSimpleNode.setReferenceToADocument(returningADoc);

					axmlDomNodesToADocNodesMap.put(currPNode.node, newSimpleNode);
					
				}
			}
			
			// if currNode has children then it has to visit them
			if(currPNode.node.hasChildNodes())
			{
				NodeList tempNodeList = currPNode.node.getChildNodes();
				int childNumber = 0;
				for(int i=0;i<tempNodeList.getLength();i++)
				{
					Node tempChildNode = tempNodeList.item(i);
					if(tempChildNode.getNodeType() == Node.ELEMENT_NODE)
						childNumber++;
					List<Integer> tempChildPath = new ArrayList<Integer>(currPNode.path);
					tempChildPath.add(childNumber);
					pNodesToBeVisited.addLast(new NodesWithPath(tempChildNode, tempChildPath));
				}
			}
		}
		
		System.out.println(returningADoc.newScCurrNodeIdsToOld);
		System.out.println(returningADoc.oldScCurrNodeIdsToNew);
		
		// In this (second) loop we construct the ADNode tree having as reference the dom tree
		// that was given to us.
		LinkedList<Node> nodesToBeVisited = new LinkedList<Node>();
		nodesToBeVisited.add(origRoot);
		while(!nodesToBeVisited.isEmpty())
		{
			Node currNode = nodesToBeVisited.removeFirst();
			
			if(currNode.hasChildNodes())
			{
				ADNode currADNode = axmlDomNodesToADocNodesMap.get(currNode);
				ADNode ancestorOfCurrADNode = currADNode;
				if(ancestorOfCurrADNode == null)
				{
					Node ancestorOfCurrNode = currNode.getParentNode();
					while(ancestorOfCurrADNode == null)
					{
						ancestorOfCurrADNode = axmlDomNodesToADocNodesMap.get(ancestorOfCurrNode);
						ancestorOfCurrNode = ancestorOfCurrNode.getParentNode();
					}
				}

				for(Node tempChildNode = currNode.getFirstChild(); tempChildNode != currNode.getLastChild(); tempChildNode = tempChildNode.getNextSibling())
				{
					ADNode tempChildADNode = axmlDomNodesToADocNodesMap.get(tempChildNode);
					if(tempChildADNode == null)
						continue;
					if(ancestorOfCurrADNode instanceof Service && (((Service)ancestorOfCurrADNode).getServiceName().toString().equals(Parameters.SEND)||((Service)ancestorOfCurrADNode).getServiceName().toString().equals(Parameters.RECEIVE)))
						ancestorOfCurrADNode.insertAsChildInBeginning(tempChildADNode);
					else
						ancestorOfCurrADNode.insertAsChild(tempChildADNode);
				}
				ADNode tempLastChildADNode = axmlDomNodesToADocNodesMap.get(currNode.getLastChild());
				if(tempLastChildADNode != null)
				{
					if(ancestorOfCurrADNode instanceof Service && (((Service)ancestorOfCurrADNode).getServiceName().toString().equals(Parameters.SEND)||((Service)ancestorOfCurrADNode).getServiceName().toString().equals(Parameters.RECEIVE)))
						ancestorOfCurrADNode.insertAsChildInBeginning(tempLastChildADNode);
					else
						ancestorOfCurrADNode.insertAsChild(tempLastChildADNode);

				}
			}
			
			// if currNode has children then it has to visit them
			if(currNode.hasChildNodes())
			{
				NodeList tempNodeList = currNode.getChildNodes();
				for(int i=0;i<tempNodeList.getLength();i++)
				{
					Node tempChildNode = tempNodeList.item(i);
					nodesToBeVisited.addLast(tempChildNode);
				}
			}
		}
		
		if(axmlDomNodesToADocNodesMap.get(origRoot) != null)
			returningADoc.setRootOfADoc(axmlDomNodesToADocNodesMap.get(origRoot));
		else
		{
			System.err.println("<---------------------->");
		}
		
		if(!ADocument.isSubTreeConsistent(returningADoc.getRootOfADoc()))
			System.exit(0);
		
		return returningADoc;
	}
}
