package fr.inria.gemo.axml.optimax.io.output;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.adocument.EqualADocAtom;
import fr.inria.gemo.axml.optimax.adocument.RuleDescriptionContainer;
import fr.inria.gemo.axml.optimax.info.Statistics;

public class ADocumentRelationshipGraphModule {
	
	private StringBuffer createFullDataOfGraphvizFile()
	{
		float minCost = -1;
		
		// first we find the minimum cost
		Iterator<RuleDescriptionContainer> totalRulesAppliedIter = Statistics.totalRulesApplied.iterator();
//		while(totalRulesAppliedIter.hasNext())
//		{
//			RuleDescriptionContainer currRule = totalRulesAppliedIter.next();
//			if(minCost == -1 || currRule.oldCost < minCost)
//				minCost = currRule.oldCost;
//			if(minCost == -1 || currRule.newCost < minCost)
//				minCost = currRule.newCost;
//		}
		
		StringBuffer returningStringBuffer = new StringBuffer();
		returningStringBuffer.append("digraph \"DocumentRelationshipGraph\" {\n");
		
		totalRulesAppliedIter = Statistics.totalRulesApplied.iterator();
		while(totalRulesAppliedIter.hasNext())
		{
			RuleDescriptionContainer currRule = totalRulesAppliedIter.next();
			
//			if(currRule.oldCost == minCost)
//				returningStringBuffer.append(currRule.numberOfOldADoc + " [shape = doublecircle];\n");
//			else
//				returningStringBuffer.append(currRule.numberOfOldADoc + " [shape = circle];\n");
//			if(currRule.newCost == minCost)
//				returningStringBuffer.append(currRule.numberOfNewADoc + " [shape = doublecircle];\n");
//			else
//				returningStringBuffer.append(currRule.numberOfOldADoc + " [shape = circle];\n");
			
			// We change the colour of the arrow according to the plan cost change
			if(currRule.increasesCost)
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + currRule.numberOfNewADoc + " [color=red label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
			else if(currRule.decreasesCost)
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + currRule.numberOfNewADoc + " [color=darkgreen label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
			else
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + currRule.numberOfNewADoc + " [label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
		}
		Iterator<EqualADocAtom> totalADocEqualities = Statistics.aDocEqualities.iterator();
		while(totalADocEqualities.hasNext())
		{
			EqualADocAtom currEquality = totalADocEqualities.next();
			returningStringBuffer.append(currEquality.numberOfADoc1 + " -> " + currEquality.numberOfADoc2 + " [style=dashed];\n");
		}
		
		returningStringBuffer.append("}\n");
		
		return returningStringBuffer;
	}
	
	private StringBuffer createReducedDataOfGraphvizFile()
	{
		StringBuffer returningStringBuffer = new StringBuffer();

		// we insert the equalities of docs in a map. We know that the
		// first value of a record in totalADocEqualities is the dropped document
		Map<Integer,Integer> droppedDocToKeptDocs = new HashMap<Integer,Integer>();
		Iterator<EqualADocAtom> totalADocEqualities = Statistics.aDocEqualities.iterator();
		while(totalADocEqualities.hasNext())
		{
			EqualADocAtom currEquality = totalADocEqualities.next();
			droppedDocToKeptDocs.put(currEquality.numberOfADoc1, currEquality.numberOfADoc2);
		}
		
		returningStringBuffer.append("digraph \"DocumentRelationshipGraph\" { node [shape = circle]; node [style=filled];\n");
		Iterator<RuleDescriptionContainer> totalRulesAppliedIter = Statistics.totalRulesApplied.iterator();
		while(totalRulesAppliedIter.hasNext())
		{
			RuleDescriptionContainer currRule = totalRulesAppliedIter.next();
			// We change the colour of the arrow according to the plan cost change
			int numberOfNewADoc = currRule.numberOfNewADoc;
			if(droppedDocToKeptDocs.containsKey(numberOfNewADoc))
				numberOfNewADoc = droppedDocToKeptDocs.get(numberOfNewADoc);
			
			if(currRule.increasesCost)
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + numberOfNewADoc + " [color=red label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
			else if(currRule.decreasesCost)
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + numberOfNewADoc + " [color=darkgreen label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
			else
				returningStringBuffer.append(currRule.numberOfOldADoc + " -> " + numberOfNewADoc + " [label = \"" + currRule.nameOfRule + "(" + currRule.arg1 + "," + currRule.arg2 + ")\" ];\n" );
		}
		
		returningStringBuffer.append("}\n");
		
		return returningStringBuffer;
	}
	
	public void createRelationshipGraph()
	{
		String fileName = "optimaxDocumentRelationshipGraph";
		
		if(Parameters.getProperty("drawRel").equals(Parameters.FALSE))
			return;
		
		if(Parameters.getProperty("drawRel").equals(Parameters.CONCISE_GRAPH))
			FileOutputModule.saveDataFromStringBufferToFile(createReducedDataOfGraphvizFile(), new String(fileName + ".txt"));
		
		if(Parameters.getProperty("drawRel").equals(Parameters.VERBOSE_GRAPH))
			FileOutputModule.saveDataFromStringBufferToFile(createFullDataOfGraphvizFile(), new String(fileName + ".txt"));
		
		GraphvizModule.runGraphviz(fileName);
//		Deleting.deleteSpecificFile(new String(fileName + ".txt"));
	}
}
