package fr.inria.gemo.axml.optimax.io.output;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import fr.inria.gemo.axml.optimax.PlanRepository;
import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;

public class ConciseADocumentDrawingModule {

	  // from now on, functions for a more compact GraphViz output, 
	// more suitable for a demo
	
	/**
	 * Structure:
	 * on the graphViz-assigned ID put
	 * - when the node is first encountered, an empty String
	 * - when its child is visited, the String encapsulating the actual ID
	 * (starting with *) and the 
	 */
	public static HashMap<Integer, String> sendNodeIDs;
	public static HashMap<Integer, String> receiveNodeIDs;
	
	public static StringBuffer createConciseGraphVizOutput(ADocument aDoc){
		sendNodeIDs = new HashMap<Integer, String>();
		receiveNodeIDs = new HashMap<Integer, String>();
		StringBuffer sb = new StringBuffer();
		sb.append("digraph doc" + aDoc.getADocNumber() + " {\n");
		toDotString(aDoc.getRootOfADoc(), sb, false, -1, 0);
		addSendReceiveLinks(aDoc.getRootOfADoc(), sb);
		sb.append("}\n");
		return sb;
	}
	
	private static void addSendReceiveLinks(ADNode rootOfADoc, StringBuffer sb) {
		Iterator<Integer> it = sendNodeIDs.keySet().iterator();
		while (it.hasNext()){
			Integer iSend = it.next();
			String sID = sendNodeIDs.get(iSend);
			Iterator<Integer>it2 = receiveNodeIDs.keySet().iterator();
			while (it2.hasNext()){
				Integer iReceive = it2.next();
				String rID = receiveNodeIDs.get(iReceive);
				if (rID.compareTo(sID) == 0){
					sb.append(iSend + " -> " + iReceive + " [color=\"forestgreen\" style=\"dotted\"];\n");
				}
			}
		}
	}


	private static final int toDotString(ADNode n, StringBuffer sb, boolean underReceive, int parentID, int nextAvailable) {
		int myOwn = nextAvailable;
		int futureAvailable = nextAvailable;
				
		String name = n.toString();
		String shortName = name.substring(0, name.indexOf('('));
		
		boolean thisNodeVisible = false;
		boolean thisNodeIsReceive = false;
		//System.out.println("\nEntering toDotString on " + shortName + " underReceive: " +
		//		underReceive);
		
		String colorString = "dimgrey";
		
		if (n instanceof Service){
			// Service node
			thisNodeVisible = true;
			
			// send or receive?
			if (shortName.startsWith("receive") || shortName.startsWith("send")){
				if (!underReceive){
					colorString = "forestgreen";
				}
				sb.append(nextAvailable + " [label=\"" + ((Service)n).serviceName + 
						"@" + ((Service)n).executionPeer + 
				"\" color=\""+ colorString + "\" fontcolor=\"" + 
				colorString + "\" ];\n");
				if (shortName.startsWith("receive")){
					thisNodeIsReceive = true;
					String recvCurrentID = new String(n.getCurrNodeID()+
							"." + n.getCurrDocID() + "." + n.getCurrPeerID());
					//System.out.println(">>> receive: " + myOwn + " " + recvCurrentID);
					receiveNodeIDs.put(new Integer(myOwn), recvCurrentID);
				}
				if (shortName.startsWith("send")){
					//System.out.println(">>> send: " + myOwn);
					sendNodeIDs.put(new Integer(myOwn), "");
				}
			}
			else{
				// newNode?
				if(shortName.startsWith("newNode")){
					if (!underReceive){
						colorString = "blue3";
					}
				}
				else{
					if (!underReceive){
						colorString="red";
					}
				}
				sb.append(nextAvailable + " [label=\"" + ((Service)n).serviceName + 
						"@" + ((Service)n).executionPeer + 
						"\" color=\"" + colorString + "\" fontcolor=\"" +
						colorString + "\" ];\n");
			}
		}
		
		else{ // NOT  A SERVICE NODE
			boolean isAnAddress = false;
			
		    SimpleNode sn = (SimpleNode)n;
		    if (sn.getNodeData().length()==0 && !
		    		sn.getRemoteOriginalIdAndHostingPeer().isEmpty()){
		    	// this node is an address 
		    	isAnAddress = true;
				shortName = sn.getRemoteOriginalIdAndHostingPeer().id.toString();
		      	
		    	String usefulName = shortName;
				if (sendNodeIDs.get(new Integer(parentID))!=null){
					// my daddy was a send
					//System.out.println(">>> send: " + parentID + " " + usefulName);
					sendNodeIDs.put(new Integer(parentID), usefulName);
				}
				thisNodeVisible = false;
			}
			if (shortName.compareTo("axml:ws-soap")!=0 &&
					shortName.compareTo("svc:executeQuery")!= 0 &&
					shortName.compareTo("query:param")!=0 &&
					(!isAnAddress)){
				thisNodeVisible = true;
				sb.append(nextAvailable + " [label=\"" + shortName);
				if (parentID < 0){
					sb.append("@"+n.getCurrPeerID());
				}
				sb.append("\"];\n");
			}
		}
			
		if (thisNodeVisible){
			if (parentID >= 0){
				sb.append(parentID + " -> " + myOwn);
				if (underReceive){
					sb.append(" [color=\"dimgrey\"]");
				}						
				sb.append(" [dir=none];\n");
			}
			futureAvailable ++;
			//System.out.println("For " + shortName + " futureAvailable is " + futureAvailable);
		}

		List<ADNode> children = n.getChildren();
		Iterator<ADNode> it = children.iterator();
		while(it.hasNext()){
			ADNode child = it.next();
			if (thisNodeVisible){
				futureAvailable = toDotString(child, sb, (underReceive || thisNodeIsReceive), myOwn, futureAvailable);
			}
			else{
				futureAvailable = toDotString(child, sb, (underReceive || thisNodeIsReceive), parentID, futureAvailable);
			}	
		}
		return futureAvailable;
	}
	
	public static void createOutputImage(ADocument givenADocument, String fileName)
	{
//		deletePreviousTxtAndJpgFiles();
		FileOutputModule.saveDataFromStringBufferToFile(createConciseGraphVizOutput(givenADocument), new String(fileName + ".txt"));
		GraphvizModule.runGraphviz(fileName);
	}
}
