package fr.inria.gemo.axml.optimax.io.output;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.PlanRepository;
import fr.inria.gemo.axml.optimax.SearchAlgorithm;
import fr.inria.gemo.axml.optimax.helping.Utilities;
import fr.inria.gemo.axml.optimax.info.Statistics;

public class CostPlotModule {
	
	private StringBuffer constructScriptDataForGnuPlot(String mainPartOfTheNameOfFiles)
	{
		float maxSimpleValue = -1;
		float maxRealValue = -1;
		
		Iterator<Float> simpleIter = Statistics.listWithCurrSimpleCostOfAlreadyVisitedPlans.iterator();
		while(simpleIter.hasNext())
		{
			Float currValue = simpleIter.next();
			if(maxSimpleValue < currValue)
				maxSimpleValue = currValue;
		}

		Iterator<Float> realIter = Statistics.listWithCurrServiceAndWeightedNetworkCostOfAlreadyVisitedPlans.iterator();
		while(realIter.hasNext())
		{
			Float currValue = realIter.next();
			if(maxRealValue < currValue)
				maxRealValue = currValue;
		}
		
		maxSimpleValue++;
		maxRealValue++;

		StringBuffer dataToBeReturned = new StringBuffer();
		dataToBeReturned.append("set data style points;\n");
		dataToBeReturned.append("set grid;\n");
		dataToBeReturned.append("set terminal jpeg;\n");
		dataToBeReturned.append("set parametric;\n");
		dataToBeReturned.append("set output \"" + Parameters.OUTPUT_DIR + mainPartOfTheNameOfFiles + ".jpg" + "\";\n");
		dataToBeReturned.append("set multiplot;\n");
		dataToBeReturned.append("set size 1,0.50;\n");
		
		for(int i=0; i<Statistics.startOfSteps.size(); i++)
		{
			dataToBeReturned.append("const" + i + "=" + Statistics.startOfSteps.get(i) + ";\n");
		}
		
		dataToBeReturned.append("set origin 0.0,0.0;\n");
		dataToBeReturned.append("set xlabel \"number of explored plans\";\n");
		dataToBeReturned.append("set ylabel \"Service and weighted network cost\";\n");
		dataToBeReturned.append("set trange [0:" + maxRealValue + "];\n");
		dataToBeReturned.append("plot " + "\"" + Parameters.OUTPUT_DIR + mainPartOfTheNameOfFiles + "DataWithMinRealCost" + ".dat" + "\" title \"min plan cost\"" + " , " + "\"" + Parameters.OUTPUT_DIR + mainPartOfTheNameOfFiles + "DataWithCurrRealCost" + ".dat" + "\" title \"curr plan cost\"");
		for(int i=0; i<Statistics.startOfSteps.size(); i++)
		{
			dataToBeReturned.append(", const" + i + ",t  title \"\"");
		}
		dataToBeReturned.append(";\n");
		
		dataToBeReturned.append("set origin 0.0,0.50;\n");
		dataToBeReturned.append("set xlabel \"number of explored plans\";\n");
		dataToBeReturned.append("set ylabel \"Simple cost\";\n");
		dataToBeReturned.append("set trange [0:" + maxSimpleValue + "];\n");
		dataToBeReturned.append("plot " + "\"" + Parameters.OUTPUT_DIR + mainPartOfTheNameOfFiles + "DataWithMinSimpleCost" + ".dat" + "\" title \"min plan cost\"" + " , " + "\"" + Parameters.OUTPUT_DIR + mainPartOfTheNameOfFiles + "DataWithCurrSimpleCost" + ".dat" + "\" title \"curr plan cost\"");
		for(int i=0; i<Statistics.startOfSteps.size(); i++)
		{
			dataToBeReturned.append(", const" + i + ",t  title \"\"");
		}
		dataToBeReturned.append(";\n");
		
		dataToBeReturned.append("unset multiplot;\n");
		
		return dataToBeReturned;
	}
	
	private StringBuffer constructDataFileDataForGnuPlot(boolean minimum, boolean simpleCost)
	{
		StringBuffer dataToBeReturned = new StringBuffer();
		
		List<Float> listWithData = null;

		if(minimum && simpleCost)
			listWithData = Statistics.listWithMinimumSimpleCostOfAlreadyVisitedPlans;
		else if(!minimum && simpleCost)
			listWithData = Statistics.listWithCurrSimpleCostOfAlreadyVisitedPlans;
		else if(!minimum && !simpleCost)
			listWithData = Statistics.listWithCurrServiceAndWeightedNetworkCostOfAlreadyVisitedPlans;
		else if(minimum && !simpleCost)
			listWithData = Statistics.listWithMinimumServiceAndWeightedNetworkCostOfAlreadyVisitedPlans;
		
		Iterator<Float> listWithDataIter = listWithData.iterator();
		int counter = 0;
		while(listWithDataIter.hasNext())
		{
			counter++;
			dataToBeReturned.append(Integer.toString(counter) + " " + Float.toString(listWithDataIter.next()) + "\n");
		}
		
		return dataToBeReturned;
	}
	
	private void constructDataAndScriptFiles(String mainPartOfTheNameOfFiles)
	{
		// construct the data of the script files
		StringBuffer scriptFile  = constructScriptDataForGnuPlot(mainPartOfTheNameOfFiles);

		// store the script files at the disk
		FileOutputModule.saveDataFromStringBufferToFile(scriptFile, new String(mainPartOfTheNameOfFiles + "Script" + ".gnu"));
		
		// construct the data for the data files
		StringBuffer dataFileMS  = constructDataFileDataForGnuPlot(true, true);
		StringBuffer dataFileMR  = constructDataFileDataForGnuPlot(true, false);
		StringBuffer dataFileCS  = constructDataFileDataForGnuPlot(false, true);
		StringBuffer dataFileCR  = constructDataFileDataForGnuPlot(false, false);

		// store the data files at the disk
		FileOutputModule.saveDataFromStringBufferToFile(dataFileMS, new String(mainPartOfTheNameOfFiles + "DataWithMinSimpleCost" + ".dat"));
		FileOutputModule.saveDataFromStringBufferToFile(dataFileMR, new String(mainPartOfTheNameOfFiles + "DataWithMinRealCost" + ".dat"));
		FileOutputModule.saveDataFromStringBufferToFile(dataFileCS, new String(mainPartOfTheNameOfFiles + "DataWithCurrSimpleCost" + ".dat"));
		FileOutputModule.saveDataFromStringBufferToFile(dataFileCR, new String(mainPartOfTheNameOfFiles + "DataWithCurrRealCost" + ".dat"));
	}
	
	private void deleteGnuDataAndScriptFiles()
	{
		File path = new File(Parameters.OUTPUT_DIR);
		if(path.isDirectory())
		{
			String[] dirContents = path.list();
			for(int i=0; i<dirContents.length; i++)
			{
				if(dirContents[i].endsWith(".dat") || dirContents[i].endsWith(".gnu"))
				{
					File specificFile = new File(Parameters.OUTPUT_DIR + dirContents[i]);
					specificFile.delete();
				}
			}
		}
	}
	
	public void createGnuPlotGraphs()
	{
		// If it is not specified in the command line to draw cost graphs
		// the we do not print
		if(Parameters.getProperty("drawCost").equals(Parameters.FALSE))
			return;
		
		String mainPartOfTheNameOfFiles = "optimaxGnuPlot";
		
		// We construct the needed files to run gnuplot
		constructDataAndScriptFiles(mainPartOfTheNameOfFiles);
		
		// We run gnuplot
		runGnuPlot(new String(mainPartOfTheNameOfFiles + "Script" + ".gnu"));
		
		// We erase the unneeded files
//		deleteGnuDataAndScriptFiles();
		
	}
	
	private static void runGnuPlot(String scriptFileName) 
	{
	     List<String> command = new ArrayList<String>();
	     command.add("gnuplot");
	     command.add(Parameters.OUTPUT_DIR + scriptFileName);
	     
	     ProcessBuilder builder = new ProcessBuilder(command);
	     File currDir = new File(".");
	     builder.directory(currDir);

	     try 
	     {
			@SuppressWarnings("unused")
			final Process process = builder.start();
			process.waitFor();
	     } 
	     catch (IOException e) 
	     {
			// TODO Auto-generated catch block
			e.printStackTrace();
	     } catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
}
