package fr.inria.gemo.axml.optimax.rules;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.IdContainer;
import fr.inria.gemo.axml.optimax.adocument.IdWithHostContainer;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.io.output.GraphvizModule;

public class CSE {
	// this method will try tell us the nodes which should be deleted. It takes in to account that
	// the givenNode subtree will be deleted and this should not affect at all the nodes of 
	// nodesMarkedForDeletion.
	private boolean erase(ADocument givenADoc, ADNode givenNode, List<ADNode> nodesThatShouldNotBeHarmed, ExtendedLinkedList<ADNode> nodesMarkedForDeletion)
	{
		if(givenADoc.getADocNumber() == 16 && givenNode.getCurrNodeID().toString().equals("296"))
			System.out.println("xaxa");
		
		// find all the nodes of the subtree at which givenNode
		// is a root (givenNode will be included in this list)
		ExtendedLinkedList<ADNode> givenNodeOrGivenNodeDescendants = new ExtendedLinkedList<ADNode>();
		
		givenNodeOrGivenNodeDescendants.addAll(UsefulMethods.getAllTheNodesOfSubTree(givenNode));
		
		// we check if the givenNodeOrGivenNodeDescendants contains any node 
		// of nodesThatShouldNotBeHarmed
		Iterator<ADNode> nodesThatShouldNotBeHarmedIter = nodesThatShouldNotBeHarmed.iterator();
		while(nodesThatShouldNotBeHarmedIter.hasNext())
		{
			if(givenNodeOrGivenNodeDescendants.refContains(nodesThatShouldNotBeHarmedIter.next()))
				return false;
		}
		
		// Now that we are sure that givenNodeOrGivenNodeDescendants does not contain
		// any nodes of nodesThatShouldNotBeHarmed we can put it in nodesMarkedForDeletion
		// By doing this check every time we reassure that nodesMarkedForDeletion does not
		// contain any node of nodesThatShouldNotBeHarmed. 
		// ps: the first nodesMarkedForDeletion list should be empty.
		
		// we put that the nodes of givenNodeOrGivenNodeDescendants in the set
		// of the nodes that are marked for deletion
		nodesMarkedForDeletion.addAll(givenNodeOrGivenNodeDescendants);
	
		// now we are going to search if there is an incopatibility or if
		// we have to add more nodes to these two sets.
		
		// foreach node in givenNodeOrGivenNodeDescendants
		Iterator<ADNode> givenNodeOrGivenNodeDescendantsIter = givenNodeOrGivenNodeDescendants.iterator();
		while(givenNodeOrGivenNodeDescendantsIter.hasNext())
		{
			// we mention currADNode as m in our pseudo-algorithm
			ADNode currADNode = givenNodeOrGivenNodeDescendantsIter.next();
			
			// we find all the addresses which point to currADNode
			List<SimpleNode> addressesThatPointToCurrADNode = currADNode.getOriginalIdWithHostAddressContainersThatPointToThisNode();
			// foreach such address
			Iterator<SimpleNode> addressesThatPointToCurrADNodeIter = addressesThatPointToCurrADNode.iterator();
			while(addressesThatPointToCurrADNodeIter.hasNext())
			{
				SimpleNode currAddress = addressesThatPointToCurrADNodeIter.next();
				if(givenNodeOrGivenNodeDescendants.refContains(currAddress) || nodesMarkedForDeletion.refContains(currAddress))
				{
					// nothing more to be erased
				}
				else // is outside givenNode subtree or was not marked for deletion
				{
					// if currAddress is the where-to (second) child of a send
					Service fatherOfCurrAddress = (Service)currAddress.getFather();
					if(fatherOfCurrAddress.getServiceName().toString().equals("send"))
					{
						if(fatherOfCurrAddress.getChildren().get(fatherOfCurrAddress.getChildren().size()-1) == currAddress)
						{
							// currADNode (m) is a receive

							if(erase(givenADoc, fatherOfCurrAddress, nodesThatShouldNotBeHarmed, nodesMarkedForDeletion) == false)
								return false;
						}
						else
							return false;
					}
				}
			}
		}
		return true;
	}
	
	public List<ADocument> commonSubExpressionElimination(ADocument givenADoc)
	{
		List<ADocument> returningADocs = new ArrayList<ADocument>();
		
		if(givenADoc.getADocNumber() == 10)
		{
			System.out.println("xaxa");
		}
		
		// We search for all the services in the document
		List<Service> servicesOfTheTree = ADocument.findServicesOfADoc(givenADoc);
		
		for(Iterator<Service> servicesOfTheTreeIter = servicesOfTheTree.iterator(); servicesOfTheTreeIter.hasNext();)
		{
			// with the loop bellow we keep all the scs that are either simple
			// either they are receive scs that do not have any other receive
			// above
			Service currService = servicesOfTheTreeIter.next();
			if(currService.getIsSpecialService())
			{
				if(currService.getServiceName().toString().equals("receive"))
				{
					ADNode ancestorOfServicesOfTheTreeGetI = currService.getFather();
					while(ancestorOfServicesOfTheTreeGetI != null && !(ancestorOfServicesOfTheTreeGetI instanceof Service))
						ancestorOfServicesOfTheTreeGetI = ancestorOfServicesOfTheTreeGetI.getFather();
					
					// if we reached the root
					if(ancestorOfServicesOfTheTreeGetI == null)
						continue;
					else if(ancestorOfServicesOfTheTreeGetI instanceof Service)
					{
						Service serviceancestorOfServicesOfTheTreeGetI = (Service) ancestorOfServicesOfTheTreeGetI;
						if(!serviceancestorOfServicesOfTheTreeGetI.getIsSpecialService())
							continue;
					}
				}
				
				servicesOfTheTreeIter.remove();
			}
			else
				continue;
		}
		
		// For each service call in the list we search for all other equivalent service calls
		while(!servicesOfTheTree.isEmpty())
		{
			Service currService = servicesOfTheTree.remove(0);
			List<Service> servicesThatWeCanApplyCSE = new ArrayList<Service>();
			servicesThatWeCanApplyCSE.add(currService);
						
			for(int i=0; i<servicesOfTheTree.size();)
			{
				Service innerService = servicesOfTheTree.get(i);
				if(currService.equivalent(innerService))
				{
					servicesThatWeCanApplyCSE.add(innerService);
					servicesOfTheTree.remove(i);
				}
				else
					i++;
			}
			
			// servicesThatWeCanApplyCSE contains the biggest set of services
			// of a specific type that we can apply CSE to them.
			if(servicesThatWeCanApplyCSE.size() > 1)
			{
				// now we going to search for all possible subsets
				// of servicesThatWeCanApplyCSE
				List<List<Service>> allPossibleListsAndSubListsWith2Elements = findAllPossibleListsAndSubListsOfSize2(servicesThatWeCanApplyCSE);
				Iterator<List<Service>> allPossibleListsAndSubListsWith2ElementsIter = allPossibleListsAndSubListsWith2Elements.iterator();
				while(allPossibleListsAndSubListsWith2ElementsIter.hasNext())
				{
					List<Service> currListOrSubListWith2Elements = allPossibleListsAndSubListsWith2ElementsIter.next();
					ADocument newADoc = checkApplyCSEandMakeNeededUpdates(givenADoc, currListOrSubListWith2Elements.get(0), currListOrSubListWith2Elements.get(1));
					if(newADoc != null)
						returningADocs.add(newADoc);
				}
			}
		}
		
		return returningADocs;
	}

	private ADocument checkApplyCSEandMakeNeededUpdates(ADocument givenADoc, Service supplierService, Service toBeEliminatedService)
	{
		ExtendedLinkedList<ADNode> nodesThatShouldNotBeHarmedOnGivenDoc = dataDepends(givenADoc, supplierService);
		
//		// we check if we the supplier service is already "suppling data" to
//		// the a service call at which toBeEliminatedService data depends
//		if(nodesThatShouldNotBeHarmedOnGivenDoc.refContains(supplierService))
//			return null;
		
		// We check if a conflict will occur by the elimination of toBeEliminatedService
		ExtendedLinkedList<ADNode> nodesMarkedForDeletionGivenADoc = new ExtendedLinkedList<ADNode>();
		if(erase(givenADoc, toBeEliminatedService, nodesThatShouldNotBeHarmedOnGivenDoc, nodesMarkedForDeletionGivenADoc) == false)
			return null;
		
		// If we do not want complete CSE we make also the following check
		if(Parameters.getProperty("completeCSE").equals(Parameters.FALSE))
		{
			// we want to compute how many simple scs are in the nodesMarkedForDeletionGivenADoc list
			// so we delete the not simple ones
			Iterator<ADNode> nodesMarkedForDeletionGivenADocIter = nodesMarkedForDeletionGivenADoc.iterator();
			while(nodesMarkedForDeletionGivenADocIter.hasNext())
			{
				ADNode currADNode = nodesMarkedForDeletionGivenADocIter.next();
				if(currADNode instanceof SimpleNode)
					nodesMarkedForDeletionGivenADocIter.remove();
				else if(currADNode instanceof Service)
				{
					Service serviceCurrADNode = (Service) currADNode;
					if(serviceCurrADNode.getIsSpecialService())
						nodesMarkedForDeletionGivenADocIter.remove();
				}
			}
			// and we calculate the nodesMarkedForDeletionGivenADoc size
			// If the size is 0 then we abort CSE
			if(nodesMarkedForDeletionGivenADoc.size() == 0)
				return null;
		}
		
		ADocument newADoc = new ADocument(givenADoc);
		
		Service supplierServiceAtNewADoc = (Service) ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, supplierService);
		Service toBeEliminatedServiceAtNewADoc = (Service) ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, toBeEliminatedService);
		
		ExtendedLinkedList<ADNode> nodesThatShouldBeDeletedOnNewDoc = new ExtendedLinkedList<ADNode>();
		List<ADNode> nodesThatShouldNotBeHarmedOnNewDoc = dataDepends(newADoc, supplierServiceAtNewADoc);
		
		erase(newADoc, toBeEliminatedServiceAtNewADoc, nodesThatShouldNotBeHarmedOnNewDoc, nodesThatShouldBeDeletedOnNewDoc);
		
		List<Service> addedSendAndReceiveContainer = applySubExpressionEliminationToGivenADocAndDeleteNotNeededNodes(newADoc, supplierServiceAtNewADoc, toBeEliminatedServiceAtNewADoc, nodesThatShouldBeDeletedOnNewDoc);
		
//		if(givenADoc.getADocNumber() == 28 && newADoc.getADocNumber() == 114)
//		{
//			GraphvizModule.createOutputImage(newADoc, "xaxaxa2");
//			System.out.println("xaxa");
//		}
		
		UsefulMethods.ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(newADoc, addedSendAndReceiveContainer.get(0));
		UsefulMethods.ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(newADoc, addedSendAndReceiveContainer.get(1));
		// must update all the receives for the new receive & for the new send
		
		// calculate network cost from scratch
		newADoc.calculateNetworkCost();
		
		// We add at the newADoc the rule that we applied
		boolean decreaseCost = false;
		boolean increaseCost = false;

		if(givenADoc.getSimpleNetworkCost() < newADoc.getSimpleNetworkCost())
			increaseCost = true;

		if(givenADoc.getSimpleNetworkCost() > newADoc.getSimpleNetworkCost())
			decreaseCost = true;
		
		newADoc.insertAppliedRuleToADoc("cse", givenADoc.getADocNumber(), supplierService.getCurrNodeID().toString(), toBeEliminatedService.getCurrNodeID().toString(), newADoc.getADocNumber(), increaseCost, decreaseCost, givenADoc.getSimpleNetworkCost(), newADoc.getSimpleNetworkCost());
		
		if(!newADoc.isADocumentConsistent())
		{
			System.err.println("checkApplyCSEandMakeNeededUpdates: Produced inconsistent document");
			System.exit(0);
		}
		
		return newADoc;
	}
	
	
	// This method will scan the subtree rooted by givenNode and it will return a
	// List with references to all the subtree nodes (except newNodes and sends that
	// are not reachable from the receives that are encountered. When it finds a receive sc, 
	// it follows the receiving pointer to find the respective send. It continues
	// the searching at send's arguments. The last is being done recursivelly
	private ExtendedLinkedList<ADNode> dataDepends(ADocument givenADoc, ADNode givenNode)
	{
		if(givenADoc.getADocNumber() == 16)
		{
			System.out.println("xaxa");
		}
		
		ExtendedLinkedList<ADNode> returningNodes = new ExtendedLinkedList<ADNode>();
		if(givenNode instanceof Service)
		{
			Service serviceGivenNode = (Service) givenNode;
			if(serviceGivenNode.getServiceName().toString().equals("receive"))
			{
				// we add receive and its descendants in returningNodes
				returningNodes.addAll(UsefulMethods.getAllTheNodesOfSubTree(serviceGivenNode));
				
				// we will find the respective send of serviceGivenNode
				
				// we find send's address (original Id with hosting peer)
				SimpleNode sendsAddressContainer = (SimpleNode) serviceGivenNode.getChildren().get(serviceGivenNode.getChildren().size() - 1);
				IdWithHostContainer sendsAddress = sendsAddressContainer.getRemoteOriginalIdAndHostingPeer();
				
				// we locate the send
				Service referenceToSend = (Service) givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(sendsAddress.id);

				// we add it in returningNodes
				returningNodes.add(referenceToSend);
				// we add its second (pointer to recieve) child in the returning nodes
				returningNodes.add(referenceToSend.getChildren().get(referenceToSend.getChildren().size() - 1));
				
				// we check if the what part of the send is a pointer
				ADNode firstChildOfSend = referenceToSend.getChildren().get(0);
				
				ADNode newFirstChildOfSend = firstChildOfSend;
				if(firstChildOfSend instanceof SimpleNode)
				{
					SimpleNode simpleNodeFirstChildOfSend = (SimpleNode) firstChildOfSend;
					// the send's what part is pointing to another node
					if(simpleNodeFirstChildOfSend.getNodeData().length() == 0 && !simpleNodeFirstChildOfSend.getRemoteOriginalIdAndHostingPeer().isEmpty())
					{
						// we add the pointer in the returning nodes
						returningNodes.add(referenceToSend.getChildren().get(0));
						
						// we locate the other service and we store its reference at newFirstChildOfSend
						IdWithHostContainer pointedServiceAddress = simpleNodeFirstChildOfSend.getRemoteOriginalIdAndHostingPeer();
						newFirstChildOfSend = givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(pointedServiceAddress.id);
					}
				}
				
				returningNodes.addAll(dataDepends(givenADoc, newFirstChildOfSend));
			}
			// send and newNode do not bring data directly (only receive does).
			// So we care only about the receives and the sends which are connected
			// with these receives
			else if(serviceGivenNode.getServiceName().toString().equals("send"));
			else if(serviceGivenNode.getServiceName().toString().equals("newNode"));

			// if it is a simple service
			else
			{
				returningNodes.add(givenNode);

				for(int i=0; i<givenNode.getChildren().size(); i++)
					returningNodes.addAll(dataDepends(givenADoc, givenNode.getChildren().get(i)));
			}
		}
		// if it is a simpleNode
		else
		{
			returningNodes.add(givenNode);

			for(int i=0; i<givenNode.getChildren().size(); i++)
				returningNodes.addAll(dataDepends(givenADoc, givenNode.getChildren().get(i)));
		}

		return returningNodes;
	}

	
	private List<Service> applySubExpressionEliminationToGivenADocAndDeleteNotNeededNodes(ADocument givenADoc, Service supplierService, Service toBeEliminatedService, List<ADNode> nodesThatShouldBeDeleted)
	{
		LinkedList<Service> addedSendAndReceiveConatiner = new LinkedList<Service>();
		
		// Creating the receive service that is going to receive the results
		// which would be transmitted by the local send
		Service receiveService = new Service();
		receiveService.setServiceName(new StringBuffer("receive"));
		toBeEliminatedService.getFather().insertAsChildNextToGivenChild(receiveService, toBeEliminatedService);
		receiveService.setIsSpecialService(true);
		receiveService.setExecutionPeer(new StringBuffer(ADocument.returnCallingPeerOfSpecificWebService(toBeEliminatedService, givenADoc)));
		
//		GraphvizModule.createOutputImage(givenADoc, "OLE");
		
		// We add the address of the supplier service in order to be possible for the 
		// receive to see who is the supplier
		IdContainer tempIdContainer4 = new IdContainer(supplierService.getOrigNodeID().toString(), supplierService.getOrigDocID().toString(), supplierService.getOrigPeerID().toString());
		String callingPeerOfSupplierService = UsefulMethods.returnCallingPeerOfGivenService(supplierService.getReferenceToADocument(), supplierService);
		IdWithHostContainer tempIdWithHostContainer4 = new IdWithHostContainer(tempIdContainer4, callingPeerOfSupplierService);
		SimpleNode supplierServiceNewADocAddr2 = new SimpleNode(new StringBuffer(), tempIdWithHostContainer4, new LinkedList<StringBuffer>());
		receiveService.insertAsChild(supplierServiceNewADocAddr2);
				
		// We update the mapping tables in order the currNodeId of the added receive to map to the 
		// old currNodeId of the serviceToEliminateAtNewADoc  
		String serviceToBeEliminatedOldCurrNodeId = givenADoc.newScCurrNodeIdsToOld.get(toBeEliminatedService.getCurrNodeID().toString());
		if(serviceToBeEliminatedOldCurrNodeId != null)
		{
			givenADoc.newScCurrNodeIdsToOld.remove(toBeEliminatedService.getCurrNodeID().toString());
			givenADoc.oldScCurrNodeIdsToNew.remove(serviceToBeEliminatedOldCurrNodeId);
			
			givenADoc.newScCurrNodeIdsToOld.put(receiveService.getCurrNodeID().toString(), serviceToBeEliminatedOldCurrNodeId);
			givenADoc.oldScCurrNodeIdsToNew.put(serviceToBeEliminatedOldCurrNodeId, receiveService.getCurrNodeID().toString());
		}
		
		// If there is a send ancestor of serviceToEliminateAtNewADoc that points to it with a afterActivated
		// or afterterminated pointer, we have to update these pointers too (in order to point to the
		// receive)
		LinkedList<ADNode> path = new LinkedList<ADNode>();
		ADNode currentNode = toBeEliminatedService;
		while(currentNode != null)
		{
			path.addFirst(currentNode);
			currentNode = currentNode.getFather();
		}
				
		while(path.size() != 0)
		{
			ADNode tempNode = path.removeFirst();
			if(tempNode instanceof Service)
			{
				Service serviceTempNode = (Service) tempNode;
				if(serviceTempNode.getServiceName().toString().equals("send"))
				{
					if(serviceTempNode.getAfterActivated().contains(toBeEliminatedService.getCurrNodeID().toString()))
					{
						serviceTempNode.getAfterActivated().remove(toBeEliminatedService.getCurrNodeID().toString());
						serviceTempNode.getAfterActivated().add(receiveService.getCurrNodeID().toString());
					}
					if(serviceTempNode.getAfterTerminated().contains(toBeEliminatedService.getCurrNodeID().toString()))
					{
						serviceTempNode.getAfterTerminated().remove(toBeEliminatedService.getCurrNodeID().toString());
						serviceTempNode.getAfterTerminated().add(receiveService.getCurrNodeID().toString());
					}
				}
			}
		}
				
		// Now we create the send that would transfer data to the above receive 
		Service sendService = new Service();
		sendService.setServiceName(new StringBuffer("send"));
		sendService.setIsSpecialService(true);
		sendService.setFather(supplierService.getFather());
		sendService.setExecutionPeer(ADocument.returnCallingPeerOfSpecificWebService(supplierService, givenADoc));
		
		// We add the address of the supplier service in order to be possible for the 
		// send to take supplier service's results (and send them to the receives)
		IdContainer tempIdContainer = new IdContainer(supplierService.getOrigNodeID().toString(), supplierService.getOrigDocID().toString(), supplierService.getOrigPeerID().toString());
		IdWithHostContainer tempIdWithHostContainer = new IdWithHostContainer(tempIdContainer, callingPeerOfSupplierService);
		
		SimpleNode supplierServiceNewADocAddr = new SimpleNode(new StringBuffer(), tempIdWithHostContainer, new LinkedList<StringBuffer>());
		sendService.insertAsChild(supplierServiceNewADocAddr);

		// We add also the address of the respective receive
		IdContainer tempIdContainer2 = new IdContainer(receiveService.getCurrNodeID().toString(), receiveService.getCurrDocID().toString(), receiveService.getCurrPeerID().toString());
		IdWithHostContainer tempIdWithHostContainer2 = new IdWithHostContainer(tempIdContainer2, receiveService.getExecutionPeer().toString());
		
		SimpleNode receiveAddr = new SimpleNode(new StringBuffer(), tempIdWithHostContainer2, new LinkedList<StringBuffer>());
		sendService.insertAsChild(receiveAddr);
		
		// We put an afterActivated tag to the send service call in order to wait
		// for the activation of the master service call
		sendService.getAfterActivated().add(supplierService.getCurrNodeID().toString());
		
		// Here we are setting the return address of the receive
		IdContainer tempIdContainer3 = new IdContainer(sendService.getCurrNodeID().toString(), sendService.getCurrDocID().toString(), sendService.getCurrPeerID().toString());
		IdWithHostContainer tempIdWithHostContainer3 = new IdWithHostContainer(tempIdContainer3, sendService.getExecutionPeer().toString());
		
		SimpleNode sendAddr = new SimpleNode(new StringBuffer(), tempIdWithHostContainer3, new LinkedList<StringBuffer>());
		receiveService.insertAsChild(sendAddr);
		
		// TODO What I mention bellow and I have commented is wrong but
		// TODO I should look at it again 
//		// Here I set the hashCodeOfTreeRootedByCurrService value of all the services that are ancestors
//		// of the cse part to 0. I do this because the previous hashCodeOfTreeRootedByCurrService
//		// value has no meaning and there is no obvious way to calculate a new value.
//		if(receiveService.getFather() != null)
//		{
//			ADNode father = receiveService.getFather();
//			while(father != null)
//			{
//				if(father instanceof Service)
//				{
//					Service fatherService = (Service) father;
//					fatherService.setHashCodeOfTreeRootedByCurrService(0);
//				}
//				father = father.getFather();
//			}
//		}

		// We add send sc as a child of an ancestor newNode (of the supplierService) or as 
		// child of the root of the document if the no newNode ancestor
		// is found
		ADNode ancestorOfTheSupplierService = UsefulMethods.searchForTheClosestAncestorOfGivenNodeThatIsUnderNewNodeOrRoot(givenADoc, supplierService);
		ancestorOfTheSupplierService.insertAsChild(sendService);
		
		// Here I check if the father of the supplierServiceAtReturningADocument is a send. If this is the case,
		// the send should have an afterActivated pointer to the supplierServiceAtReturningADocument. If this is
		// true then I add another afterActivated pointer to the newly added send subtree
		ADNode fatherOfSupplierServiceAtReturningADocument = supplierService.getFather();
		if(fatherOfSupplierServiceAtReturningADocument != null && fatherOfSupplierServiceAtReturningADocument instanceof Service)
		{
			Service serviceFatherOfSupplierServiceAtReturningADocument = (Service) fatherOfSupplierServiceAtReturningADocument;
			if(serviceFatherOfSupplierServiceAtReturningADocument.getServiceName().toString().equals("send"))
			{
				if(serviceFatherOfSupplierServiceAtReturningADocument.getAfterActivated().contains(supplierService.getCurrNodeID().toString()))
					serviceFatherOfSupplierServiceAtReturningADocument.getAfterActivated().add(sendService.getCurrNodeID().toString());
			}
		}
			
			
//			// Here I set the hashCodeOfTreeRootedByCurrService value of all the services that are ancestors
//			// of masterServiceNewADoc to 0. I do this because the previous hashCodeOfTreeRootedByCurrService
//			// value has no meaning and there is no obvious way to calculate a new value.
//			if(masterServiceNewADoc.getFather() != null)
//			{
//				ADNode father = masterServiceNewADoc.getFather();
//				while(father != null)
//				{
//					if(father instanceof Service)
//					{
//						Service fatherService = (Service) father;
//						fatherService.setHashCodeOfTreeRootedByCurrService(0);
//					}
//					father = father.getFather();
//				}
//			}
			
			// We remove the unnecessary sends that were left from the elimination
			// of the eliminated subtrees
//			removeFromGivenADocTheGivenSendDestinationsAndRespectiveSends(returningADocument, sendDestinationsToBeRemoved);
//			if(returningADocument.getADocNumber() == 232)
//			{
//				System.out.println("lalal");
//				removeFromGivenADocTheGivenSendDestinationsAndRespectiveSends(returningADocument, sendDestinationsToBeRemoved);
////				GraphvizModule.createOutputImage(returningADocument, "spyros1");
//			}
			
			
			// update the number of simple transfers
//			String masterServiceCallingPeer = ADocument.returnCallingPeerOfSpecificWebService(supplierServiceAtReturningADocument, returningADocument).toString();
//			returningADocument.updateNetworkCostCSE(eliminatedServices, masterServiceCallingPeer);
			
		givenADoc.increaseNumberOfRewritings();
		givenADoc.increaseNumberOfCSE();
		
		givenADoc.calculateHashCode();
		
		// We add the added send and receive service calls to the addedSendAndReceiveConatiner
		addedSendAndReceiveConatiner.add(sendService);
		addedSendAndReceiveConatiner.add(receiveService);
		
		// We update originalIdsOfNodesNotUnderReceivesToADNodeRefs map of
		// returningADocument with the values of the new send, receive
		// and the service call that is being pointed by send
		
		// a record for receive
		givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(tempIdContainer2, receiveService);
		// a record for send
		givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(tempIdContainer3, sendService);
		// a record for the service call that is being pointed by send
		givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(tempIdContainer, supplierService);

		// We update also the originalIdWithHostAddressesThatPointToThisNode
		// of the new send and the new receive
		receiveService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(receiveAddr);
		sendService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(sendAddr);
		supplierService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(supplierServiceNewADocAddr);
		supplierService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(supplierServiceNewADocAddr2);
		
		// We delete from the givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs and 
		// someNode's originalIdWithHostAddressesThatPointToThisNode the records that belong
		// to the nodesThatShouldBeDeletedIter
		Iterator<ADNode> nodesThatShouldBeDeletedIter = nodesThatShouldBeDeleted.iterator();
		while(nodesThatShouldBeDeletedIter.hasNext())
		{
			ADNode currNode = nodesThatShouldBeDeletedIter.next();
			ADNode fatherOfCurrNode = currNode.getFather();
			
			currNode.getFather().getChildren().refRemove(currNode);
			currNode.setFather(null);
			currNode.getChildren().clear();

			// we check if the father of currNode is a newNode which is now left without
			// any children. If this is the case, this newNode has to be removed
			if(fatherOfCurrNode != null && fatherOfCurrNode instanceof Service)
				UsefulMethods.deleteRecursivellyAncestorNewNodesWithNoChildren((Service)fatherOfCurrNode);
			
			// If this currNode is pointing to some other node (originalId pointer), we 
			// remove it from the pointed's node originalIdWithHostAddressesThatPointToThisNode
			if(currNode instanceof SimpleNode)
			{
				SimpleNode simpleNodeCurrNode = (SimpleNode) currNode;
				if(simpleNodeCurrNode.getNodeData().length() == 0 && !simpleNodeCurrNode.getRemoteOriginalIdAndHostingPeer().isEmpty())
				{
					IdWithHostContainer pointer = simpleNodeCurrNode.getRemoteOriginalIdAndHostingPeer();
					ADNode nodeWhichIsBeingPointed = givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(pointer.id);
					nodeWhichIsBeingPointed.getOriginalIdWithHostAddressContainersThatPointToThisNode().refRemove(currNode);
				}
			}
		}
		
		nodesThatShouldBeDeletedIter = nodesThatShouldBeDeleted.iterator();
		while(nodesThatShouldBeDeletedIter.hasNext())
		{
			ADNode currNode = nodesThatShouldBeDeletedIter.next();
			// If there is a record for this node at givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs
			// we remove it
			IdContainer currNodeOrigId = new IdContainer(currNode.getOrigNodeID().toString(), currNode.getOrigDocID().toString(), currNode.getOrigPeerID().toString());
			if(givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(currNodeOrigId) != null)
				givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.remove(currNodeOrigId);
			
		}
		
		return addedSendAndReceiveConatiner;
	}
	
	// The following function returns all possible sets and subsets of the given
	// set that have a size greater that 1 and maximum size 2	
	private List<List<Service>> findAllPossibleListsAndSubListsOfSize2(List<Service> givenSet)
	{
		List<List<Service>> returningResult = new ArrayList<List<Service>>();
		
		if(givenSet.size() < 2)
			return returningResult;
		
		for(int i=0; i<givenSet.size(); i++)
		{
			for(int j=0; j<givenSet.size(); j++)
			{
				if(i == j)
					continue;
				
				List<Service> newList = new ArrayList<Service>();
				newList.add(givenSet.get(i));
				newList.add(givenSet.get(j));
				returningResult.add(newList);
			}
		}
		
		return returningResult;
	}
}
