package fr.inria.gemo.axml.optimax.rules;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.sun.org.apache.xerces.internal.util.NamespaceSupport;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;
import fr.inria.gemo.axml.optimax.helping.DinstinctNumberGenerator;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.helping.NamespaceContextWrapper;
import fr.inria.gemo.axml.optimax.helping.Utilities;
import fr.inria.gemo.axml.optimax.info.DefaultNamespaceDeclarator;

public class Composition {

	NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
	String axmlPrefix = null;
	String queryPrefix = null;
	
	public Composition() {
		axmlPrefix = myNamespaceSupport.getPrefix(Parameters.axmlURI);
		queryPrefix = myNamespaceSupport.getPrefix(Parameters.queryURI);
	}
	
	class GqsServiceChild {
		Service service;
		String paramName;
	}
	
	class GqsWithGqsChildren {
		Service gqs;
		LinkedList<GqsServiceChild> gqsChildren = new LinkedList<GqsServiceChild>();
	}
	
	public List<ADocument> queryComposition(ADocument givenADoc)
	{
		LinkedList<ADocument> listOfReturningADocs = new LinkedList<ADocument>();
	
//		VerboseADocumentDrawingModule.createOutputImage(givenADoc, "fileName");
		
		// Here we are going to find all the service calls of the document
		// that are calls to a generic query service
		ExtendedLinkedList<Service> sCsNotUnderReceive = UsefulMethods.returnServicesUnderneathASpecificNodeStopOnReceive(givenADoc.getRootOfADoc());
		Iterator<Service> sCsNotUnderReceiveIter = sCsNotUnderReceive.iterator();
		while(sCsNotUnderReceiveIter.hasNext())
		{
			Service currService = sCsNotUnderReceiveIter.next();
			if(!currService.getServiceName().toString().equals(Parameters.GQS))
				sCsNotUnderReceiveIter.remove();
		}
		
		// We are going to search for each GCS if it is a continuous GCS
		// and if it has any other services in its parameters. 
		sCsNotUnderReceiveIter = sCsNotUnderReceive.iterator();
		while(sCsNotUnderReceiveIter.hasNext())
		{
			// for each of the found children scs
			Service currService = sCsNotUnderReceiveIter.next();
			List<GqsServiceChild> children = returnServiceDirectDiscendantsOfContGQS(currService);
			// we are going to keep only the direct descendants which are continuous GQS calls
			Iterator<GqsServiceChild> childrenIter = children.iterator();
			while(childrenIter.hasNext())
			{
				if(checkIfItIsContGQS(childrenIter.next().service) == null)
					childrenIter.remove();
			}
			
			if(children.isEmpty())
				continue;
			
			// for all the possible combinations of the left children, try to apply query composition
			List<List<GqsServiceChild>> allPossibleCombinations = findAllThePossibleCombinations(children);
			
			// for each of these combinations execute applyQueryComposition
			Iterator<List<GqsServiceChild>> allPossibleCombinationsIter = allPossibleCombinations.iterator();
			while(allPossibleCombinationsIter.hasNext())
			{
				List<GqsServiceChild> currCombination = allPossibleCombinationsIter.next();
				
				List<Service> listOfServ = new LinkedList<Service>();
				listOfServ.add(currService);
				Iterator<GqsServiceChild> currCombinationIter = currCombination.iterator();
				while(currCombinationIter.hasNext())
					listOfServ.add(currCombinationIter.next().service);
				
				if(!checkIfTheGivenScHaveTheSameExecutionPeer(listOfServ))
					continue;
				
				ADocument returnedADoc = applyQueryComposition(givenADoc, currService, currCombination);
				System.out.println(returnedADoc);
				listOfReturningADocs.add(returnedADoc);
			}
		}
		return listOfReturningADocs;
	}
	
	private boolean checkIfTheGivenScHaveTheSameExecutionPeer(List<Service> givenServices)
	{
		Iterator<Service> givenServicesIter =  givenServices.iterator();
		String commonExecutionPeer = "";
		while(givenServicesIter.hasNext())
		{
			Service currService  = givenServicesIter.next();
			if(commonExecutionPeer.length() == 0)
				commonExecutionPeer = currService.getExecutionPeer().toString();
			else
			{
				if(!currService.getExecutionPeer().toString().equals(commonExecutionPeer))
					return false;
			}
		}
		return true;
	}
	
	private ADocument applyQueryComposition(ADocument givenADoc, Service givenService, List<GqsServiceChild> givenGqsServiceChildren)
	{
		ADocument newADoc = new ADocument(givenADoc);
		Service givenServiceAtNewADoc = (Service)ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, givenService);
		
		List<GqsServiceChild> givenGqsServiceChildrenAtNewADoc = new LinkedList<GqsServiceChild>();
		Iterator<GqsServiceChild> givenGqsServiceChildrenIter = givenGqsServiceChildren.iterator();
		while(givenGqsServiceChildrenIter.hasNext())
		{
			GqsServiceChild givenGqsServiceChild = givenGqsServiceChildrenIter.next();
			Service givenServiceChildAtNewADoc = (Service)ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, givenGqsServiceChild.service);
			GqsServiceChild givenGqsServiceChildAtNewNode = new GqsServiceChild();
			givenGqsServiceChildAtNewNode.paramName = givenGqsServiceChild.paramName;
			givenGqsServiceChildAtNewNode.service = givenServiceChildAtNewADoc;
			givenGqsServiceChildrenAtNewADoc.add(givenGqsServiceChildAtNewNode);
		}
		
		// We keep the hashcode of the old xquery of the father GQS
		// First we locate the continuousQuery node
		SimpleNode continuousQueryNodeOfFather = checkIfItIsContGQS(givenServiceAtNewADoc);
		// then we search for the declaration node
		SimpleNode declarationNodeOfFather = null;
		Iterator<ADNode> childrenOfCQNIter = continuousQueryNodeOfFather.getChildren().iterator();
		while(childrenOfCQNIter.hasNext())
		{
			ADNode currChild = childrenOfCQNIter.next();
			if(currChild instanceof SimpleNode)
			{
				SimpleNode simpleNodeCurrChild = (SimpleNode) currChild;
				if(simpleNodeCurrChild.getNodeData().toString().equals(new String(queryPrefix + ":declaration")))
					declarationNodeOfFather = simpleNodeCurrChild;
			}
		}
		Document domWithXQueryOfFather = declarationNodeOfFather.declaration;
		int oldHashOfXQueryOfFather = Utilities.modHashCode(Utilities.DomToStringSerializerWithoutIdentationAndDeclar(domWithXQueryOfFather).toString());
		
		// here we make the compositions
		Iterator<GqsServiceChild> givenGqsServiceChildrenAtNewADocIter = givenGqsServiceChildrenAtNewADoc.iterator();
		while(givenGqsServiceChildrenAtNewADocIter.hasNext())
		{
			composeFatherWithChild(givenServiceAtNewADoc, givenGqsServiceChildrenAtNewADocIter.next());
		}
		
		// we calculate the hashcode of the new xquery of the father of GGS
		int newHashOfXQueryOfFather = Utilities.modHashCode(Utilities.DomToStringSerializerWithoutIdentationAndDeclar(domWithXQueryOfFather).toString());
		
		// we update the hashcode of the father GQS. We substract modulo the old hashcode
		// and we modulo add the newHash code
		int tempHash = givenServiceAtNewADoc.getHashCodeOfTreeRootedByCurrService();
		tempHash = Utilities.amodN(((long)tempHash - (long)oldHashOfXQueryOfFather), Integer.MAX_VALUE).intValue();
		tempHash = Utilities.amodN(((long)tempHash + (long)newHashOfXQueryOfFather), Integer.MAX_VALUE).intValue();
		givenServiceAtNewADoc.setHashCodeOfTreeRootedByCurrService(tempHash);
		
		// calculate network cost from scratch
		newADoc.calculateNetworkCost();
		
		// We add at the newADoc the rule that we applied
		boolean decreaseCost = false;
		boolean increaseCost = false;

		if(givenADoc.getSimpleNetworkCost() < newADoc.getSimpleNetworkCost())
			increaseCost = true;

		if(givenADoc.getSimpleNetworkCost() > newADoc.getSimpleNetworkCost())
			decreaseCost = true;
		
		newADoc.insertAppliedRuleToADoc("comp", givenADoc.getADocNumber(), givenServiceAtNewADoc.getCurrNodeID().toString(), "", newADoc.getADocNumber(), increaseCost, decreaseCost, givenADoc.getSimpleNetworkCost(), newADoc.getSimpleNetworkCost());
		
		return newADoc; 
	}
	
	private int findMinPos(int a, int b)
	{
		if(a < 0 && b < 0) // if both are negative
			return -1;
		if(a < 0) // if a is negative
			return b;
		if(b < 0) // if b is negative
			return a;
		else // if both are positive
		{
			int res = a;
			if(b<a)
				res = b;
			return res;
		}
	}
	
	private int findMinPos(int a, int b, int c)
	{
		return findMinPos(a, findMinPos(a, b));
	}
	
	private void composeFatherWithChild(Service fatherSc, GqsServiceChild childSc)
	{
		Document fatherDeclaration = null;
		// we get the simpleNode that keeps the declaration Dom
		List<ADNode> childrenOfFatherSc = checkIfItIsContGQS(fatherSc).getChildren();
		Iterator<ADNode> childrenOfFatherScIter = childrenOfFatherSc.iterator();
		while(childrenOfFatherScIter.hasNext())
		{
			ADNode currChild = childrenOfFatherScIter.next();
			if(currChild instanceof SimpleNode)
			{
				SimpleNode simpleNodeCurrChild = (SimpleNode) currChild;
				if(simpleNodeCurrChild.getNodeData().toString().equals(new String(queryPrefix + ":declaration")))
					fatherDeclaration = simpleNodeCurrChild.declaration;
			}
		}
		Set variablesOfFatherSc = findVariables(fatherDeclaration);
		
		
		Document childDeclaration = null;
		// we get the simpleNode that keeps the declaration Dom
		List<ADNode> childrenOfChildSc = checkIfItIsContGQS(childSc.service).getChildren();
		Iterator<ADNode> childrenOfChildScIter = childrenOfChildSc.iterator();
		while(childrenOfChildScIter.hasNext())
		{
			ADNode currChild = childrenOfChildScIter.next();
			if(currChild instanceof SimpleNode)
			{
				SimpleNode simpleNodeCurrChild = (SimpleNode) currChild;
				if(simpleNodeCurrChild.getNodeData().toString().equals(new String(queryPrefix + ":declaration")))
					childDeclaration = simpleNodeCurrChild.declaration;
			}
		}
		Set<String> variablesOfChildSc = findVariables(childDeclaration);
		
		// we will check if any of the variables of variablesOfChildSc are
		// being contained in variablesOfFatherSc. If yes we are going to
		// replace that variable with a new one in order to avoid conflict
		Map<String,String> oldToNewChildVariables = new HashMap<String,String>();
		LinkedList<String> conflicts = new LinkedList<String>();
		DinstinctNumberGenerator dinstinctNumberGeneratorInstance = new DinstinctNumberGenerator();
		
		Iterator<String> variablesOfChildScIter = variablesOfChildSc.iterator();
		while(variablesOfChildScIter.hasNext())
		{
			String currVariableOfChild = variablesOfChildScIter.next();
			// if we have a conflict...
			if(variablesOfFatherSc.contains(currVariableOfChild))
			{
				String newVariableName = new String("$var" + dinstinctNumberGeneratorInstance.getNumberInt());
				oldToNewChildVariables.put(currVariableOfChild, newVariableName);
				conflicts.add(currVariableOfChild);
			}
		}
		
		// we will scan the childDeclaration and change the variables to the new ones
		// according to the map
		changeVariablesAccordingToMap(childDeclaration, oldToNewChildVariables, conflicts);
		
		// we locate the node where we want to put the child in the father definition
		String paramNameOfChild = childSc.paramName;
		Node positionWhereWeShouldAddData = findParamNode(fatherDeclaration, paramNameOfChild);
		
		// we take all the child elements of the root of the child and we import them
		Node rootOfChild = childDeclaration.getDocumentElement();
		
//		System.out.println(Utilities.ElementToStringSerializer((Element)positionWhereWeShouldAddData, false, true));
//		
//		System.out.println("--------");
//		
//		System.out.println(Utilities.DomToStringSerializerWithIdentation(fatherDeclaration));
//		System.exit(0);
		
		Node firstTxtNode = fatherDeclaration.createTextNode("{");
		positionWhereWeShouldAddData.appendChild(firstTxtNode);
		
		for(int i=0 ; i<rootOfChild.getChildNodes().getLength() ; i++)
		{
			// for each node that we import
			Node subtreeToBeAddedToTheFather = fatherDeclaration.importNode(rootOfChild.getChildNodes().item(i), true);
			// we place it at the right position
			positionWhereWeShouldAddData.appendChild(subtreeToBeAddedToTheFather);
		}
		
		Node secondTxtNode = fatherDeclaration.createTextNode("}");
		positionWhereWeShouldAddData.appendChild(secondTxtNode);
		
		// we locate the params of the childSc and we copy them at the parameters of fatherSc
		List<ADNode> paramsOfChild = returnParamsOfContGQS(childSc.service);
		
		// we remove the child with its paramContainer from the father
		ADNode paramContainerOfChild = childSc.service.getFather();
		ADNode fatherOfParamContainerOfChild = paramContainerOfChild.getFather();
		paramContainerOfChild.removeNode();
		
		// we add all the paramsOfChild as parameters of fatherSc
		while(!paramsOfChild.isEmpty())
		{
			paramsOfChild.get(0).removeNode();
			fatherOfParamContainerOfChild.insertAsChild(paramsOfChild.get(0));
			paramsOfChild.remove(0);
		}
	}
	
	private void changeVariablesAccordingToMap(Document givenDoc, Map<String,String> oldToNewVar, List<String> conflicts)
	{
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
		try {
			NodeList returnedNodes = (NodeList) xPath.evaluate("//node()" , givenDoc, XPathConstants.NODESET);
			for(int i=0; i<returnedNodes.getLength(); i++)
			{
				Node currNode = returnedNodes.item(i);
				if(currNode.getNodeType() == Node.TEXT_NODE)
				{
					String textOfTextNode = currNode.getTextContent();
					Iterator<String> conflictsIter = conflicts.iterator();
					while(conflictsIter.hasNext())
					{
						String currConflict = conflictsIter.next();
					
						if(textOfTextNode.contains(currConflict))
						{
							textOfTextNode.replace(currConflict, oldToNewVar.get(currConflict));
							currNode.setTextContent(textOfTextNode);
						}
					}
				}
			}
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	private Node findParamNode(Document givenDom, String nameOfParam)
	{
		Node returningNode = null;
		
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
//		
//		 + ":param name=\"" + nameOfParam + "\"
//		
		
		try {
			returningNode = (Node) xPath.evaluate("//" + queryPrefix + ":param[@name=\"" + nameOfParam + "\"]", givenDom, XPathConstants.NODE);
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return returningNode;
	}
	
	private Set<String> findVariables(Document givenDom)
	{
		Set<String> variables = new HashSet<String>();
		
		XPathFactory factory = XPathFactory.newInstance();
		XPath xPath = factory.newXPath();
		NamespaceSupport myNamespaceSupport = DefaultNamespaceDeclarator.getNamespaceSupportWithDefaultNamespaces();
		xPath.setNamespaceContext( new NamespaceContextWrapper(myNamespaceSupport));
		
		try {
			NodeList returnedNodes = (NodeList) xPath.evaluate("//node()" , givenDom, XPathConstants.NODESET);
			for(int i=0; i<returnedNodes.getLength(); i++)
			{
				Node currNode = returnedNodes.item(i);
				if(currNode.getNodeType() == Node.TEXT_NODE)
				{
					String textOfTextNode = currNode.getTextContent();
					if(textOfTextNode.contains("$"))
					{
						StringBuffer myStr = new StringBuffer(textOfTextNode);
						while(myStr.length() != 0)
						{
							int pos1 = myStr.indexOf("$");
							if(pos1 != -1)
								myStr.delete(0, pos1);
							else
								break;
							int pos2 = myStr.indexOf(" ");
							int pos3 = myStr.indexOf("/");
//							int pos4 = myStr.indexOf("");
							int pos4 = -1;
							int pos5 = findMinPos(pos2, pos3, pos4);
							if(pos5 != -1)
							{
								StringBuffer valueToBeAdded = new StringBuffer(myStr.substring(0, pos5));
								if(valueToBeAdded.charAt(valueToBeAdded.length() - 1) == '\n')
									valueToBeAdded.delete(valueToBeAdded.length() - 1, valueToBeAdded.length());
								variables.add(valueToBeAdded.toString());
								myStr.delete(0, pos5);
							}
							else 
								break;
						}
					}
				}
			}
		} catch (XPathExpressionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return variables;
	}
	
	// here we are going to find all the possible combinations of the givenServices list
	// TODO find them
	private List<List<GqsServiceChild>> findAllThePossibleCombinations(List<GqsServiceChild> givenServices)
	{
		List<List<GqsServiceChild>> returnignList = new LinkedList<List<GqsServiceChild>>();
		returnignList.add(givenServices);
		return returnignList;
	}
	
	private List<ADNode> returnParamsOfContGQS(Service givenService)
	{
		List<ADNode> returningParamList = new LinkedList<ADNode>();
		
		// first of all we are going to check if it is a continuous GQS
		SimpleNode qContinuousQuery = checkIfItIsContGQS(givenService);
		// if it is not we return an empty list
		if(qContinuousQuery == null)
			return returningParamList;
		// if it is we search for its children
		else
		{
			// we search the q:param children of the q:continuousQuery
			SimpleNode  qParam= null;
			for(int i=0; i<qContinuousQuery.getChildren().size(); i++)
			{
				if(qContinuousQuery.getChildren().get(i) instanceof SimpleNode)
				{
					SimpleNode currSimpleChild = (SimpleNode)qContinuousQuery.getChildren().get(i);
					if(currSimpleChild.getNodeData().toString().equals(new String(queryPrefix + ":param")))
					{
						qParam = currSimpleChild;
						returningParamList.add(qParam);
					}
				}
			}
			// if no q:param found then we report it and
			// we continue with the next QCS call
			if(qParam == null)
			{
				System.err.println("queryComposition: The GQS (" + givenService + ") does not have a q:param descendant");
				return returningParamList;
			}
		}
		return returningParamList;
	}
	
	// It check if the given query is a continuous query and if yes
	// it returns its continuousQuery node. Otherwise it returns null
	public SimpleNode checkIfItIsContGQS(Service givenService)
	{
		// we search for the axmlWsSoap child
		SimpleNode axmlWsSoap = null;
		for(int i=0; i<givenService.getChildren().size(); i++)
		{
			if(givenService.getChildren().get(i) instanceof SimpleNode)
			{
				SimpleNode currSimpleChild = (SimpleNode)givenService.getChildren().get(i);
				if(currSimpleChild.getNodeData().toString().equals(new String(axmlPrefix + ":ws-soap")))
				{
					axmlWsSoap = currSimpleChild;
				}
			}
		}
		// if no axmlWsSoap found then something is going wrong
		// we continue with the next QCS call
		if(axmlWsSoap == null)
		{
			System.err.println("queryComposition: The GQS (" + givenService + ") does not have a axmlWsSoap child");
			return null;
		}
		
		// we search for the q:continuousQuery child of the axml:ws-soap
		SimpleNode  qContinuousQuery= null;
		for(int i=0; i<axmlWsSoap.getChildren().size(); i++)
		{
			if(axmlWsSoap.getChildren().get(i) instanceof SimpleNode)
			{
				SimpleNode currSimpleChild = (SimpleNode)axmlWsSoap.getChildren().get(i);
				if(currSimpleChild.getNodeData().toString().equals(new String(queryPrefix + ":continuousQuery")))
				{
					qContinuousQuery = currSimpleChild;
				}
			}
		}
		// if no q:continuousQuery found then we report it and
		// we continue with the next QCS call
		if(qContinuousQuery == null)
		{
			System.err.println("queryComposition: The GQS (" + givenService + ") does not have a q:continuousQuery descendant");
			return null;
		}
		else
			return qContinuousQuery;
	}
	
	private List<GqsServiceChild> returnServiceDirectDiscendantsOfContGQS(Service givenService)
	{
		List<GqsServiceChild> returningList = new LinkedList<GqsServiceChild>();
		
		// first of all we are going to check if it is a continuous GQS
		SimpleNode qContinuousQuery = checkIfItIsContGQS(givenService);
		// if it is not we return an empty list
		if(qContinuousQuery == null)
			return returningList;
		// if it is we search for its children
		else
		{
			// we search the q:param children of the q:continuousQuery
			SimpleNode  qParam= null;
			for(int i=0; i<qContinuousQuery.getChildren().size(); i++)
			{
				if(qContinuousQuery.getChildren().get(i) instanceof SimpleNode)
				{
					SimpleNode currSimpleChild = (SimpleNode)qContinuousQuery.getChildren().get(i);
					if(currSimpleChild.getNodeData().toString().equals(new String(queryPrefix + ":param")))
					{
						qParam = currSimpleChild;
						if(qParam.getChildren().size() == 1)
						{
							// TODO we have made the convention that in the parameters, if we are going
							// TODO to have a service call, this would be just one
							GqsServiceChild gqsServiceChild = new GqsServiceChild();
							gqsServiceChild.paramName = qParam.paramName;
							if(qParam.getChildren().getFirst() instanceof Service)
							gqsServiceChild.service = (Service)qParam.getChildren().getFirst();
							returningList.add(gqsServiceChild);
						}
					}
				}
			}
			// if no q:param found then we report it and
			// we continue with the next QCS call
			if(qParam == null)
			{
				System.err.println("queryComposition: The GQS (" + givenService + ") does not have a q:param descendant");
				return returningList;
			}
		}
		return returningList;
	}
}
