package fr.inria.gemo.axml.optimax.rules;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.IdContainer;
import fr.inria.gemo.axml.optimax.adocument.IdWithHostContainer;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;
import fr.inria.gemo.axml.optimax.io.output.GraphvizModule;

public class Delegation {

	public List<ADocument> delegationRuleForADocs(ADocument givenADoc, List<StringBuffer> listOfPeers)
	{
		List<ADocument> returningADocs = new ArrayList<ADocument>();
		
		// Here we are going to find the services that are capable of delegation
		List<Service> servicesOfADoc = ADocument.findServicesOfADoc(givenADoc);
		Iterator<Service> servicesOfADocIter = servicesOfADoc.iterator();
		while(servicesOfADocIter.hasNext())
		{
			Service currService = servicesOfADocIter.next();
			if(currService.getIsSpecialService())
				servicesOfADocIter.remove();
		}
		
		// We will try to delegate one by one the service calls
		// that were left to all possible peers
		servicesOfADocIter = servicesOfADoc.iterator();
		System.out.println("servicesOfADoc:" + servicesOfADoc.size());
		while(servicesOfADocIter.hasNext())
		{
			Service currDelegService = servicesOfADocIter.next();
			for(int i=0; i<listOfPeers.size(); i++)
			{
				StringBuffer currDelegPeer = listOfPeers.get(i);
				// if this peer is not proper for delegation then
				// we will go to the next one
				if(checkIfGivenDelegationsWillCauseProblemToGivenServiceOrToItsSubTree(currDelegService, currDelegPeer.toString()) == true)
					continue;
				
				// Here we are going to find all the nodes of the subtree to be delegated
				ExtendedLinkedList<Service> servicesOfSubTreeToBeDelegatedWithFrontier = UsefulMethods.returnServicesUnderneathASpecificNodeStopOnNewNodeAndReceive(currDelegService);
				// 1)We are going to search if in the toBeDelegated Sc nodes (newNode & receive frontier) there is a send 
				// with a (symbolical - originalId) pointer in its what part and this pointer is pointing to a sc which is 
				// outside the toBeDelegated subtree (newNode & receive frontier). If such a case exists we will have a 
				// problem because the to be accessed data would not be in the same doc and (maybe) peer
				// 2)In the same loop we search for scs (set A) that are being pointed (what part) by sends or receives which 
				// are outside the to be delegated subtree (newNode & receive frontier). If this happens its an error because 
				// the setA will change document and possibly delegation peer
				boolean abortThisDelegation = false;
				for(int j=0; j<servicesOfSubTreeToBeDelegatedWithFrontier.size() && abortThisDelegation == false; j++)
				{
					Service serviceCurrNode = (Service) servicesOfSubTreeToBeDelegatedWithFrontier.get(j);
					// Here we examine case #1
					if(serviceCurrNode.getServiceName().toString().equals("send"))
					{
						ADNode whatChildOfSend = serviceCurrNode.getChildren().get(0);
						if(whatChildOfSend instanceof SimpleNode)
						{
							SimpleNode simpleNodeWhatChildOfSend = (SimpleNode) whatChildOfSend;
							// if the what part is a symbolic (original id) pointer
							if(simpleNodeWhatChildOfSend.getNodeData().length() == 0 && !simpleNodeWhatChildOfSend.getRemoteOriginalIdAndHostingPeer().isEmpty())
							{
								// we will check if it is pointing to a service which is outside the toBeDelegated subtree
								IdContainer pointedServiceCallsOriginalId = simpleNodeWhatChildOfSend.getRemoteOriginalIdAndHostingPeer().id;
								Service pointedServiceCall = (Service) givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(pointedServiceCallsOriginalId);
								// ...if yes...
								if(!servicesOfSubTreeToBeDelegatedWithFrontier.refContains(pointedServiceCall))
								{
									abortThisDelegation = true;
									break;
								}
							}
						}
					}
					// Here we examine case #2
					// if some services are pointing to this service call
					if(!serviceCurrNode.getOriginalIdWithHostAddressContainersThatPointToThisNode().isEmpty())
					{
						Iterator<SimpleNode> addressContainersThatPointToThisNodeIter = serviceCurrNode.getOriginalIdWithHostAddressContainersThatPointToThisNode().iterator();
						while(addressContainersThatPointToThisNodeIter.hasNext())
						{
							SimpleNode currAddressContainerThatPointsToThisNode = addressContainersThatPointToThisNodeIter.next();
							// we check if it is the what part of the send or receive. We can do this
							// by checking if it is the first child of a send or a receive
							if(currAddressContainerThatPointsToThisNode.getFather().getChildren().get(0) == currAddressContainerThatPointsToThisNode)
							{
								// if yes...
								// we check if this send is in the servicesOfSubTreeToBeDelegatedWithFrontier
								if(!servicesOfSubTreeToBeDelegatedWithFrontier.refContains(currAddressContainerThatPointsToThisNode.getFather()))
								{
									abortThisDelegation = true;
									break;
								}
							}
							
						}
					}
				}
				if(abortThisDelegation == true)
				{
//					GraphvizModule.createOutputImage(givenADoc, Integer.toString(givenADoc.getADocNumber()) + currDelegPeer + currDelegService.getCurrNodeID());
					continue;
				}
				
				
				ADocument newADoc = new ADocument(givenADoc);
				Service newCurrDelegService = (Service) ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, currDelegService);
				
				// Here we are going to perform the delegation. Note that the address
				// (originalID) of the newCurrDelegService remains the same
				List<Service> receiveNewNodeAndDelegScRefContainer = performDelegation(newADoc, newCurrDelegService, currDelegPeer.toString());
				
				// Here we are going to search for receives and sends and simple scs in the delegated subtree (newNode frontier)
				// Whenever we find such a service call (simple, send or receive), we... 
				// 1)if it is a send we change the execution peer and the symbolic pointer of the respective service
				// call that is pointing to it
				// 2)if it is a simple sc then the execution peer is not changed but we have to update the symbolic
				// pointers (if any) that are pointing to this service
				List<Service> servicesOfTheDelegatedSubTreeNewNodeFrontrier = UsefulMethods.returnServicesUnderneathASpecificNodeStopOnNewNodeAndReceive(receiveNewNodeAndDelegScRefContainer.get(2));
				for(int j=0; j<servicesOfTheDelegatedSubTreeNewNodeFrontrier.size(); j++)
				{
					Service currService = servicesOfTheDelegatedSubTreeNewNodeFrontrier.get(j);
					// ... change the execution peer of each of the found sends - receives
					if(currService.getServiceName().toString().equals("send") || currService.getServiceName().toString().equals("receive"))
						currService.setExecutionPeer(new StringBuffer(currDelegPeer));
					// ... for all the addresses that point to one of the found services, change
					// the execution peer that the address is mentioning
					List<SimpleNode> simpleAddressesToBeChanged = currService.getOriginalIdWithHostAddressContainersThatPointToThisNode();
					for(int k=0; k<simpleAddressesToBeChanged.size(); k++)
						simpleAddressesToBeChanged.get(k).getRemoteOriginalIdAndHostingPeer().setHost(currDelegPeer.toString());
				}
				
				// Here we are going to update the receive descriptions that may
				// need to be updated
				UsefulMethods.ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(newADoc, newCurrDelegService);
				
				// TODO Note that we do not unify "confluent paths". The two final plans differ in the
				// TODO sense of who is installing the new Nodes. In a distributed setting is important
				// TODO this because it gives or restrict from some peers the possibility to rewrite
				// TODO the predifined strategy.
				// Here we are going to search for receives (not under newNodes) in the delegated subtree.
				// If we find receives that depend on sends that don't belong to this delegated
				// subtree and their father is a newNode then they are moved under the new newNode call
				// of the delegation. We make this in order to make sure that f(g) -> deleg(f(g())) ->
				// deleg(f(deleg(g()))) == deleg(f(deleg(g()))) <- f(deleg(g())) <- f(g())
//				checkAndMoveSendsWhichAreSendingDataToTheDelegatedSubTree(newADoc, receiveNewNodeAndDelegScRefContainer.get(1));
				
				// calculate network cost from scratch
				newADoc.calculateNetworkCost();
				
				// We add at the newADoc the rule that we applied
				boolean decreaseCost = false;
				boolean increaseCost = false;

				if(givenADoc.getSimpleNetworkCost() < newADoc.getSimpleNetworkCost())
					increaseCost = true;

				if(givenADoc.getSimpleNetworkCost() > newADoc.getSimpleNetworkCost())
					decreaseCost = true;
				
				newADoc.insertAppliedRuleToADoc("deleg", givenADoc.getADocNumber(), currDelegService.getCurrNodeID().toString(), currDelegPeer.toString(), newADoc.getADocNumber(), increaseCost, decreaseCost, givenADoc.getSimpleNetworkCost(), newADoc.getSimpleNetworkCost());
				
//				// We check if the produced newADoc is consistent
//				if(!newADoc.isADocumentConsistent())
//				{
//					System.err.println("delegationRuleForADocs: Produced inconsistent document");
//					System.exit(0);
//				}
					
				returningADocs.add(newADoc);
			}
		}
		
		return returningADocs;
	}
	
	// Here we are going to search for receives (not under newNodes) in the delegated subtree.
	// If we find receives that depend on sends that don't belong to this delegated
	// subtree and their father is a newNode then they are moved under the new newNode call
	// of the delegation. We make this in order to make sure that f(g) -> deleg(f(g())) ->
	// deleg(f(deleg(g()))) == deleg(f(deleg(g()))) <- f(deleg(g())) <- f(g())
	private void checkAndMoveSendsWhichAreSendingDataToTheDelegatedSubTree(ADocument givenADoc, Service newNodeOfTheDelegation)
	{
		ExtendedLinkedList<ADNode> newNodeSubTreeNodes = new ExtendedLinkedList<ADNode>();
		newNodeSubTreeNodes.addAll(UsefulMethods.getAllTheNodesOfSubTree(newNodeOfTheDelegation));
		
		List<Service> newNodeSubTreeReceiveNodes = UsefulMethods.returnReceivesUnderneathASpecificNodeStopOnNewNodeAndReceive(newNodeOfTheDelegation.getChildren().get(0));
		for(Iterator<Service> newNodeSubTreeReceiveNodesIter = newNodeSubTreeReceiveNodes.iterator(); newNodeSubTreeReceiveNodesIter.hasNext();)
		{
			Service currReceive = newNodeSubTreeReceiveNodesIter.next();
			
			// if is not a delegation receive skip it
			if(currReceive.getChildren().get(0) instanceof SimpleNode)
			{
				SimpleNode firstChildOfReceive = ((SimpleNode)currReceive.getChildren().get(0));
				if(firstChildOfReceive.getNodeData().length() == 0 && !firstChildOfReceive.getRemoteOriginalIdAndHostingPeer().isEmpty())
					continue;
			}
			
			// we get the container that contains the address of the respective send
			SimpleNode sendAddrCont = (SimpleNode) currReceive.getChildren().get(currReceive.getChildren().size() - 1);
			
			// we search for the respective send
			Service sendRef = (Service) givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(sendAddrCont.getRemoteOriginalIdAndHostingPeer().id);
			
			// we check if the send is under newNode
			if(!newNodeSubTreeNodes.refContains(sendRef))
			{
				// if not, then we check if the parent of send is another newNode. 
				if(sendRef.getFather() instanceof Service)
				{
					Service serviceSendsFather = (Service) sendRef.getFather();
					if(serviceSendsFather.getServiceName().toString().equals("newNode"))
					{
						Service newNodeOfSend = serviceSendsFather;
						Service copyOfNewNodeOfSend = (Service) newNodeOfSend.returnNewNodeWithNoChildrenAndFatherButWithNewCurrIds();
						
						// we create a new newNode under newNodeOfTheDelegation
						newNodeOfTheDelegation.insertAsChild(copyOfNewNodeOfSend);
						
						// we move under the copyOfNewNodeOfSend the sendRef
						sendRef.removeNode();
						copyOfNewNodeOfSend.insertAsChild(sendRef);

						// if the old father of send is empty then we delete it
						// (we do this recursivelly)
						UsefulMethods.deleteRecursivellyAncestorNewNodesWithNoChildren(newNodeOfSend);

						// we run recursivelly the algorithm for send's descendants
						checkAndMoveSendsWhichAreSendingDataToTheDelegatedSubTree(givenADoc, (Service)sendRef.getFather());
					}
				}
			}
		}
	}
	
	private List<Service> performDelegation(ADocument givenADoc, Service serviceCallToBeDelegated, String delegationPeer)
	{
		List<Service> receiveNewNodeAndServiceCallToBeDelegatedRefContainer = new LinkedList<Service>();
		
		// If we try to delegate a special service then we abort
		if(serviceCallToBeDelegated.getIsSpecialService() == true)
		{
			System.out.println("The givenService that was sent to the delegationRule must not be a special service");
			return null;
		}
		
		// Creating the receive service that is going to receive the results
		// which would be transmitted by the following send
		Service receiveService = new Service();
		receiveService.setServiceName(new StringBuffer("receive"));
		receiveService.setIsSpecialService(true);
		// the execution peer of receive sc will be the calling peer of serviceCallToBeDelegated
		String executionPeerOfReceive = UsefulMethods.returnCallingPeerOfGivenService(givenADoc, serviceCallToBeDelegated);
		receiveService.setExecutionPeer(new StringBuffer(executionPeerOfReceive));
		
		// NOTE receive description will have the same original ids
		// as the delegated subtree. 
		Service descriptionForReceive = new Service(serviceCallToBeDelegated, true, true);
		receiveService.insertAsChild(descriptionForReceive);
		
		serviceCallToBeDelegated.getFather().insertAsChildNextToGivenChild(receiveService, serviceCallToBeDelegated);

		// Creating the newNode sc that is going to create
		// a send sc at a remote peer
		Service newNodeServiceCall = new Service();
		newNodeServiceCall.setServiceName(new StringBuffer("newNode"));
		newNodeServiceCall.setIsSpecialService(true);
		newNodeServiceCall.setExecutionPeer(new StringBuffer(delegationPeer));
		// Creating the arguments of the newNode service call
		
			// The argument of the newNode service call would be a send service call
			// that would be executed at the remote peer. This send sc will execute
			// first of all the delegated sc and then it would send back the results
			// of the execution. 
			Service sendServiceCall = new Service();
			sendServiceCall.setServiceName(new StringBuffer("send"));
			sendServiceCall.setIsSpecialService(true);
			sendServiceCall.setExecutionPeer(new StringBuffer(delegationPeer));
			// Creating the arguments of the send service
			
			// Removing serviceCallToBeDelegated from its old position
			serviceCallToBeDelegated.removeNode();
			
			// Adding as first argument the serviceCallToBeDelegated
			sendServiceCall.insertAsChild(serviceCallToBeDelegated);

			// Setting the afterActivated for the send service
			sendServiceCall.getAfterActivated().add(serviceCallToBeDelegated.getCurrNodeID().toString());
//			sendServiceCall.setFather(newNodeServiceCall);

			// Now we are going to constract the send's reference to the recieve sc
			IdContainer sendDestinationAddressIdContainer = new IdContainer(receiveService.getOrigNodeID().toString(), receiveService.getOrigDocID().toString(), receiveService.getOrigPeerID().toString());
			IdWithHostContainer sendDestinationAddressIdWithHostContainer = new IdWithHostContainer(sendDestinationAddressIdContainer, receiveService.getExecutionPeer().toString());
			SimpleNode sendDestinationAddress = new SimpleNode(new StringBuffer(), sendDestinationAddressIdWithHostContainer, new LinkedList<StringBuffer>());				
			sendServiceCall.insertAsChild(sendDestinationAddress);
			
		// Adding the above as argument to newNode sc
		newNodeServiceCall.insertAsChild(sendServiceCall);
			
		// Here we have to set at the receive the send's original id
		IdContainer receiveReceivingAddressIdContainer = new IdContainer(sendServiceCall.getOrigNodeID().toString(), sendServiceCall.getOrigDocID().toString(), sendServiceCall.getOrigPeerID().toString());
		IdWithHostContainer receiveReceivingAddressIdWithHostContainer = new IdWithHostContainer(receiveReceivingAddressIdContainer, sendServiceCall.getExecutionPeer().toString());
		SimpleNode receiveReceivingAddress = new SimpleNode(new StringBuffer(), receiveReceivingAddressIdWithHostContainer, new LinkedList<StringBuffer>());
		receiveReceivingAddress.setFather(receiveService);
		// We add to the receive's arguments the sends address
		receiveService.insertAsChild(receiveReceivingAddress);
		
		// We add newNode sc as a child of an ancestor newNode (of the receive) or as 
		// child of the root of the document if the no newNode ancestor
		// is found
		ADNode ancestorOfTheReceive = UsefulMethods.searchForTheClosestAncestorOfGivenNodeThatIsUnderNewNodeOrRoot(givenADoc, receiveService);
		
		// We set the newNodeServiceCall as a child of the ancestorOfTheReceive
		ancestorOfTheReceive.insertAsChild(newNodeServiceCall);
		
		// We update originalIdsOfNodesNotUnderReceivesToADNodeRefs map of
		// givenADoc with the values of the new send and receive
		givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(sendDestinationAddressIdContainer, receiveService);
		givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.put(receiveReceivingAddressIdContainer, sendServiceCall);
		
		// We update also the OriginalIdWithHostAddressContainersThatPointToThisNode
		// of the new send and the new receive
		receiveService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(sendDestinationAddress);
		sendServiceCall.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(receiveReceivingAddress);
		
		// We update the OriginalIdWithHostAddressContainersThatPointToThisNode references
		// of serviceCallToBeDelegated to symbolically (through originalID) point to the
		// receive service of the delegation. OriginalIdWithHostAddressContainersThatPointToThisNode
		// of serviceCallToBeDelegated has to be deleted and the same pointers have to be added
		// in OriginalIdWithHostAddressContainersThatPointToThisNode of the receive
		
		// We also keep the SCs that symbolically pointed to the delegated sc (send & recieves). In the
		// loop bellow we are going to change this pointer to logically point to the receive of the delegation.
		// The service calls which depend on these sends & receives (which have these send & recieves in thier
		// what part, have to be updated).
		LinkedList<Service> updatedServiceCalls = new LinkedList<Service>();
		if(!serviceCallToBeDelegated.getOriginalIdWithHostAddressContainersThatPointToThisNode().isEmpty())
		{
			Iterator<SimpleNode> originalIdWithHostAddressContainersThatPointToScToBeDelegIter = serviceCallToBeDelegated.getOriginalIdWithHostAddressContainersThatPointToThisNode().iterator();
			while(originalIdWithHostAddressContainersThatPointToScToBeDelegIter.hasNext())
			{
				// we find the node that was pointing to the simple delegated sc (this would be
				// the pointer of a send or receive)
				SimpleNode origIdAddressContPointingToServiceCallToBeDelegated = originalIdWithHostAddressContainersThatPointToScToBeDelegIter.next();
				IdWithHostContainer addrWithHost  = origIdAddressContPointingToServiceCallToBeDelegated.getRemoteOriginalIdAndHostingPeer();
				// the father of origIdAddressContPointingToServiceCallToBeDelegated is a send or receive with
				// a pointer that will be changed. We are interested in this sc
				updatedServiceCalls.add((Service)origIdAddressContPointingToServiceCallToBeDelegated.getFather());
				
				// we change that pointer to point now (symbolically) to the receive. The hosting peer remains the
				// same but what is changing is the origin iD.
				IdContainer receivesOrigId = new IdContainer(receiveService.getOrigNodeID().toString(), receiveService.getOrigDocID().toString(), receiveService.getOrigPeerID().toString());
				addrWithHost.setID(receivesOrigId);
				
				// now it is no longer pointing to the serviceCallToBeDelegated but to the receiveService
				// so the OriginalIdWithHostAddressContainersThatPointToThisNode of these sc have to be updated
				// ...we remove from serviceCallToBeDelegated's
				originalIdWithHostAddressContainersThatPointToScToBeDelegIter.remove();
				// ...we add to receiveService's
				receiveService.getOriginalIdWithHostAddressContainersThatPointToThisNode().add(origIdAddressContPointingToServiceCallToBeDelegated);
			}
			
			// we remove serviceCallToBeDelegated from the originalIdsOfNodesNotUnderReceivesToADNodeRefs 
			// records of givenADoc (no one is pointing to this simple service any more)
			IdContainer serviceCallToBeDelegatedId = new IdContainer(serviceCallToBeDelegated.getOrigNodeID().toString(), serviceCallToBeDelegated.getOrigDocID().toString(), serviceCallToBeDelegated.getOrigPeerID().toString());
			givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.remove(serviceCallToBeDelegatedId);
		}
		
		// Here we are going to update the receive descriptions that are related with the updatedServiceCalls
		for(Iterator<Service> updatedServiceCallsIter = updatedServiceCalls.iterator(); updatedServiceCallsIter.hasNext();)
		{
			Service currUpdatedServiceCall = updatedServiceCallsIter.next();
			UsefulMethods.ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(givenADoc, currUpdatedServiceCall);
		}
		
		// We update the mapping tables in order the currNodeId of the added receive to map to the 
		// old currNodeId of the delegated service  
		String scToBeDelegatedOldCurrNodeId = givenADoc.newScCurrNodeIdsToOld.get(serviceCallToBeDelegated.getCurrNodeID().toString());
		if(scToBeDelegatedOldCurrNodeId != null)
		{
			givenADoc.newScCurrNodeIdsToOld.remove(serviceCallToBeDelegated.getCurrNodeID().toString());
			givenADoc.oldScCurrNodeIdsToNew.remove(scToBeDelegatedOldCurrNodeId);
			
			givenADoc.newScCurrNodeIdsToOld.put(receiveService.getCurrNodeID().toString(), scToBeDelegatedOldCurrNodeId);
			givenADoc.oldScCurrNodeIdsToNew.put(scToBeDelegatedOldCurrNodeId, receiveService.getCurrNodeID().toString());
		}
		
		// We update the some statistic data that givenADoc keeps
		givenADoc.increaseNumberOfRewritings();
		givenADoc.increaseNumberOfDelegations();
		
		// We insert the receive and newNode sc references in receiveAndNewNodeRefContainer
		receiveNewNodeAndServiceCallToBeDelegatedRefContainer.add(receiveService);
		receiveNewNodeAndServiceCallToBeDelegatedRefContainer.add(newNodeServiceCall);
		receiveNewNodeAndServiceCallToBeDelegatedRefContainer.add(serviceCallToBeDelegated);
		return receiveNewNodeAndServiceCallToBeDelegatedRefContainer;
	}
	

	// This method searches for delegations at the first services that you find
	// while descending the subtree rooted by givenService
	private Set<String> findFirstDelegationPeersOfImmediateDescendantScs(ADocument givenADoc, Service givenService)
	{
		Set<String> delegationPeersToBeReturned = new HashSet<String>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenService);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currADNode = nodesToBeVisited.removeFirst();
			if(currADNode instanceof SimpleNode)
				nodesToBeVisited.addAll(currADNode.getChildren());
			else if(currADNode instanceof Service)
			{
				Service currService = (Service) currADNode;
				if(currService.getServiceName().toString().equals("receive"))
				{
					SimpleNode sendAddr = (SimpleNode) currService.getChildren().get(currService.getChildren().size() - 1);
					Service sendRef = (Service) givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(sendAddr.getRemoteOriginalIdAndHostingPeer().id);
					delegationPeersToBeReturned.add(sendRef.getExecutionPeer().toString());
				}
				// send & newNode scs are not taken into account
				else if(currService.getServiceName().toString().equals("send"));
				else if(currService.getServiceName().toString().equals("newNode"));
				// else if it is a simple sc we continue the search
				// in its descendants
				else
					nodesToBeVisited.addAll(currService.getChildren());
			}
		}
		
		return delegationPeersToBeReturned;
	}
	
	// we search for all the simple scs that are "under" the givenSc. newNode nodes and sends
	// are ignored (and we don't search at their descendants). receives are used to find the
	// respective send and continue our search from their what part. Whenever we encounter a
	// simple sc we put it in the toBeReturned list and we don't search at its descendants.
	private List<Service> getTheFirstSimpleServiceCallDescendantsFound(ADNode givenADNode)
	{
		if(givenADNode instanceof Service && ((Service)givenADNode).getIsSpecialService())
			return new LinkedList<Service>();
		
		LinkedList<Service> toBeReturnedServiceCalls = new LinkedList<Service>();
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.addAll(givenADNode.getChildren());
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				
				if(serviceCurrNode.getServiceName().toString().equals("receive"))
				{
					// we are going to find the respective send of this receive
					
					// we get the second child of the recieve
					SimpleNode secondChildOfReceive = (SimpleNode) serviceCurrNode.getChildren().get(1);
					// we "extract" from it the sends originalId
					IdContainer sendsOriginalId = secondChildOfReceive.getRemoteOriginalIdAndHostingPeer().id;
					// we locate the send
					Service sendRef = (Service) serviceCurrNode.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(sendsOriginalId);
					// we get the what part (1st child) of send
					ADNode whatPartOfSend = sendRef.getChildren().get(0);
					// if it is a service node then we put it directly
					// in nodesToBeVisited
					if(whatPartOfSend instanceof Service)
						nodesToBeVisited.add(whatPartOfSend);
					// if not...
					else if(whatPartOfSend instanceof SimpleNode)
					{
						SimpleNode simpleNodeWhatPartOfSend = (SimpleNode) whatPartOfSend;
						// we check if it is addressContainer (pointing to
						// another service call in order to get data from 
						// that sc)
						if(simpleNodeWhatPartOfSend.getNodeData().length() == 0 && !simpleNodeWhatPartOfSend.getRemoteOriginalIdAndHostingPeer().isEmpty())
						{
							// we locate the service call
							IdContainer pointedServiceAddress = simpleNodeWhatPartOfSend.getRemoteOriginalIdAndHostingPeer().id;
							Service pointedService = (Service) simpleNodeWhatPartOfSend.getReferenceToADocument().originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(pointedServiceAddress);
							nodesToBeVisited.add(pointedService); // even if pointedService is a receive it is going to be resolved in a following step
						}
						// if it is just a simpleNode (not an addressContainer)
						else
							nodesToBeVisited.add(simpleNodeWhatPartOfSend);
					}
				}
				// we are not interested in searching under these nodes
				else if(serviceCurrNode.getServiceName().toString().equals("send"));
				else if(serviceCurrNode.getServiceName().toString().equals("newNode"));
				// if it is a simple service call
				else
				{
					// we do not search bellow but we keep the service call
					toBeReturnedServiceCalls.add(serviceCurrNode);
				}
			}
			// if it is a simple node we add its children in nodesToBeVisited
			// and we continue our search
			else
				nodesToBeVisited.addAll(currNode.getChildren());
		}
		return toBeReturnedServiceCalls;
	}
	
	// This method checks if givenSc can be delegated to givenDelegationDest. In order to do this, it checks
	// 1) if there is a conflict of givenDelegationDest with the previous delegations of givenSc until now
	// 2) if there is a conflict with the calling peer of the father sc of givenSc
	// 3) if there is a conflict with the delegations of the first descendant simple scs 
	private boolean checkIfGivenDelegationsWillCauseProblemToGivenServiceOrToItsSubTree(Service givenSc, String givenDelegationDest)
	{
		if(givenSc.getReferenceToADocument().getADocNumber() == 24 && givenSc.getCurrNodeID().toString().equals("18") && givenDelegationDest.equals("peer2"))
		{
			System.out.println("lala");
		}
		
		// here we are going to store the original calling peer
		// of givenService
		String originalCallingPeerOfGivenSc;
		
		// here we are going to find the calling peer of the previous simple sc ancestor
		Service simpleFatherScOfServiceToBeDelegated = UsefulMethods.findSimpleScFatherOfGivenSc(givenSc.getReferenceToADocument(), givenSc);
		String callingPeerOfPreviousNormalScAncestor = null;
		// if is has a simple service as father or ancestor
		if(simpleFatherScOfServiceToBeDelegated != null)
			callingPeerOfPreviousNormalScAncestor =  UsefulMethods.returnCallingPeerOfGivenService(simpleFatherScOfServiceToBeDelegated.getReferenceToADocument(), simpleFatherScOfServiceToBeDelegated);
		// if not...
		else
			callingPeerOfPreviousNormalScAncestor = givenSc.getReferenceToADocument().getPeerThatHostsADoc().toString();
		
		// the original calling peer of givenService (the calling peer of givenService
		// without any delegations) is the current callingPeerOfPreviousNormalScAncestor.
		// So...
		originalCallingPeerOfGivenSc = callingPeerOfPreviousNormalScAncestor;
		
		if(checkIfDelegationsMakeALoopForGivenSc(givenSc, originalCallingPeerOfGivenSc, givenDelegationDest) == true)
			return true;
		
		List<Service> serviceCallsThatWouldBeAffectedByTheDeleg = getTheFirstSimpleServiceCallDescendantsFound(givenSc);
		while(!serviceCallsThatWouldBeAffectedByTheDeleg.isEmpty())
		{
			// for each of the scs that would be affected by the delegation we ckeck if 
			// a problem will occur
			Service toBeAffectedServiceCall = serviceCallsThatWouldBeAffectedByTheDeleg.remove(0);
			if(checkIfDelegationsMakeALoopForGivenSc(toBeAffectedServiceCall, givenDelegationDest, null) == true)
				return true;
		}
		
		return false;
	}
	
	// This method checks if the delegations of givenSc make a circle or not. We can also give an additionalDelegation
	// to be added to the already existing and found delegs of givenSc.
	// If we want generally to check if a delegation of givenSc can be made (and calculate all the conflicts), we have
	// to call the method checkIfGivenDelegationsWillCauseProblemToGivenServiceOrToItsSubTree
	private boolean checkIfDelegationsMakeALoopForGivenSc(Service givenSc, String originalCallingPeerOfGivenSc, String additionalDelegation)
	{
		Set<String> delegationSet = new HashSet<String>();
		delegationSet.add(originalCallingPeerOfGivenSc);

		List<String> delegationOfGivenScUntilNow = findDelegationPeersOfCurrScUntilPreviousNormalScAncestor(givenSc);
		Iterator<String> delegationsOfGivenScUntilNowIter = delegationOfGivenScUntilNow.iterator();
		while(delegationsOfGivenScUntilNowIter.hasNext())
		{
			String delegPeer = delegationsOfGivenScUntilNowIter.next();
			// we add the found peers in the delegationSet and if the set
			// already contains this peer a loop is discovered
			if(delegationSet.add(delegPeer) == false)
				return true;
		}
		
		// we add the proposed additional delegation for givenSc.
		// we check if a loop will occur if we apply this deleg too.
		if(additionalDelegation != null && delegationSet.add(additionalDelegation) == false)
			return true;

		// if we reached this point then there is no loop
		return false;
	}
	
	// This method finds all the peers where the givenService was delegated.
	private List<String> findDelegationPeersOfCurrScUntilPreviousNormalScAncestor(Service givenService)
	{
		List<String> delegationPeers = new LinkedList<String>();
		
		// We search for the ancestor service call of the current service call
		ADNode tempADNode = givenService.getFather();
		while(tempADNode != null && !(tempADNode instanceof Service))
			tempADNode = tempADNode.getFather();
		
		while(tempADNode != null)
		{
			if(tempADNode instanceof Service)
			{
				Service serviceTempADNode = (Service) tempADNode;
				if(serviceTempADNode.getServiceName().toString().equals("send"))
				{
					// A delegation is detected by finding a send that is pointing
					// to a receive (and both of them have different execution peer - 
					// if they have the same execution peer logically is a wrong
					// delegation)
					Service sendRef = serviceTempADNode;
					SimpleNode addressOfTheRespectiveReceive = (SimpleNode) sendRef.getChildren().get(serviceTempADNode.getChildren().size() - 1);
					ADocument currADoc = addressOfTheRespectiveReceive.getReferenceToADocument();
					Service receiveRef = (Service) currADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(addressOfTheRespectiveReceive.getRemoteOriginalIdAndHostingPeer().id);
					
					if(receiveRef.getServiceName().toString().equals("receive") && !receiveRef.getExecutionPeer().toString().equals(sendRef.getExecutionPeer().toString()))
					{
						// a delegation is detected
						delegationPeers.add(sendRef.getExecutionPeer().toString());
					}
					
					tempADNode = receiveRef.getFather();
				}
				// in a valid delegation we should find first send
				// and as a parent of send newNode. So we assume that
				// there is no delegation here
				else if(serviceTempADNode.getServiceName().equals("newNode"))
					return delegationPeers;
				
				// We are searching for delegations in receive's description
				// We have done a big mistake
				else if(serviceTempADNode.getServiceName().equals("receive"))
					return null;
				
				// if it is another simple service call
				else
					return delegationPeers;
			}
			else
				tempADNode = tempADNode.getFather();
		}
		
		// if tempADNode == null, the root of the document
		// was found
		return delegationPeers;
	}
}
