package fr.inria.gemo.axml.optimax.rules;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import fr.inria.gemo.axml.optimax.Parameters;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.info.PeerInformationModule;

public class Instantiation {

	public List<ADocument> instatiationForADocs(ADocument givenADoc)
	{
		List<ADocument> returningADocs = new ArrayList<ADocument>();
		PeerInformationModule peerInformationModuleInst = PeerInformationModule.getInstance();
		
		List<Service> serviceCallsOfADoc = UsefulMethods.returnServicesUnderneathASpecificNodeStopOnReceive(givenADoc.getRootOfADoc());
		// we delete the service calls that are special scs or
		// they have execution peer different than Parameters.ANY
		Iterator<Service> serviceCallIter = serviceCallsOfADoc.iterator();
		while(serviceCallIter.hasNext())
		{
			Service currSc = serviceCallIter.next();
			if(currSc.getIsSpecialService() || !currSc.getExecutionPeer().toString().equals(Parameters.ANY))
				serviceCallIter.remove();
		}
		
		// we will traverse the serviceCallsOfADoc and for each of the peers left
		// we will try to instantiate one service call at a time. For each instatiation
		// we create a new adoc and we put it in the returningADocs list
		serviceCallIter = serviceCallsOfADoc.iterator();
		while(serviceCallIter.hasNext())
		{
			Service currSc = serviceCallIter.next();
			
			// for each possible execution peer
			List<StringBuffer> listOfPeers = peerInformationModuleInst.getListOfPeers();
			Iterator<StringBuffer> listOfPeersIter = listOfPeers.iterator();
			while(listOfPeersIter.hasNext())
			{
				StringBuffer currPeer = listOfPeersIter.next();
				
				// we create a new doc
				ADocument newADoc = new ADocument(givenADoc);
				Service newCurrSc = (Service) ADocument.returnRespectiveReferenceToNewADoc(givenADoc, newADoc, currSc);
				
				// we change the execution peer of newCurrSc to currPeer
				newCurrSc.setExecutionPeer(new StringBuffer(currPeer));
				
				// we ascend the tree and we search for receives that might need updating
				UsefulMethods.ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(newADoc, newCurrSc);
				
				// calculate network cost from scratch
				newADoc.calculateNetworkCost();
				
				// We add at the newADoc the rule that we applied
				boolean decreaseCost = false;
				boolean increaseCost = false;

				if(givenADoc.getSimpleNetworkCost() < newADoc.getSimpleNetworkCost())
					increaseCost = true;

				if(givenADoc.getSimpleNetworkCost() > newADoc.getSimpleNetworkCost())
					decreaseCost = true;
				
				newADoc.insertAppliedRuleToADoc("inst", givenADoc.getADocNumber(), newCurrSc.getCurrNodeID().toString(), newCurrSc.getExecutionPeer().toString(), newADoc.getADocNumber(), increaseCost, decreaseCost, givenADoc.getSimpleNetworkCost(), newADoc.getSimpleNetworkCost());
				
				returningADocs.add(newADoc);
			}
		}
		
		return returningADocs;
	}
	
}
