package fr.inria.gemo.axml.optimax.rules;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import fr.inria.gemo.axml.optimax.adocument.ADNode;
import fr.inria.gemo.axml.optimax.adocument.ADocument;
import fr.inria.gemo.axml.optimax.adocument.IdWithHostContainer;
import fr.inria.gemo.axml.optimax.adocument.Service;
import fr.inria.gemo.axml.optimax.adocument.SimpleNode;
import fr.inria.gemo.axml.optimax.helping.ExtendedLinkedList;

public class UsefulMethods {

	static void ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(ADocument givenADoc, ADNode givenADNode)
	{
		ADNode currNode = givenADNode;
		while(currNode != null)
		{
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				// if we find a send while ascending the tree
				if(serviceCurrNode.getServiceName().toString().equals("send"))
				{
					// if it is pointing to a simple service call by its what part (first arg)
					// then no update needs to be done
					ADNode firstChildOfSend = serviceCurrNode.getChildren().get(0);
					if(firstChildOfSend instanceof SimpleNode)
					{
						SimpleNode simpleNodeFirstChildOfSend = (SimpleNode) firstChildOfSend;
						if(simpleNodeFirstChildOfSend.getNodeData().length() == 0 && !simpleNodeFirstChildOfSend.getRemoteOriginalIdAndHostingPeer().isEmpty())
						{
							currNode = currNode.getFather();
							continue;
						}
					}
					
					// we get the address of the corresponding receive
					SimpleNode addressContainer = (SimpleNode) serviceCurrNode.getChildren().get(serviceCurrNode.getChildren().size() - 1);
					IdWithHostContainer originalIdWithHostAddressOfReceive = addressContainer.getRemoteOriginalIdAndHostingPeer();
					
					// we locate the coresponding receive
					Service receiveRef = (Service) givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(originalIdWithHostAddressOfReceive.id);
					
					// we update the specific receive's what part with the send's what
					// part 
					//NOTE (we assume that the send has only two children. Its what part and 
					//NOTE the respective receive's address)
					//NOTE (we append a copy of serviceCurrNode.getChildren().get(0) to
					// the what part of the receive)
					ADNode whatPartOfTheSend = serviceCurrNode.getChildren().get(0);
					ADNode copyOfTheWhatPartOfSend = null;
					if(whatPartOfTheSend instanceof SimpleNode)
						copyOfTheWhatPartOfSend = new SimpleNode((SimpleNode)serviceCurrNode.getChildren().get(0), true, true);
					else if(whatPartOfTheSend instanceof Service)
						copyOfTheWhatPartOfSend = new Service((Service)serviceCurrNode.getChildren().get(0), true, true);
					
					appendNewWhatPartToReceive(receiveRef, copyOfTheWhatPartOfSend);
					ascendTheTreeAndIfYouFindASendChangeTheWhatPartOfTheCorrespondingRecieve(givenADoc, receiveRef);
				}
			}
			currNode = currNode.getFather();
		}
	}
	
	// The following method deletes all the chidren of givenReceive sc (except the respective
	// send's address) and adds the rootOfNewWhatPart
	private static void appendNewWhatPartToReceive(Service givenReceive, ADNode rootOfNewWhatPart)
	{
		List<ADNode> childrenOfGivenReceive = givenReceive.getChildren();
		SimpleNode sendsAddressContainer = (SimpleNode) childrenOfGivenReceive.get(childrenOfGivenReceive.size() - 1);
		sendsAddressContainer.removeNode();
		childrenOfGivenReceive.get(0).removeNode();
		
		givenReceive.insertAsChild(rootOfNewWhatPart);
		givenReceive.insertAsChild(sendsAddressContainer);
	}
	
	
	// This method will find the calling peer of a given service
	static String returnCallingPeerOfGivenService(ADocument givenADoc, Service givenService)
	{
		// We search for the ancestor service call of the current service call
		ADNode tempADNode = givenService.getFather();
		while(tempADNode != null && !(tempADNode instanceof Service))
			tempADNode = tempADNode.getFather();
		
		while(tempADNode != null)
		{
			if(tempADNode instanceof Service)
			{
				Service serviceTempADNode = (Service) tempADNode;
				// if we find send service call, then the execution peer of
				// send is the calling peer of the service
				if(serviceTempADNode.getServiceName().toString().equals("send"))
					return serviceTempADNode.getExecutionPeer().toString();
				// if we find newNode service call, then the execution peer of
				// new node is the calling peer of the service
				else if(serviceTempADNode.getServiceName().toString().equals("newNode"))
					return serviceTempADNode.getExecutionPeer().toString();
				
				// We are searching for delegations in receive's description
				// We have done a big mistake
				else if(serviceTempADNode.getServiceName().toString().equals("receive"))
					return null;
			}
			
			// If it is another simple service call or a simple node
			// then we proceed to the father
			tempADNode = tempADNode.getFather();
		}
		
		// if we reached the root of the givenADoc, then the calling peer
		// is the hosting peer of givenADoc
		return givenADoc.getPeerThatHostsADoc().toString();
	}
	
	// This methor returns a reference to the father (ancestor) service call of givenService
	// If father is a simple service then null is returned
	static Service findSimpleScFatherOfGivenSc(ADocument givenADoc, Service givenService)
	{
		// We search for the ancestor service call of the current service call
		ADNode tempADNode = givenService.getFather();
		while(tempADNode != null && !(tempADNode instanceof Service))
			tempADNode = tempADNode.getFather();
		
		while(tempADNode != null)
		{
			if(tempADNode instanceof Service)
			{
				Service serviceTempADNode = (Service) tempADNode;
				if(serviceTempADNode.getServiceName().toString().equals("send"))
				{
					SimpleNode addressOfTheRespectiveReceive = (SimpleNode) serviceTempADNode.getChildren().get(serviceTempADNode.getChildren().size() - 1);
					tempADNode = givenADoc.originalIdsOfNodesNotUnderReceivesToADNodeRefs.get(addressOfTheRespectiveReceive.getRemoteOriginalIdAndHostingPeer().id);
					tempADNode = tempADNode.getFather();
				}
				// in a valid delegation we should find first send
				// and as a parent of send newNode. So we assume that
				// there is no delegation here. So the givenService
				// does not have a simple Sc father
				else if(serviceTempADNode.getServiceName().toString().equals("newNode"))
					return null;
				
				// We are searching for delegations in receive's description
				// We have done a big mistake
				else if(serviceTempADNode.getServiceName().toString().equals("receive"))
					return null;
				
				// if it is another simple service call
				else
					return serviceTempADNode;
			}
			else
				tempADNode = tempADNode.getFather();
		}
		return null;
	}
	
	static ADNode searchForTheClosestAncestorOfGivenNodeThatIsUnderNewNodeOrRoot(ADocument givenADoc, ADNode givenADNode)
	{
		// We add newNode sc as a child of an ancestor newNode (of the receive) or as 
		// child of the root of the document if the no newNode ancestor
		// is found
		ADNode closestAncestorOfGivenADNode = givenADNode.getFather();
		boolean newNodeFound = false;
		while(closestAncestorOfGivenADNode != null && newNodeFound == false)
		{
			while(closestAncestorOfGivenADNode != null && !(closestAncestorOfGivenADNode instanceof Service))
				closestAncestorOfGivenADNode = closestAncestorOfGivenADNode.getFather();
			
			// if it is not null then it is a service, so we will check
			// if it is a newNode service
			if(closestAncestorOfGivenADNode != null)
			{
				Service serviceAncestorOfTheReceive = (Service) closestAncestorOfGivenADNode;
				if(serviceAncestorOfTheReceive.getServiceName().toString().equals("newNode"))
				{
					newNodeFound = true;
					break;
				}
				else
					closestAncestorOfGivenADNode = closestAncestorOfGivenADNode.getFather();	
			}
		}
		
		// If we reached the root of the document then we will
		// return the root of the doc 
		if(closestAncestorOfGivenADNode == null)
			closestAncestorOfGivenADNode = givenADoc.getRootOfADoc();
		
		return closestAncestorOfGivenADNode;
	}
	
	static ExtendedLinkedList<ADNode> getAllTheNodesOfSubTree(ADNode givenNode)
	{
		ExtendedLinkedList<ADNode> nodesToBeReturned = new ExtendedLinkedList<ADNode>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currADNode = nodesToBeVisited.removeFirst();
			nodesToBeReturned.add(currADNode);
			nodesToBeVisited.addAll(currADNode.getChildren());
		}
		
		return nodesToBeReturned;
	}
	
	static ExtendedLinkedList<Service> returnServicesUnderneathASpecificNodeStopOnNewNodeAndReceive(ADNode givenNode)
	{
		ExtendedLinkedList<Service> theScsToBeReturned = new ExtendedLinkedList<Service>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				theScsToBeReturned.add(serviceCurrNode);

				if(serviceCurrNode.getServiceName().toString().equals("newNode") || serviceCurrNode.getServiceName().toString().equals("receive"))
					continue;
			}
			nodesToBeVisited.addAll(currNode.getChildren());
		}
		
		return theScsToBeReturned;
	}
	
	public static ExtendedLinkedList<Service> returnServicesUnderneathASpecificNodeStopOnReceive(ADNode givenNode)
	{
		ExtendedLinkedList<Service> theScsToBeReturned = new ExtendedLinkedList<Service>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				theScsToBeReturned.add(serviceCurrNode);

				if(serviceCurrNode.getServiceName().toString().equals("receive"))
					continue;
			}
			nodesToBeVisited.addAll(currNode.getChildren());
		}
		
		return theScsToBeReturned;
	}
	
	static List<Service> returnReceivesUnderneathASpecificNodeStopOnNewNodeAndReceive(ADNode givenNode)
	{
		List<Service> theReceiveScsToBeReturned = new LinkedList<Service>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				if(serviceCurrNode.getServiceName().toString().equals("receive"))
				{
					theReceiveScsToBeReturned.add(serviceCurrNode);
					continue;
				}
				else if(serviceCurrNode.getServiceName().toString().equals("newNode"))
					continue;
			}
			nodesToBeVisited.addAll(currNode.getChildren());
		}
		
		return theReceiveScsToBeReturned;
	}
	
	static List<Service> returnSendsReceivesUnderneathASpecificNodeStopOnNewNodeAndReceive(ADNode givenNode)
	{
		List<Service> theSendReceiveScsToBeReturned = new LinkedList<Service>();
		
		LinkedList<ADNode> nodesToBeVisited = new LinkedList<ADNode>();
		nodesToBeVisited.add(givenNode);
		while(!nodesToBeVisited.isEmpty())
		{
			ADNode currNode = nodesToBeVisited.removeFirst();
			if(currNode instanceof Service)
			{
				Service serviceCurrNode = (Service) currNode;
				if(serviceCurrNode.getServiceName().toString().equals("receive") || serviceCurrNode.getServiceName().toString().equals("send"))
				{
					theSendReceiveScsToBeReturned.add(serviceCurrNode);
					continue;
				}
				else if(serviceCurrNode.getServiceName().toString().equals("newNode"))
					continue;
			}
			nodesToBeVisited.addAll(currNode.getChildren());
		}
		
		return theSendReceiveScsToBeReturned;
	}
	
	static void deleteRecursivellyAncestorNewNodesWithNoChildren(Service givenNewNode)
	{
		if(givenNewNode != null && givenNewNode.getServiceName().toString().equals("newNode"))
		{
			ADNode fatherOfNewNode = givenNewNode.getFather();
			
			// if the givenNewNode does not have any children
			// then we remove it from the system
			if(givenNewNode.getChildren().isEmpty())
				givenNewNode.removeNode();
			else
				return;
			
			if(fatherOfNewNode != null && fatherOfNewNode instanceof Service)
			{
				Service serviceFatherOfNewNode = (Service) fatherOfNewNode;
				if(serviceFatherOfNewNode.getServiceName().toString().equals("newNode"))
					deleteRecursivellyAncestorNewNodesWithNoChildren(serviceFatherOfNewNode);
			}
		}
	}
}
